/* ========================================================================= //
#?
#? NAME
#?      EnDeMaps.js
#?
#? SYNOPSIS
#?      <SCRIPT language="javascript" type="text/javascript" src="EnDeMaps.js"></SCRIPT>
#?      <SCRIPT language="javascript" type="text/javascript" src="EnDeMap.js"></SCRIPT>
#?
#? DESCRIPTION
#?      Character maps for encoding, decoding and converting from and to plain
#?      text. Maps will be initialized in  EnDeMap.js.
#?
#? HACKER's INFO
#?      // ToDo: //
#       http://www.w3.org/TR/html401/sgml/entities.html
#           HTML character references -- aka HTML Entity
#           NCE - named character entity - aka character entity reference
#                 http://en.wikipedia.org/wiki/Character_entity_reference
#           NCR - numeric character references
#                 http://en.wikipedia.org/wiki/Numeric_character_reference
#           UCS - universal character set
#                 http://en.wikipedia.org/wiki/Universal_Character_Set
#                 BOM - Byte Order Mark
#                 BMP - Basic Multilingual Plane
#                 UCS-2, a single code value between 0 and 65535
#                 UTF-16, an extension of UCS-2, to represents code points
#                         outside the BMP
#                 UCS-4, a single code value between 0 and (theoretically)
#                         hexadecimal 7FFFFFFF
#                         UCS-4 thereby permits a binary representation of
#                         every code point in the UCS
#           ==> articles about Unicode will mistakenly refer to  UCS-2  as
#               "UCS-16". UCS-16 does not exist.
#
#           "Character 65533 (FFFD hexadecimal) is the last valid character in
#            UCS-2. 65534 (FFFE hexadecimal) is unassigned and reserved as the
#            byte-swapped version of  ZERO WIDTH NON-BREAKING SPACE  for byte-
#            order detection purposes. 65535 (FFFF hexadecimal) is unassigned."
#
#       CP-1252 (Windows-1252 Western European)
#           "Entities have also been added for the  remaining characters
#            occurring in CP-1252 which do not occur in the  HTMLlat1 or
#            HTMLsymbol entity sets. These all occur in the  128 to  159
#            range within the CP-1252 charset. These entities permit the
#            characters to be denoted in a platform-independent manner."
#            CP-1252 is same or similar to ANSI see also http://www.goascii.de/
#       See also: http://unicode.org/Public/MAPPINGS/VENDORS/MICSFT/WindowsBestFit/bestfit1252.txt
#
#       from: http://www.w3.org/TR/REC-html40/sgml/entities.html
#       Problems/Inconsistency:
#            #9825 vs 9829 hearts
#            #59146 b.epsiv und 59147 (da fehlt was)
#            NCR seen somewhere: gravea, cir, congr, ero, etago,
#
# ? // ToDo: http://unicode.org/Public/MAPPINGS/VENDORS/MISC/
# ?          http://unicode.org/Public/MAPPINGS/VENDORS/MICSFT/WindowsBestFit/bestfit1252.txt
# ?          http://www.unicode.org/charts/symbols.html
#?
#? SEE ALSO
#?      http://www.w3.org/TR/REC-MathML/chap6/bycodes.html
#?      EnDe.js
#?      EnDeMap.js
#?
#? VERSION
#?      @(#) EnDeMaps.js 1.30 09/06/14 14:01:51
#?
#? AUTHOR
#?      05-jun-07 Achim Hoffmann, mailto: EnDe (at) my (dash) stp (dot) net
#?
 * ========================================================================= */

/* vi:set ts=16 */

/*
 * NOTE: following variables are a string enclosed in ";
 * The variables start right after the comment // {_charsMapAll
 * and ends right before the comment // _charsMapAll}
 * This string does not (must not) contain any literal " and no literal \n .
 * The string should be read as table with 5 columns (excluding final \n),
 * where columns are separated by \t .
 * Array must be sorted numericaly according left most column.
 * Group names have to be separated by , .
 * The group name  :alias:  will be handled special in the
 * initialization functions.
 */
// {_charsMapAll
var _charsMapAll = "	\n\
	\n\
#	00	\n\
	\n\
#------+-------+---------------+---------------+----------      \n\
#                                  Character Name               \n\
# Int           standard        Entity          Group           Description	\n\
#------+-------+---------------+---------------+----------      \n\
0	-	-	-	NUL null character	\n\
1	-	-	-	SOH start of header	\n\
2	-	-	-	STX start of text	\n\
3	-	-	-	ETX end of text	\n\
4	-	-	-	EOT end of transmission	\n\
5	-	-	-	ENQ enquiry	\n\
6	-	-	-	ACK accknowledge	\n\
7	-	-	-	BEL bell; ring	\n\
8	-	-	-	BS  backspace	\n\
9	-	-	-	HT  horizontal tab	\n\
9	-	Tab	Lextra	tabulator stop; horizontal tabulation	\n\
10	-	NewLine	Lextra	force a line break; line feed	\n\
11	-	-	-	VT  vertical tabulation	\n\
12	-	-	-	FF  form feed	\n\
33	-	excl	ISOnum	exclamation mark	\n\
34	HTML-40	quot	ISOnum	quotation mark	\n\
35	-	num	ISOnum	number sign	\n\
36	-	dollar	ISOnum	dollar sign	\n\
37	-	percnt	ISOnum	percent sign	\n\
38	HTML-40	amp	ISOnum	ampersand	\n\
39	-	apos	ISOnum	apostrophe	\n\
40	-	lpar	ISOnum	left parenthesis	\n\
41	-	rpar	ISOnum	right parenthesis	\n\
42	-	midast	ISOamsb,:alias:	asterisk	\n\
42	-	ast	ISOnum	asterisk	\n\
43	-	plus	ISOnum	plus sign	\n\
44	-	comma	ISOnum	comma	\n\
45	-	hyphen	:alias:	hyphen	\n\
45	-	minus	:alias:	minus	\n\
46	-	period	ISOnum	full stop, period	\n\
47	-	sol	ISOnum	solidus	\n\
58	-	colon	ISOnum	colon	\n\
59	-	semi	ISOnum	semicolon	\n\
60	HTML-40	lt	ISOnum	less-than sign	\n\
61	-	equals	ISOnum	equals sign	\n\
62	HTML-40	gt	ISOnum	greater-than sign	\n\
63	-	quest	ISOnum	question mark	\n\
64	-	commat	ISOnum	commercial at	\n\
91	-	lbrack	:alias:	ISOnum lsqb	\n\
91	-	lsqb	ISOnum	left square bracket	\n\
92	-	bsol	ISOnum	reverse solidus (backslash)	\n\
93	-	rbrack	:alias:	ISOnum rsqb	\n\
93	-	rsqb	ISOnum	right square bracket	\n\
94	-	circ	ISOdia	circumflex accent	\n\
95	-	lowbar	ISOnum	low line	\n\
96	-	DiacriticalGrave	:alias:	ISOdia grave	\n\
96	-	grave	ISOdia	grave accent	\n\
109	-	-	-	m (for meter)	\n\
123	-	lbrace	:alias:	ISOnum lcub	\n\
123	-	lcub	ISOnum	left curly bracket	\n\
124	-	vert	:alias:	ISOnum verbar	\n\
124	-	verbar	ISOnum	vertical bar	\n\
125	-	rbrace	:alias:	ISOnum rcub	\n\
125	-	rcub	ISOnum	right curly bracket	\n\
126	-	tilde	:alias:	tilde	\n\
160	HTML-40	nbsp	ISOnum	no break (required) space	\n\
161	HTML-40	iexcl	ISOnum	inverted exclamation mark	\n\
162	HTML-40	cent	ISOnum	cent sign	\n\
163	HTML-40	pound	ISOnum	pound sign	\n\
164	HTML-40	curren	ISOnum	general currency sign	\n\
165	HTML-40	yen	ISOnum	yen sign	\n\
166	HTML-40	brvbar	ISOnum	broken (vertical) bar	\n\
167	HTML-40	sect	ISOnum	section sign	\n\
168	HTML-40	DoubleDot	:alias:	ISOdia die	\n\
168	HTML-40	die	ISOdia,:alias:	dieresis	\n\
168	HTML-40	uml	ISOdia,:alias:	umlaut mark	\n\
168	HTML-40	Dot	ISOtech	dieresis or umlaut mark	\n\
169	HTML-40	copy	ISOnum	copyright sign	\n\
170	HTML-40	ordf	ISOnum	ordinal indicator, feminine	\n\
171	HTML-40	laquo	ISOnum	angle quotation mark, left	\n\
172	HTML-40	not	ISOnum	neg; not sign	\n\
173	HTML-40	shy	ISOnum	soft hyphen	\n\
174	HTML-40	circledR	:alias:	ISOnum reg	\n\
174	HTML-40	reg	ISOnum	registered sign	\n\
175	HTML-40	OverBar	Lextra,:alias:	over bar	\n\
175	HTML-40	macr	ISOdia	macron	\n\
176	HTML-40	deg	ISOnum	degree sign	\n\
177	HTML-40	PlusMinus	:alias:	ISOnum plusmn	\n\
177	HTML-40	pm	:alias:	ISOnum plusmn	\n\
177	HTML-40	plusmn	ISOnum	plus-or-minus sign	\n\
178	HTML-40	sup2	ISOnum	superscript two	\n\
179	HTML-40	sup3	ISOnum	superscript three	\n\
180	HTML-40	DiacriticalAcute	:alias:	ISOdia acute	\n\
180	HTML-40	acute	ISOdia	acute accent	\n\
181	HTML-40	micro	ISOnum	micro sign	\n\
182	HTML-40	para	ISOnum	pilcrow (paragraph sign)	\n\
183	HTML-40	CenterDot	:alias:	ISOnum middot	\n\
183	HTML-40	centerdot	:alias:	ISOnum middot	\n\
183	HTML-40	middot	ISOnum	middle dot	\n\
184	HTML-40	Cedilla	:alias:	ISOdia cedil	\n\
184	HTML-40	cedil	ISOdia	cedilla	\n\
185	HTML-40	sup1	ISOnum	superscript one	\n\
186	HTML-40	ordm	ISOnum	ordinal indicator, masculine	\n\
187	HTML-40	raquo	ISOnum	angle quotation mark, right	\n\
188	HTML-40	frac14	ISOnum	fraction one-quarter	\n\
189	HTML-40	half	ISOnum,:alias:	fraction one-half	\n\
189	HTML-40	frac12	ISOnum	fraction one-half	\n\
190	HTML-40	frac34	ISOnum	fraction three-quarters	\n\
191	HTML-40	iquest	ISOnum	inverted question mark	\n\
192	HTML-40	Agrave	ISOlat1	capital A, grave accent	\n\
193	HTML-40	Aacute	ISOlat1	capital A, acute accent	\n\
194	HTML-40	Acirc	ISOlat1	capital A, circumflex accent	\n\
195	HTML-40	Atilde	ISOlat1	capital A, tilde	\n\
196	HTML-40	Auml	ISOlat1	capital A, dieresis or umlaut mark	\n\
197	HTML-40	Aring	ISOlat1	capital A, ring	\n\
198	HTML-40	AElig	ISOlat1	capital AE diphthong (ligature)	\n\
199	HTML-40	Ccedil	ISOlat1	capital C, cedilla	\n\
200	HTML-40	Egrave	ISOlat1	capital E, grave accent	\n\
201	HTML-40	Eacute	ISOlat1	capital E, acute accent	\n\
202	HTML-40	Ecirc	ISOlat1	capital E, circumflex accent	\n\
203	HTML-40	Euml	ISOlat1	capital E, dieresis or umlaut mark	\n\
204	HTML-40	Igrave	ISOlat1	capital I, grave accent	\n\
205	HTML-40	Iacute	ISOlat1	capital I, acute accent	\n\
206	HTML-40	Icirc	ISOlat1	capital I, circumflex accent	\n\
207	HTML-40	Iuml	ISOlat1	capital I, dieresis or umlaut mark	\n\
208	HTML-40	ETH	ISOlat1	capital Eth, Icelandic	\n\
209	HTML-40	Ntilde	ISOlat1	capital N, tilde	\n\
210	HTML-40	Ograve	ISOlat1	capital O, grave accent	\n\
211	HTML-40	Oacute	ISOlat1	capital O, acute accent	\n\
212	HTML-40	Ocirc	ISOlat1	capital O, circumflex accent	\n\
213	HTML-40	Otilde	ISOlat1	capital O, tilde	\n\
214	HTML-40	Ouml	ISOlat1	capital O, dieresis or umlaut mark	\n\
215	HTML-40	times	ISOnum	multiply sign	\n\
216	HTML-40	Oslash	ISOlat1	capital O, slash	\n\
217	HTML-40	Ugrave	ISOlat1	capital U, grave accent	\n\
218	HTML-40	Uacute	ISOlat1	capital U, acute accent	\n\
219	HTML-40	Ucirc	ISOlat1	capital U, circumflex accent	\n\
220	HTML-40	Uuml	ISOlat1	capital U, dieresis or umlaut mark	\n\
221	HTML-40	Yacute	ISOlat1	capital Y, acute accent	\n\
222	HTML-40	THORN	ISOlat1	capital THORN, Icelandic	\n\
223	HTML-40	szlig	ISOlat1	small sharp s, German (sz ligature)	\n\
224	HTML-40	agrave	ISOlat1	small a, grave accent	\n\
225	HTML-40	aacute	ISOlat1	small a, acute accent	\n\
226	HTML-40	acirc	ISOlat1	small a, circumflex accent	\n\
227	HTML-40	atilde	ISOlat1	small a, tilde	\n\
228	HTML-40	auml	ISOlat1	small a, dieresis or umlaut mark	\n\
229	HTML-40	aring	ISOlat1	small a, ring	\n\
230	HTML-40	aelig	ISOlat1	small ae diphthong (ligature)	\n\
231	HTML-40	ccedil	ISOlat1	small c, cedilla	\n\
232	HTML-40	egrave	ISOlat1	small e, grave accent	\n\
233	HTML-40	eacute	ISOlat1	small e, acute accent	\n\
234	HTML-40	ecirc	ISOlat1	small e, circumflex accent	\n\
235	HTML-40	euml	ISOlat1	small e, dieresis or umlaut mark	\n\
236	HTML-40	igrave	ISOlat1	small i, grave accent	\n\
237	HTML-40	iacute	ISOlat1	small i, acute accent	\n\
238	HTML-40	icirc	ISOlat1	small i, circumflex accent	\n\
239	HTML-40	iuml	ISOlat1	small i, dieresis or umlaut mark	\n\
240	HTML-40	eth	ISOlat1	small eth, Icelandic	\n\
241	HTML-40	ntilde	ISOlat1	small n, tilde	\n\
242	HTML-40	ograve	ISOlat1	small o, grave accent	\n\
243	HTML-40	oacute	ISOlat1	small o, acute accent	\n\
244	HTML-40	ocirc	ISOlat1	small o, circumflex accent	\n\
245	HTML-40	otilde	ISOlat1	small o, tilde	\n\
246	HTML-40	ouml	ISOlat1	small o, dieresis or umlaut mark	\n\
247	HTML-40	div	:alias:	ISOnum divide	\n\
247	HTML-40	divide	ISOnum	divide sign	\n\
248	HTML-40	oslash	ISOlat1	small o with stroke	\n\
249	HTML-40	ugrave	ISOlat1	small u, grave accent	\n\
250	HTML-40	uacute	ISOlat1	small u, acute accent	\n\
251	HTML-40	ucirc	ISOlat1	small u, circumflex accent	\n\
252	HTML-40	uuml	ISOlat1	small u, dieresis or umlaut mark	\n\
253	HTML-40	yacute	ISOlat1	small y, acute accent	\n\
254	HTML-40	thorn	ISOlat1	small thorn, Icelandic	\n\
255	HTML-40	yuml	ISOlat1	small y, dieresis or umlaut mark	\n\
	\n\
#	01	\n\
	\n\
256	-	Amacr	ISOlat2	capital A, macron	\n\
257	-	amacr	ISOlat2	small a, macron	\n\
258	-	Abreve	ISOlat2	capital A, breve	\n\
259	-	abreve	ISOlat2	small a, breve	\n\
260	-	Aogon	ISOlat2	capital A, ogonek	\n\
261	-	aogon	ISOlat2	small a, ogonek	\n\
262	-	Cacute	ISOlat2	capital C, acute accent	\n\
263	-	cacute	ISOlat2	small c, acute accent	\n\
264	-	Ccirc	ISOlat2	capital C, circumflex accent	\n\
265	-	ccirc	ISOlat2	small c, circumflex accent	\n\
266	-	Cdot	ISOlat2	capital C, dot above	\n\
267	-	cdot	ISOlat2	small c, dot above	\n\
268	-	Ccaron	ISOlat2	capital C, caron	\n\
269	-	ccaron	ISOlat2	small c, caron	\n\
270	-	Dcaron	ISOlat2	capital D, caron	\n\
271	-	dcaron	ISOlat2	small d, caron	\n\
272	-	Dstrok	ISOlat2	capital D, stroke	\n\
273	-	dstrok	ISOlat2	small d, stroke	\n\
274	-	Emacr	ISOlat2	capital E, macron	\n\
275	-	emacr	ISOlat2	small e, macron	\n\
278	-	Edot	ISOlat2	capital E, dot above	\n\
279	-	edot	ISOlat2	small e, dot above	\n\
280	-	Eogon	ISOlat2	capital E, ogonek	\n\
281	-	eogon	ISOlat2	small e, ogonek	\n\
282	-	Ecaron	ISOlat2	capital E, caron	\n\
283	-	ecaron	ISOlat2	small e, caron	\n\
284	-	Gcirc	ISOlat2	capital G, circumflex accent	\n\
285	-	gcirc	ISOlat2	small g, circumflex accent	\n\
286	-	Gbreve	ISOlat2	capital G, breve	\n\
287	-	gbreve	ISOlat2	small g, breve	\n\
288	-	Gdot	ISOlat2	capital G, dot above	\n\
289	-	gdot	ISOlat2	small g, dot above	\n\
290	-	Gcedil	ISOlat2	capital G, cedilla	\n\
292	-	Hcirc	ISOlat2	capital H, circumflex accent	\n\
293	-	hcirc	ISOlat2	small h, circumflex accent	\n\
294	-	Hstrok	ISOlat2	capital H, stroke	\n\
295	-	hstrok	ISOlat2	small h, stroke	\n\
296	-	Itilde	ISOlat2	capital I, tilde	\n\
297	-	itilde	ISOlat2	small i, tilde	\n\
298	-	Imacr	ISOlat2	capital I, macron	\n\
299	-	imacr	ISOlat2	small i, macron	\n\
302	-	Iogon	ISOlat2	capital I, ogonek	\n\
303	-	iogon	ISOlat2	small i, ogonek	\n\
304	-	Idot	ISOlat2	capital I, dot above	\n\
305	-	imath	ISOamso	small i, no dot	\n\
305	-	inodot	ISOlat2	small i without dot	\n\
306	-	IJlig	ISOlat2	capital IJ ligature	\n\
307	-	ijlig	ISOlat2	small ij ligature	\n\
308	-	Jcirc	ISOlat2	capital J, circumflex accent	\n\
309	-	jcirc	ISOlat2	small j, circumflex accent	\n\
310	-	Kcedil	ISOlat2	capital K, cedilla	\n\
311	-	kcedil	ISOlat2	small k, cedilla	\n\
312	-	kgreen	ISOlat2	small k, Greenlandic	\n\
313	-	Lacute	ISOlat2	capital L, acute accent	\n\
314	-	lacute	ISOlat2	small l, acute accent	\n\
315	-	Lcedil	ISOlat2	capital L, cedilla	\n\
316	-	lcedil	ISOlat2	small l, cedilla	\n\
317	-	Lcaron	ISOlat2	capital L, caron	\n\
318	-	lcaron	ISOlat2	small l, caron	\n\
319	-	Lmidot	ISOlat2	capital L, middle dot	\n\
320	-	lmidot	ISOlat2	small l, middle dot	\n\
321	-	Lstrok	ISOlat2	capital L, stroke	\n\
322	-	lstrok	ISOlat2	small l, stroke	\n\
323	-	Nacute	ISOlat2	capital N, acute accent	\n\
324	-	nacute	ISOlat2	small n, acute accent	\n\
325	-	Ncedil	ISOlat2	capital N, cedilla	\n\
326	-	ncedil	ISOlat2	small n, cedilla	\n\
327	-	Ncaron	ISOlat2	capital N, caron	\n\
328	-	ncaron	ISOlat2	small n, caron	\n\
329	-	napos	ISOlat2	small n, apostrophe	\n\
330	-	ENG	ISOlat2	capital ENG, Lapp	\n\
331	-	eng	ISOlat2	small eng, Lapp	\n\
332	-	Omacr	ISOlat2	capital O, macron	\n\
333	-	omacr	ISOlat2	small o, macron	\n\
336	-	Odblac	ISOlat2	capital O, double acute accent	\n\
337	-	odblac	ISOlat2	small o, double acute accent	\n\
338	HTML-40	OElig	ISOlat2	capital OE ligature	\n\
339	HTML-40	oelig	ISOlat2	small oe ligature	\n\
340	-	Racute	ISOlat2	capital R, acute accent	\n\
341	-	racute	ISOlat2	small r, acute accent	\n\
342	-	Rcedil	ISOlat2	capital R, cedilla	\n\
343	-	rcedil	ISOlat2	small r, cedilla	\n\
344	-	Rcaron	ISOlat2	capital R, caron	\n\
345	-	rcaron	ISOlat2	small r, caron	\n\
346	-	Sacute	ISOlat2	capital S, acute accent	\n\
347	-	sacute	ISOlat2	small s, acute accent	\n\
348	-	Scirc	ISOlat2	capital S, circumflex accent	\n\
349	-	scirc	ISOlat2	small s, circumflex accent	\n\
350	-	Scedil	ISOlat2	capital S, cedilla	\n\
351	-	scedil	ISOlat2	small s, cedilla	\n\
352	HTML-40	Scaron	ISOlat2	capital S, caron	\n\
353	HTML-40	scaron	ISOlat2	small s, caron	\n\
354	-	Tcedil	ISOlat2	capital T, cedilla	\n\
355	-	tcedil	ISOlat2	small t, cedilla	\n\
356	-	Tcaron	ISOlat2	capital T, caron	\n\
357	-	tcaron	ISOlat2	small t, caron	\n\
358	-	Tstrok	ISOlat2	capital T, stroke	\n\
359	-	tstrok	ISOlat2	small t, stroke	\n\
360	-	Utilde	ISOlat2	capital U, tilde	\n\
361	-	utilde	ISOlat2	small u, tilde	\n\
362	-	Umacr	ISOlat2	capital U, macron	\n\
363	-	umacr	ISOlat2	small u, macron	\n\
364	-	Ubreve	ISOlat2	capital U, breve	\n\
365	-	ubreve	ISOlat2	small u, breve	\n\
366	-	Uring	ISOlat2	capital U, ring	\n\
367	-	uring	ISOlat2	small u, ring	\n\
368	-	Udblac	ISOlat2	capital U, double acute accent	\n\
369	-	udblac	ISOlat2	small u, double acute accent	\n\
370	-	Uogon	ISOlat2	capital U, ogonek	\n\
371	-	uogon	ISOlat2	small u, ogonek	\n\
372	-	Wcirc	ISOlat2	capital W, circumflex accent	\n\
373	-	wcirc	ISOlat2	small w, circumflex accent	\n\
374	-	Ycirc	ISOlat2	capital Y, circumflex accent	\n\
375	-	ycirc	ISOlat2	small y, circumflex accent	\n\
376	HTML-40	Yuml	ISOlat2	capital Y, dieresis or umlaut mark	\n\
377	-	Zacute	ISOlat2	capital Z, acute accent	\n\
378	-	zacute	ISOlat2	small z, acute accent	\n\
379	-	Zdot	ISOlat2	capital Z, dot above	\n\
380	-	zdot	ISOlat2	small z, dot above	\n\
381	-	Zcaron	ISOlat2	capital Z, caron	\n\
382	-	zcaron	ISOlat2	small z, caron	\n\
402	HTML-40	fnof	ISOtech	function of (italic small f), florian	\n\
501	-	gacute	ISOlat2	small g, acute accent	\n\
	\n\
#	02	\n\
	\n\
603	-	varepsilon	:alias:	ISOgrk3 epsiv	\n\
603	-	epsiv	ISOgrk3	varepsilon	\n\
710	HTML-40	circ	ISOdia,ISOpub	modifier letter circumflex accent	\n\
711	-	Hacek	:alias:	ISOdia caron	\n\
711	-	caron	ISOdia	caron	\n\
728	-	Breve	:alias:	ISOdia breve	\n\
728	-	breve	ISOdia	breve	\n\
729	-	DiacriticalDot	:alias:	ISOdia dot	\n\
729	-	dot	ISOdia	dot above	\n\
730	-	ring	ISOdia	ring	\n\
731	-	ogon	ISOdia	ogonek	\n\
732	-	DiacriticalTilde	:alias:	ISOdia tilde	\n\
732	HTML-40	tilde	ISOdia	tilde	\n\
733	-	DiacriticalDoubleAcute	:alias:	ISOdia dblac	\n\
733	-	dblac	ISOdia	double acute accent	\n\
	\n\
#	03	\n\
	\n\
770	-	Hat	:alias:	circumflex accent (circ in ISOdia)	\n\
785	-	DownBreve	Lextra	breve, inverted (non-spacing)	\n\
818	-	UnderBar	Lextra	combining low line	\n\
902	-	Aacgr	ISOgrk2	capital Alpha, accent, Greek	\n\
904	-	Eacgr	ISOgrk2	capital Epsilon, accent, Greek	\n\
905	-	EEacgr	ISOgrk2	capital Eta, accent, Greek	\n\
906	-	Iacgr	ISOgrk2	capital Iota, accent, Greek	\n\
908	-	Oacgr	ISOgrk2	capital Omicron, accent, Greek	\n\
910	-	Uacgr	ISOgrk2	capital Upsilon, accent, Greek	\n\
911	-	OHacgr	ISOgrk2	capital Omega, accent, Greek	\n\
912	-	idiagr	ISOgrk2	small iota, dieresis, accent, Greek	\n\
913	HTML-40	Alpha	:alias:	capital Alpha, Greek	\n\
913	-	Agr	ISOgrk1	capital Alpha, Greek	\n\
914	HTML-40	Beta	:alias:	capital Beta, Greek	\n\
914	-	Bgr	ISOgrk1	capital Beta, Greek	\n\
915	HTML-40	Gamma	ISOgrk3,:alias:	capital Gamma, Greek	\n\
915	-	Ggr	ISOgrk1	capital Gamma, Greek	\n\
916	HTML-40	Delta	ISOgrk3,:alias:	capital Delta, Greek	\n\
916	-	Dgr	ISOgrk1	capital Delta, Greek	\n\
917	HTML-40	Epsilon	:alias:	capital Epsilon, Greek	\n\
917	-	Egr	ISOgrk1	capital Epsilon, Greek	\n\
918	HTML-40	Zeta	:alias:	capital Zeta, Greek	\n\
918	-	Zgr	ISOgrk1	capital Zeta, Greek	\n\
919	HTML-40	Eta	ISOgrk1,:alias:	capital Eta, Greek	\n\
919	-	EEgr	ISOgrk1	capital Eta, Greek	\n\
920	HTML-40	Theta	ISOgrk3,:alias:	capital Theta, Greek	\n\
920	-	THgr	ISOgrk1	capital Theta, Greek	\n\
921	HTML-40	Iota	:alias:	capital Iota, Greek	\n\
921	-	Igr	ISOgrk1	capital Iota, Greek	\n\
922	HTML-40	Kappa	:alias:	capital Kappa, Greek	\n\
922	-	Kgr	ISOgrk1	capital Kappa, Greek	\n\
923	HTML-40	Lambda	ISOgrk3,:alias:	capital Lambda, Greek	\n\
923	-	Lgr	ISOgrk1	capital Lambda, Greek	\n\
924	HTML-40	Mu	:alias:	capital Mu, Greek	\n\
924	-	Mgr	ISOgrk1	capital Mu, Greek	\n\
925	HTML-40	Nu	:alias:	capital Nu, Greek	\n\
925	-	Ngr	ISOgrk1	capital Nu, Greek	\n\
926	HTML-40	Xi	ISOgrk3,:alias:	capital Xi, Greek	\n\
926	-	Xgr	ISOgrk1	capital Xi, Greek	\n\
927	HTML-40	Omicron	:alias:	capital Omicron, Greek	\n\
927	-	Ogr	ISOgrk1	capital Omicron, Greek	\n\
928	HTML-40	Pi	ISOgrk3,:alias:	capital Pi, Greek	\n\
928	-	Pgr	ISOgrk1	capital Pi, Greek	\n\
929	HTML-40	Rho	:alias:	capital Rho, Greek	\n\
929	-	Rgr	ISOgrk1	capital Rho, Greek	\n\
930	-	-	-	-	\n\
931	HTML-40	Sigma	ISOgrk3,:alias:	capital Sigma, Greek	\n\
931	-	Sgr	ISOgrk1	capital Sigma, Greek	\n\
932	HTML-40	Tau	:alias:	capital Tau, Greek	\n\
932	-	Tgr	ISOgrk1	capital Tau, Greek	\n\
933	HTML-40	Upsilon	:alias:	capital Upsilon, Greek	\n\
933	-	Ugr	ISOgrk1	capital Upsilon, Greek	\n\
934	HTML-40	Phi	ISOgrk3,:alias:	capital Phi, Greek	\n\
934	-	PHgr	ISOgrk1	capital Phi, Greek	\n\
935	HTML-40	Chi	:alias:	capital Chi, Greek	\n\
935	-	KHgr	ISOgrk1	capital Chi, Greek	\n\
936	HTML-40	Psi	ISOgrk3,:alias:	capital Psi, Greek	\n\
936	HTML-40	PSgr	ISOgrk1	capital Psi, Greek	\n\
937	HTML-40	Omega	ISOgrk3,:alias:	capital Omega, Greek	\n\
937	-	OHgr	ISOgrk1	capital Omega, Greek	\n\
938	-	Idigr	ISOgrk2	capital Iota, dieresis, Greek	\n\
939	-	Udigr	ISOgrk2	capital Upsilon, dieresis, Greek	\n\
940	-	aacgr	ISOgrk2	small alpha, accent, Greek	\n\
941	-	eacgr	ISOgrk2	small epsilon, accent, Greek	\n\
942	-	eeacgr	ISOgrk2	small eta, accent, Greek	\n\
943	-	iacgr	ISOgrk2	small iota, accent, Greek	\n\
944	-	udiagr	ISOgrk2	small upsilon, dieresis, accent, Greek	\n\
945	HTML-40	alpha	ISOgrk3,:alias:	small alpha, Greek	\n\
945	-	agr	ISOgrk1	small alpha, Greek	\n\
946	HTML-40	beta	ISOgrk3,:alias:	small beta, Greek	\n\
946	-	bgr	ISOgrk1	small beta, Greek	\n\
947	HTML-40	gamma	ISOgrk3,:alias:	small gamma, Greek	\n\
947	-	ggr	ISOgrk1	small gamma, Greek	\n\
948	HTML-40	delta	ISOgrk3,:alias:	small delta, Greek	\n\
948	-	dgr	ISOgrk1	small delta, Greek	\n\
949	-	epsi	ISOgrk3,:alias:	small epsilon, Greek	\n\
949	HTML-40	epsilon	:alias:	small epsilon, Greek	\n\
949	-	straightepsilon	:alias:	ISOgrk3 epsi	\n\
949	-	egr	ISOgrk1	small epsilon, Greek	\n\
950	-	zgr	ISOgrk1	small zeta, Greek	\n\
950	HTML-40	zeta	ISOgrk3,:alias:	small zeta, Greek	\n\
951	HTML-40	eta	ISOgrk3,:alias:	small eta, Greek	\n\
951	-	eegr	ISOgrk1	small eta, Greek	\n\
952	-	thgr	ISOgrk1	small theta, Greek	\n\
952	HTML-40	theta	ISOgrk3,:alias:	straight theta, small theta, Greek	\n\
953	HTML-40	iota	ISOgrk3,:alias:	small iota, Greek	\n\
953	-	igr	ISOgrk1	small iota, Greek	\n\
954	-	kgr	ISOgrk1	small kappa, Greek	\n\
954	HTML-40	kappa	ISOgrk3,:alias:	small kappa, Greek	\n\
955	-	lgr	ISOgrk1	small lambda, Greek	\n\
955	HTML-40	lambda	ISOgrk3,:alias:	small lambda, Greek	\n\
956	HTML-40	mu	ISOgrk3,:alias:	small mu, Greek	\n\
956	-	mgr	ISOgrk1	small mu, Greek	\n\
957	HTML-40	nu	ISOgrk3,:alias:	small nu, Greek	\n\
957	-	ngr	ISOgrk1	small nu, Greek	\n\
958	HTML-40	xi	ISOgrk3,:alias:	small xi, Greek	\n\
958	-	xgr	ISOgrk1	small xi, Greek	\n\
959	HTML-40	omicron	ISOgrk3,:alias:	small omicron, Greek	\n\
959	-	ogr	ISOgrk1	small omicron, Greek	\n\
960	HTML-40	pi	ISOgrk3,:alias:	small pi, Greek	\n\
960	-	pgr	ISOgrk1	small pi, Greek	\n\
961	HTML-40	rho	ISOgrk3,:alias:	small rho, Greek	\n\
961	-	rgr	ISOgrk1	small rho, Greek	\n\
962	-	sigmav	ISOgrk3,:alias:	varsigma	\n\
962	-	varsigma	:alias:	ISOgrk3 sigmav	\n\
962	HTML-40	sigmaf	:alias:	final sigma	\n\
962	-	sfgr	ISOgrk1	final small sigma, Greek	\n\
963	HTML-40	sigma	ISOgrk3,:alias:	small sigma, Greek	\n\
963	-	sgr	ISOgrk1	small sigma, Greek	\n\
964	HTML-40	tau	ISOgrk3,:alias:	small tau, Greek	\n\
964	-	tgr	ISOgrk1	small tau, Greek	\n\
965	-	upsi	ISOgrk3,:alias:	small upsilon, Greek	\n\
965	-	ugr	ISOgrk1	small upsilon, Greek	\n\
966	-	phgr	ISOgrk1	small phi, Greek	\n\
965	HTML-40	upsilon	:alias:	ISOgrk3 upsi	\n\
966	-	straightphi	:alias:	ISOgrk3 phi	\n\
966	HTML-40	phi	ISOgrk3,:alias:	small phi, Greek	\n\
967	-	khgr	ISOgrk1	small chi, Greek	\n\
967	HTML-40	chi	ISOgrk3,:alias:	small chi, Greek	\n\
968	HTML-40	psi	ISOgrk3,:alias:	small psi, Greek	\n\
968	-	psgr	ISOgrk1	small psi, Greek	\n\
969	HTML-40	omega	ISOgrk3,:alias:	small omega, Greek	\n\
969	-	ohgr	ISOgrk1	small omega, Greek	\n\
970	-	idigr	ISOgrk2	small iota, dieresis, Greek	\n\
971	-	udigr	ISOgrk2	small upsilon, dieresis, Greek	\n\
972	-	oacgr	ISOgrk2	small omicron, accent, Greek	\n\
973	-	uacgr	ISOgrk2	small upsilon, accent, Greek	\n\
974	-	ohacgr	ISOgrk2	small omega, accent, Greek	\n\
977	-	vartheta	:alias:	ISOgrk3 thetav	\n\
977	HTML-40	thetasym	:alias:	small theta symbol	\n\
977	-	thetav	ISOgrk3	curly or open theta	\n\
978	-	Upsi	ISOgrk3,:alias:	capital Upsilon, Greek	\n\
978	-	Upsilon	:alias:	ISOgrk3 Upsi	\n\
979	-	-	-	-	\n\
980	-	-	-	-	\n\
981	-	phiv	ISOgrk3,:alias:	curly or open phi	\n\
981	-	varphi	:alias:	ISOgrk3 phiv	\n\
982	-	varpi	:alias:	ISOgrk3 piv	\n\
982	HTML-40	pi	:alias:	small pi, Greek	\n\
982	-	piv	ISOgrk3	varpi	\n\
988	-	digamma	:alias:	ISOgrk3 gammad	\n\
988	-	gammad	ISOgrk3,:alias:	digamma	\n\
988	-	Gammad	ISOgrk3	capital digamma	\n\
1008	-	varkappa	:alias:	ISOgrk3 kappav	\n\
1008	-	kappav	ISOgrk3	varkappa	\n\
1009	-	varrho	:alias:	ISOgrk3 rhov	\n\
1009	-	rhov	ISOgrk3	varrho	\n\
	\n\
#	04	\n\
	\n\
1025	-	IOcy	ISOcyr1	capital IO, Russian	\n\
1026	-	DJcy	ISOcyr2	capital DJE, Serbian	\n\
1027	-	GJcy	ISOcyr2	capital GJE Macedonian	\n\
1028	-	Jukcy	ISOcyr2	capital JE, Ukrainian	\n\
1029	-	DScy	ISOcyr2	capital DSE, Macedonian	\n\
1030	-	Iukcy	ISOcyr2	capital I, Ukrainian	\n\
1031	-	YIcy	ISOcyr2	capital YI, Ukrainian	\n\
1032	-	Jsercy	ISOcyr2	capital JE, Serbian	\n\
1033	-	LJcy	ISOcyr2	capital LJE, Serbian	\n\
1034	-	NJcy	ISOcyr2	capital NJE, Serbian	\n\
1035	-	TSHcy	ISOcyr2	capital TSHE, Serbian	\n\
1036	-	KJcy	ISOcyr2	capital KJE, Macedonian	\n\
1038	-	Ubrcy	ISOcyr2	capital U, Byelorussian	\n\
1039	-	DZcy	ISOcyr2	capital dze, Serbian	\n\
1040	-	Acy	ISOcyr1	capital A, Cyrillic	\n\
1041	-	Bcy	ISOcyr1	capital BE, Cyrillic	\n\
1042	-	Vcy	ISOcyr1	capital VE, Cyrillic	\n\
1043	-	Gcy	ISOcyr1	capital GHE, Cyrillic	\n\
1044	-	Dcy	ISOcyr1	capital DE, Cyrillic	\n\
1045	-	IEcy	ISOcyr1	capital IE, Cyrillic	\n\
1046	-	ZHcy	ISOcyr1	capital ZHE, Cyrillic	\n\
1047	-	Zcy	ISOcyr1	capital ZE, Cyrillic	\n\
1048	-	Icy	ISOcyr1	capital I, Cyrillic	\n\
1049	-	Jcy	ISOcyr1	capital short I, Cyrillic	\n\
1050	-	Kcy	ISOcyr1	capital KA, Cyrillic	\n\
1051	-	Lcy	ISOcyr1	capital EL, Cyrillic	\n\
1052	-	Mcy	ISOcyr1	capital EM, Cyrillic	\n\
1053	-	Ncy	ISOcyr1	capital EN, Cyrillic	\n\
1054	-	Ocy	ISOcyr1	capital O, Cyrillic	\n\
1055	-	Pcy	ISOcyr1	capital PE, Cyrillic	\n\
1056	-	Rcy	ISOcyr1	capital ER, Cyrillic	\n\
1057	-	Scy	ISOcyr1	capital ES, Cyrillic	\n\
1058	-	Tcy	ISOcyr1	capital TE, Cyrillic	\n\
1059	-	Ucy	ISOcyr1	capital U, Cyrillic	\n\
1060	-	Fcy	ISOcyr1	capital EF, Cyrillic	\n\
1061	-	KHcy	ISOcyr1	capital HA, Cyrillic	\n\
1062	-	TScy	ISOcyr1	capital TSE, Cyrillic	\n\
1063	-	CHcy	ISOcyr1	capital CHE, Cyrillic	\n\
1064	-	SHcy	ISOcyr1	capital SHA, Cyrillic	\n\
1065	-	SHCHcy	ISOcyr1	capital SHCHA, Cyrillic	\n\
1066	-	HARDcy	ISOcyr1	capital HARD sign, Cyrillic	\n\
1067	-	Ycy	ISOcyr1	capital YERU, Cyrillic	\n\
1068	-	SOFTcy	ISOcyr1	capital SOFT sign, Cyrillic	\n\
1069	-	Ecy	ISOcyr1	capital E, Cyrillic	\n\
1070	-	YUcy	ISOcyr1	capital YU, Cyrillic	\n\
1071	-	YAcy	ISOcyr1	capital YA, Cyrillic	\n\
1072	-	acy	ISOcyr1	small a, Cyrillic	\n\
1073	-	bcy	ISOcyr1	small be, Cyrillic	\n\
1074	-	vcy	ISOcyr1	small ve, Cyrillic	\n\
1075	-	gcy	ISOcyr1	small ghe, Cyrillic	\n\
1076	-	dcy	ISOcyr1	small de, Cyrillic	\n\
1077	-	iecy	ISOcyr1	small ie, Cyrillic	\n\
1078	-	zhcy	ISOcyr1	small zhe, Cyrillic	\n\
1079	-	zcy	ISOcyr1	small ze, Cyrillic	\n\
1080	-	icy	ISOcyr1	small i, Cyrillic	\n\
1081	-	jcy	ISOcyr1	small short i, Cyrillic	\n\
1082	-	kcy	ISOcyr1	small ka, Cyrillic	\n\
1083	-	lcy	ISOcyr1	small el, Cyrillic	\n\
1084	-	mcy	ISOcyr1	small em, Cyrillic	\n\
1085	-	ncy	ISOcyr1	small en, Cyrillic	\n\
1086	-	ocy	ISOcyr1	small o, Cyrillic	\n\
1087	-	pcy	ISOcyr1	small pe, Cyrillic	\n\
1088	-	rcy	ISOcyr1	small er, Cyrillic	\n\
1089	-	scy	ISOcyr1	small es, Cyrillic	\n\
1090	-	tcy	ISOcyr1	small te, Cyrillic	\n\
1091	-	ucy	ISOcyr1	small u, Cyrillic	\n\
1092	-	fcy	ISOcyr1	small ef, Cyrillic	\n\
1093	-	khcy	ISOcyr1	small ha, Cyrillic	\n\
1094	-	tscy	ISOcyr1	small tse, Cyrillic	\n\
1095	-	chcy	ISOcyr1	small che, Cyrillic	\n\
1096	-	shcy	ISOcyr1	small sha, Cyrillic	\n\
1097	-	shchcy	ISOcyr1	small shcha, Cyrillic	\n\
1098	-	hardcy	ISOcyr1	small hard sign, Cyrillic	\n\
1099	-	ycy	ISOcyr1	small yeru, Cyrillic	\n\
1100	-	softcy	ISOcyr1	small soft sign, Cyrillic	\n\
1101	-	ecy	ISOcyr1	small e, Cyrillic	\n\
1102	-	yucy	ISOcyr1	small yu, Cyrillic	\n\
1103	-	yacy	ISOcyr1	small ya, Cyrillic	\n\
1105	-	iocy	ISOcyr1	small io, Russian	\n\
1106	-	djcy	ISOcyr2	small dje, Serbian	\n\
1107	-	gjcy	ISOcyr2	small gje, Macedonian	\n\
1108	-	jukcy	ISOcyr2	small je, Ukrainian	\n\
1109	-	dscy	ISOcyr2	small dse, Macedonian	\n\
1110	-	iukcy	ISOcyr2	small i, Ukrainian	\n\
1111	-	yicy	ISOcyr2	small yi, Ukrainian	\n\
1112	-	jsercy	ISOcyr2	small je, Serbian	\n\
1113	-	ljcy	ISOcyr2	small lje, Serbian	\n\
1114	-	njcy	ISOcyr2	small nje, Serbian	\n\
1115	-	tshcy	ISOcyr2	small tshe, Serbian	\n\
1116	-	kjcy	ISOcyr2	small kje Macedonian	\n\
1118	-	ubrcy	ISOcyr2	small u, Byelorussian	\n\
1119	-	dzcy	ISOcyr2	small dze, Serbian	\n\
	\n\
5760	-	-	-	ogham space mark	\n\
6158	-	-	-	mongolian vowel separator	\n\
	\n\
#	20	\n\
	\n\
8192	-	-	-	en space (1/2-em)	\n\
8193	-	-	-	em space	\n\
8194	HTML-40	ensp	ISOpub	en space (1/2-em)	\n\
8195	HTML-40	emsp	ISOpub	em space	\n\
8196	-	emsp13	ISOpub	1/3-em space	\n\
8197	-	emsp14	ISOpub	1/4-em space	\n\
8198	-	emsp16	ISOpub	1/6-em space	\n\
8199	-	numsp	ISOpub	digit space (width of a number)	\n\
8199	-	-	:alias:	figure space	\n\
8199	-	-	:alias:	tabular width space	\n\
8200	-	puncsp	ISOpub	punctuation space (width of comma))	\n\
8201	-	ThinSpace	:alias:	space of width 3/18 em (ISOpub thinsp	\n\
8201	HTML-40	thinsp	ISOpub	thin space (1/6-em)	\n\
8202	-	VeryThinSpace	:alias:	space of width 1/18 em (ISOpub hairsp)	\n\
8202	-	hairsp	ISOpub	hair space	\n\
8203	-	ZeroWidthSpace	Lextra	zero width space	\n\
8204	HTML-40	zwnj	ISO8879	zero width non-joiner	\n\
8205	HTML-40	zwj	ISO8879	zero width joiner	\n\
8206	HTML-40	lrm	ISO8879	left-to-right mark	\n\
8207	HTML-40	rlm	ISO8879	right-to-left mark	\n\
8208	-	dash	ISOpub,:alias:	hyphen (true graphic)	\n\
8208	-	hyphen	ISOnum	hyphen	\n\
8209	-	-	-	non breaking hyphen	\n\
8211	HTML-40	ndash	ISOpub	en dash	\n\
8212	HTML-40	mdash	ISOpub	em dash 	\n\
8213	-	horbar	ISOnum	horizontal bar	\n\
8214	-	Vert	:alias:	ISOtech Verbar	\n\
8214	-	Verbar	ISOtech	double vertical bar	\n\
8216	-	OpenCurlyQuote	:alias:	ISOnum lsquo	\n\
8216	HTML-40	lsquo	ISOnum	single quotation mark, left	\n\
8217	-	CloseCurlyQuote	:alias:	ISOnum rsquo	\n\
8217	-	rsquor	ISOpub,:alias:	rising single quote, right (high)	\n\
8217	HTML-40	rsquo	ISOnum	single quotation mark, right	\n\
8218	HTML-40	lsquor	ISOpub	rising single quote, left (low)	\n\
8220	-	OpenCurlyDoubleQuote	:alias:	ISOnum ldquo	\n\
8220	HTML-40	ldquo	ISOnum	double quotation mark, left	\n\
8221	-	CloseCurlyDoubleQuote	:alias:	ISOnum rdquo	\n\
8221	-	rdquor	ISOpub,:alias:	rising double quote, right (high)	\n\
8221	HTML-40	rdquo	ISOnum	double quotation mark, right	\n\
8222	HTML-40	bdquo	ISO8879,:alias:	double low-9 quotation mark	\n\
8222	-	ldquor	ISOpub	rising double quote, left (low)	\n\
8223	-	-	ISOpub	rising double quote, left (high)	\n\
8224	-	dagger	ISOamsb,:alias:	dagger relation	\n\
8224	HTML-40	dagger	ISOpub	dagger	\n\
8225	-	Dagger	ISOamsb,:alias:	double dagger relation	\n\
8225	-	ddagger	:alias:	ISOpub Dagger	\n\
8225	HTML-40	Dagger	ISOpub	double dagger	\n\
8226	-	bullet	:alias:	ISOpub bull	\n\
8226	HTML-40	bull	ISOpub	round bullet, filled	\n\
8229	-	nldr	ISOpub	double baseline dot (en leader)	\n\
8230	HTML-40	hellip	ISOpub	ellipsis (horizontal)	\n\
8230	-	mldr	ISOpub	em leader	\n\
8239	-	-	-	narrower than NO-BREAK SPACE	\n\
8240	HTML-40	permil	ISOtech	per thousand	\n\
8241	-	pertenk	ISOtech	per 10 thousand	\n\
8242	HTML-40	prime	ISOtech	prime prime or minute	\n\
8243	HTML-40	Prime	ISOtech	double prime or second	\n\
8244	-	tprime	ISOtech	triple prime	\n\
8245	-	backprime	:alias:	ISOamso bprime	\n\
8245	-	bprime	ISOamso	backprime - reverse prime	\n\
8249	HTML-40	lsaquo	ISOprop	single left-pointing angle quotation mark	\n\
8250	HTML-40	ssaquo	ISOprop	single right-pointing angle quotation mark	\n\
8254	HTML-40	oline	ISO8879	overline; spacing overscore	\n\
8257	-	caret	ISOpub	caret (insertion mark)	\n\
8259	-	hybull	ISOpub	rectangle, filled (hyphen bullet)	\n\
8260	HTML-40	frasl	ISO8879	fraction slashfraction slash	\n\
8287	-	-	-	4/18-em space (medium mathematical space)	\n\
8288	-	-	-	word joiner	\n\
8364	HTML-40	euro	ISO8879	euro sign	\n\
8411	-	TripleDot	:alias:	ISOtech tdot	\n\
8411	-	tdot	ISOtech	three dots above	\n\
8412	-	DotDot	ISOtech	four dots above	\n\
	\n\
#	21	\n\
	\n\
8450	-	Copf	ISOmopf	Bbb C, open face C	\n\
8451	-	-	-	degrees Celsius	\n\
8453	-	incare	ISOpub	in-care-of symbol	\n\
8457	-	-	-	degrees Fahrenheit	\n\
8459	-	Hscr	ISOmscr	script letter H	\n\
8459	-	hamilt	ISOtech,:alias:	Hamiltonian (ISOmscr Hscr)	\n\
8460	-	Hfr	ISOmfrk	/frak H, upper case h	\n\
8463	-	hslash	:alias:	ISOamso plankv	\n\
8463	-	plankv	ISOamso	variant Planck's over 2pi	\n\
8464	-	Iscr	ISOmscr	script letter I	\n\
8465	-	Ifr	ISOmfrk,:alias:	/frak I, upper case h, (ISOmscr Iscr)	\n\
8465	-	Im	:alias:	ISOamso image	\n\
8465	HTML-40	image	ISOamso	imaginary	\n\
8466	-	lagran	ISOtech,:alias:	Lagrangian (ISOmscr Lscr) 	\n\
8466	-	Lscr	ISOmscr	script letter L	\n\
8467	-	ell	ISOamso	cursive small l	\n\
8469	-	Nopf	ISOmopf	Bbb N, open face N	\n\
8470	-	numero	ISOcyr1	numero sign	\n\
8471	-	copysr	ISOpub	sound recording copyright sign	\n\
8472	-	wp	:alias:	ISOamso weierp	\n\
8472	HTML-40	weierp	ISOamso	Weierstrass p	\n\
8473	-	Popf	ISOmopf	Bbb P, open face P	\n\
8474	-	Qopf	ISOmopf	Bbb Q, open face Q	\n\
8476	-	Re	:alias:	ISOamso real	\n\
8476	-	Rfr	ISOmfrk,:alias:	/frak R, upper case r	\n\
8476	HTML-40	real	ISOamso	real	\n\
8477	-	Ropf	ISOmopf	Bbb R, open face R	\n\
8478	-	rx	ISOpub	pharmaceutical prescription (Rx)	\n\
8482	HTML-40	trade	ISOnum	trade mark sign	\n\
8484	-	Zopf	ISOmopf	Bbb Z, open face Z	\n\
8486	-	ohm	ISOnum	ohm sign	\n\
8487	-	mho	ISOamso	Siemens; conductance	\n\
8489	-	iiota	ISOamso	inverted iota	\n\
8490	-	-	-	Kelvin	\n\
8491	-	angst	ISOtech	Angstrom capital A, ring	\n\
8492	-	bernou	ISOtech,:alias:	Bernoulli function (ISOmscr Bscr) 	\n\
8492	-	Bscr	ISOmscr	script letter B	\n\
8495	-	escr	ISOmscr	script letter e	\n\
8496	-	Escr	ISOmscr	script letter E	\n\
8497	-	Fscr	ISOmscr	script letter F	\n\
8499	-	phmmat	ISOtech,:alias:	physics M-matrix (ISOmscr Mscr) 	\n\
8499	-	Mscr	ISOmscr	script letter M	\n\
8500	-	order	ISOtech,:alias:	order of (ISOmscr oscr) 	\n\
8500	-	oscr	ISOmscr	script letter o	\n\
8501	HTML-40	aleph	ISOtech	aleph, Hebrew	\n\
8502	-	beth	ISOamso	beth, Hebrew	\n\
8503	-	gimel	ISOamso	gimel, Hebrew	\n\
8504	-	daleth	ISOamso	daleth, Hebrew	\n\
8531	-	frac13	ISOpub	fraction one-third	\n\
8532	-	frac23	ISOpub	fraction two-thirds	\n\
8533	-	frac15	ISOpub	fraction one-fifth	\n\
8534	-	frac25	ISOpub	fraction two-fifths	\n\
8535	-	frac35	ISOpub	fraction three-fifths	\n\
8536	-	frac45	ISOpub	fraction four-fifths	\n\
8537	-	frac16	ISOpub	fraction one-sixth	\n\
8538	-	frac56	ISOpub	fraction five-sixths	\n\
8539	-	frac18	ISOnum	fraction one-eighth	\n\
8540	-	frac38	ISOnum	fraction three-eighths	\n\
8541	-	frac58	ISOnum	fraction five-eighths	\n\
8542	-	frac78	ISOnum	fraction seven-eighths	\n\
8592	-	LeftArrow	:alias:	ISOnum larr	\n\
8592	-	leftarrow	:alias:	ISOnum larr	\n\
8592	HTML-40	larr	ISOnum	leftward arrow	\n\
8593	-	UpArrow	:alias:	ISOnum uarr	\n\
8593	-	uparrow	:alias:	ISOnum uarr	\n\
8593	HTML-40	uarr	ISOnum	upward arrow	\n\
8594	-	RightArrow	:alias:	ISOnum rarr	\n\
8594	-	rightarrow	:alias:	ISOnum rarr	\n\
8594	HTML-40	rarr	ISOnum	rightward arrow	\n\
8595	-	DownArrow	:alias:	ISOnum darr	\n\
8595	-	downarrow	:alias:	ISOnum darr	\n\
8595	HTML-40	darr	ISOnum	downward arrow	\n\
8596	-	LeftRightArrow	:alias:	ISOamsa harr	\n\
8596	-	leftrightarrow	:alias:	ISOamsa harr	\n\
8596	HTML-40	harr	ISOamsa	left and right arrow	\n\
8597	-	UpDownArrow	:alias:	ISOamsa varr	\n\
8597	-	updownarrow	:alias:	ISOamsa varr	\n\
8597	-	varr	ISOamsa	up and down arrow	\n\
8598	-	UpperLeftArrow	:alias:	ISOamsa nwarr	\n\
8598	-	nwarrow	:alias:	ISOamsa nwarr	\n\
8598	-	nwarr	ISOamsa	NW pointing arrow	\n\
8599	-	UpperRightArrow	:alias:	ISOamsa nearr	\n\
8599	-	nearrow	:alias:	ISOamsa nearr	\n\
8599	-	nearr	ISOamsa	NE pointing arrow	\n\
8600	-	LowerRightArrow	:alias:	ISOamsa searr	\n\
8600	-	searrow	:alias:	ISOamsa searr	\n\
8600	-	searr	ISOamsa	SE pointing arrow	\n\
8601	-	LowerLeftArrow	:alias:	ISOamsa swarr	\n\
8601	-	swarrow	:alias:	ISOamsa swarr	\n\
8601	-	swarr	ISOamsa	SW pointing arrow	\n\
8602	-	nleftarrow	:alias:	ISOamsa nlarr	\n\
8602	-	nlarr	ISOamsa	not left arrow	\n\
8603	-	nrightarrow	:alias:	ISOamsa nrarr	\n\
8603	-	nrarr	ISOamsa	not right arrow	\n\
8606	-	twoheadleftarrow	:alias:	ISOamsa Larr	\n\
8606	-	Larr	ISOamsa	two head leftarrow	\n\
8607	-	Uarr	ISOamsa	up two-headed arrow	\n\
8608	-	twoheadrightarrow	:alias:	ISOamsa Rarr	\n\
8608	-	Rarr	ISOamsa	two head right arrow	\n\
8609	-	Darr	ISOamsa	down two-headed arrow	\n\
8610	-	leftarrowtail	:alias:	ISOamsa larrtl	\n\
8610	-	larrtl	ISOamsa	left arrow-tailed	\n\
8611	-	ratail	ISOamsa	right arrow-tail	\n\
8611	-	rightarrowtail	:alias:	ISOamsa rarrtl	\n\
8611	-	rarrtl	ISOamsa	rioght arrow-tailed	\n\
8612	-	LeftTeeArrow	Lextra,:alias:	for mapstoleft	\n\
8612	-	mapstoleft	Lextra	leftwards arrow from bar	\n\
8613	-	UpTeeArrow	Lextra,:alias:	mapstoup	\n\
8613	-	mapstoup	Lextra	upwards arrow from bar	\n\
8614	-	RightTeeArrow	:alias:	ISOamsa map	\n\
8614	-	mapsto	:alias:	ISOamsa map	\n\
8614	-	map	ISOamsa	maps to	\n\
8615	-	DownTeeArrow	Lextra,:alias:	mmlextra mapstodown	\n\
8615	-	mapstodown	Lextra	downwards arrow from bar	\n\
8617	-	hookleftarrow	:alias:	ISOamsa larrhk	\n\
8617	-	larrhk	ISOamsa	left arrow-hooked	\n\
8618	-	hookrightarrow	:alias:	ISOamsa rarrhk	\n\
8618	-	rarrhk	ISOamsa	right arrow-hooked	\n\
8619	-	looparrowleft	:alias:	ISOamsa larrlp	\n\
8619	-	larrlp	ISOamsa	left arrow-looped	\n\
8620	-	looparrowright	:alias:	ISOamsa rarrlp	\n\
8620	-	rarrlp	ISOamsa	right arrow-looped	\n\
8621	-	leftrightsquigarrow	:alias:	ISOamsa harrw	\n\
8621	-	harrw	ISOamsa	left and right arr-wavy	\n\
8622	-	nleftrightarrow	:alias:	ISOamsa nharr	\n\
8622	-	nharr	ISOamsa	not left and right arrow	\n\
8624	-	Lsh	:alias:	ISOamsa lsh	\n\
8624	-	lsh	ISOamsa	Lsh	\n\
8625	-	Rsh	:alias:	ISOamsa rsh	\n\
8625	-	rsh	ISOamsa	Rsh	\n\
8626	-	ldsh	ISOamsa	left down angled arrow	\n\
8627	-	rdsh	ISOamsa	right down angled arrow	\n\
8629	HTML-40	crarr	ISO8879	downwards arrow with corner leftwards (carriage return)	\n\
8630	-	cularr	ISOamsa	left curved arrow	\n\
8630	-	curvearrowleft	:alias:	ISOamsa cularr	\n\
8631	-	curvearrowright	:alias:	ISOamsa curarr	\n\
8631	-	curarr	ISOamsa	right curved arrow	\n\
8634	-	circlearrowleft	:alias:	ISOamsa olarr	\n\
8634	-	olarr	ISOamsa	left arrow in circle	\n\
8635	-	circlearrowright	:alias:	ISOamsa orarr	\n\
8635	-	orarr	ISOamsa	right arrow in circle	\n\
8636	-	LeftVector	:alias:	ISOamsa lharu	\n\
8636	-	leftharpoonup	:alias:	ISOamsa lharu	\n\
8636	-	lharu	ISOamsa	left harpoon-up	\n\
8637	-	DownLeftVector	:alias:	ISOamsa lhard	\n\
8637	-	leftharpoondown	:alias:	ISOamsa lhard	\n\
8637	-	lhard	ISOamsa	left harpoon-down	\n\
8638	-	RightUpVector	:alias:	ISOamsa uharr	\n\
8638	-	upharpoonright	:alias:	ISOamsa uharr	\n\
8638	-	uharr	ISOamsa	up harpoon-right	\n\
8639	-	LeftUpVector	:alias:	ISOamsa uharl	\n\
8639	-	upharpoonleft	:alias:	ISOamsa uharl	\n\
8639	-	uharl	ISOamsa	up harpoon-left	\n\
8640	-	RightVector	:alias:	ISOamsa rharu	\n\
8640	-	rightharpoonup	:alias:	ISOamsa rharu	\n\
8640	-	rharu	ISOamsa	right harpoon-up	\n\
8641	-	DownRightVector	:alias:	ISOamsa rhard	\n\
8641	-	rightharpoondown	:alias:	ISOamsa rhard	\n\
8641	-	rhard	ISOamsa	right harpoon-down	\n\
8642	-	RightDownVector	:alias:	ISOamsa dharr	\n\
8642	-	downharpoonright	:alias:	ISOamsa dharr	\n\
8642	-	dharr	ISOamsa	down harpoon-right	\n\
8643	-	LeftDownVector	:alias:	ISOamsa dharl	\n\
8643	-	downharpoonleft	:alias:	ISOamsa dharl	\n\
8643	-	dharl	ISOamsa	down harpoon-left	\n\
8644	-	RightArrowLeftArrow	:alias:	ISOamsa rlarr	\n\
8644	-	rightleftarrows	:alias:	ISOamsa rlarr	\n\
8644	-	rlarr	ISOamsa	right arrow over left arrow	\n\
8645	-	UpArrowDownArrow	:alias:	ISOamsa udarr	\n\
8645	-	udarr	ISOamsa	up arrow, down arrow	\n\
8646	-	LeftArrowRightArrow	:alias:	ISOamsa lrarr	\n\
8646	-	leftrightarrows	:alias:	ISOamsa lrarr	\n\
8646	-	lrarr	ISOamsa	left arrow over right arrow	\n\
8647	-	leftleftarrows	:alias:	ISOamsa llarr	\n\
8647	-	llarr	ISOamsa	two left arrows	\n\
8648	-	upuparrows	:alias:	ISOamsa uuarr	\n\
8648	-	uuarr	ISOamsa	two up arrows	\n\
8649	-	rightrightarrows	:alias:	ISOamsa rrarr	\n\
8649	-	rrarr	ISOamsa	two right arrows	\n\
8650	-	downdownarrows	:alias:	ISOamsa ddarr	\n\
8650	-	ddarr	ISOamsa	two down arrows	\n\
8651	-	ReverseEquilibrium	:alias:	ISOamsa lrhar	\n\
8651	-	leftrightharpoons	:alias:	ISOamsa lrhar	\n\
8651	-	lrhar	ISOamsa	left harp over right	\n\
8652	-	Equilibrium	:alias:	ISOamsa rlhar	\n\
8652	-	rightleftharpoons	:alias:	ISOamsa rlhar	\n\
8652	-	rlhar	ISOamsa	right harp over left eft	\n\
8653	-	nLeftarrow	:alias:	ISOamsa nlArr	\n\
8653	-	nlArr	ISOamsa,:alias:	nLeftarrow; not implied by	\n\
8653	-	nvlArr	ISOamsa	not, vert, left double arrow	\n\
8654	-	nLeftrightarrow	:alias:	ISOamsa nhArr	\n\
8654	-	nhArr	ISOamsa,:alias:	nLeftrightarrow; not left and right double arrow	\n\
8654	-	nvHarr	ISOamsa	not, vert, left and right double arrow 	\n\
8655	-	nRightarrow	:alias:	ISOamsa nrArr	\n\
8655	-	nrArr	ISOamsa,:alias:	nRightarrow; not implies	\n\
8655	-	nvrArr	ISOamsa	not, vert, right double arrow	\n\
8656	-	DoubleLeftArrow	:alias:	ISOtech lArr	\n\
8656	-	Leftarrow	:alias:	ISOtech lArr	\n\
8656	HTML-40	lArr	ISOtech	Leftarrow; is implied by	\n\
8657	-	DoubleUpArrow	:alias:	ISOamsa uArr	\n\
8657	-	Uparrow	:alias:	ISOamsa uArr	\n\
8657	HTML-40	uArr	ISOamsa	up double arrow	\n\
8658	-	DoubleRightArrow	:alias:	ISOtech rArr	\n\
8658	-	Implies	:alias:	ISOtech rArr	\n\
8658	-	Rightarrow	:alias:	ISOtech rArr	\n\
8658	HTML-40	rArr	ISOtech	Rightarrow implies	\n\
8659	-	DoubleDownArrow	:alias:	ISOamsa dArr	\n\
8659	-	Downarrow	:alias:	ISOamsa dArr	\n\
8659	HTML-40	dArr	ISOamsa	down double arrow	\n\
8660	-	DoubleLeftRightArrow	:alias:	ISOamsa hArr	\n\
8660	-	Leftrightarrow	:alias:	ISOamsa hArr	\n\
8660	HTML-40	hArr	ISOamsa	left and right double arrow	\n\
8661	-	DoubleUpDownArrow	:alias:	ISOamsa vArr	\n\
8661	-	Updownarrow	:alias:	ISOamsa vArr	\n\
8661	-	vArr	ISOamsa	up and down double arrow	\n\
8662	-	nwArr	ISOamsa	NW pointing double arrow	\n\
8663	-	neArr	ISOamsa	NE pointing double arrow	\n\
8664	-	seArr	ISOamsa	SE pointing double arrow	\n\
8665	-	swArr	ISOamsa	SW pointing double arrow	\n\
8666	-	Lleftarrow	:alias:	ISOamsa lAarr	\n\
8666	-	lAarr	ISOamsa	left triple arrow	\n\
8667	-	Rrightarrow	:alias:	ISOamsa rAarr	\n\
8667	-	rAarr	ISOamsa	right triple arrow	\n\
8669	-	dzigrarr	ISOamsa,:alias:	right long zig-zag arrow	\n\
8669	-	rightsquigarrow	:alias:	ISOamsa rarrw	\n\
8669	-	rarrw	ISOamsa	rightsquigarrow; right arrow-wavy	\n\
8676	-	LeftArrowBar	Lextra,:alias:	mmlextra larrb	\n\
8676	-	larrb	Lextra	leftwards arrow to bar	\n\
8677	-	RightArrowBar	Lextra,:alias:	mmlextra rarrb	\n\
8677	-	rarrb	Lextra	leftwards arrow to bar	\n\
	\n\
#	22	\n\
	\n\
8704	-	ForAll	:alias:	ISOtech forall	\n\
8704	HTML-40	forall	ISOtech	for all	\n\
8705	-	complement	:alias:	ISOamso comp	\n\
8705	-	comp	ISOamso	complement sign	\n\
8706	-	PartialD	:alias:	ISOtech part	\n\
8706	HTML-40	part	ISOtech	partial differential	\n\
8707	-	Exists	:alias:	ISOtech exist	\n\
8707	HTML-40	exist	ISOtech	at least one exists	\n\
8708	-	NotExists	:alias:	ISOamso nexist	\n\
8708	-	nexists	:alias:	ISOamso nexist	\n\
8708	-	nexist	ISOamso	negated exists	\n\
8709	-	varnothing	:alias:	ISOamso emptyv	\n\
8709	HTML-40	empty	:alias:	empty set; null set; diameter	\n\
8709	-	emptyv	ISOamso	circle, slash	\n\
8711	-	Del	:alias:	ISOtech nabla	\n\
8711	HTML-40	nabla	ISOtech	del, Hamilton operator	\n\
8712	-	Element	:alias:	ISOtech isinv	\n\
8712	-	in	:alias:	ISOtech isin 	\n\
8712	-	isinv	ISOtech,:alias:	set membership, variant	\n\
8712	HTML-40	isin	ISOtech	in; set membership 	\n\
8713	-	NotElement	:alias:	ISOtech notin	\n\
8713	HTML-40	notin	ISOtech	notin; negated set membership	\n\
8715	-	ReverseElement	:alias:	ISOtech niv	\n\
8715	-	SuchThat	:alias:	ISOtech ni	\n\
8715	-	niv	ISOtech,:alias:	contains, variant	\n\
8715	HTML-40	ni	ISOtech	owns; contains	\n\
8716	-	NotReverseElement	:alias:	ISOtech notniva	\n\
8716	-	notniva	ISOtech,:alias:	negated contains, variant	\n\
8716	-	notni	ISOtech	negated contains	\n\
8719	-	Product	Lextra,:alias:	prod	\n\
8719	HTML-40	prod	ISOamsb	product operator	\n\
8720	-	Coproduct	:alias:	ISOamsb coprod	\n\
8720	-	coprod	ISOamsb	coproduct operator	\n\
8721	-	Sum	:alias:	ISOamsb sum	\n\
8721	-	sum	ISOamsb	summation operator	\n\
8722	HTML-40	minus	ISOtech	minus sign	\n\
8723	-	MinusPlus	:alias:	ISOtech mnplus	\n\
8723	-	mp	:alias:	ISOtech mnplus	\n\
8723	-	mnplus	ISOtech	minus-or-plus sign	\n\
8724	-	dotplus	:alias:	ISOamsb plusdo	\n\
8724	-	plusdo	ISOamsb	dotplus; plus sign, dot above	\n\
8725	-	-	-	division slash	\n\
8726	-	Backslash	:alias:	ISOamsb setmn	\n\
8726	-	setminus	:alias:	ISOamsb setmn	\n\
8726	-	setmn	ISOamsb	reverse solidus	\n\
8727	HTML-40	lowast	ISOtech	low asterisk	\n\
8728	-	SmallCircle	:alias:	ISOtech compfn	\n\
8728	-	compfn	ISOtech	composite function (small circle)	\n\
8730	-	Sqrt	:alias:	ISOtech radic	\n\
8730	HTML-40	radic	ISOtech	surd radical	\n\
8733	-	Proportional	:alias:	ISOtech prop	\n\
8733	-	propto	:alias:	ISOtech prop	\n\
8733	-	varpropto	:alias:	ISOamsr vprop	\n\
8733	-	vprop	ISOamsr,:alias:	proportional, variant	\n\
8733	HTML-40	prop	ISOtech	is proportional to	\n\
8734	HTML-40	infin	ISOtech	infinity	\n\
8735	-	angrt	ISOtech	right (90 degree) angle	\n\
8736	-	angle	:alias:	ISOamso ang	\n\
8736	HTML-40	ang	ISOamso	angle	\n\
8737	-	measuredangle	:alias:	ISOamso angmsd	\n\
8737	-	angmsd	ISOamso	angle-measured	\n\
8738	-	angsph	ISOtech	angle-spherical	\n\
8739	-	VerticalBar	:alias:	ISOamsr mid	\n\
8739	-	mid	ISOamsr	mid	\n\
8740	-	NotVerticalBar	:alias:	ISOamsn nmid	\n\
8740	-	nmid	ISOamsn	nmid	\n\
8741	-	DoubleVerticalBar	:alias:	ISOtech par	\n\
8741	-	parallel	:alias:	ISOtech par	\n\
8741	-	par	ISOtech	parallel	\n\
8742	-	NotDoubleVerticalBar	:alias:	ISOamsn npar	\n\
8742	-	nparallel	:alias:	ISOamsn npar	\n\
8742	-	npar	ISOamsn	not parallel	\n\
8743	-	wedge	:alias:	ISOtech and	\n\
8743	HTML-40	and	ISOtech	logical and	\n\
8744	-	vee	:alias:	ISOtech or	\n\
8744	HTML-40	or	ISOtech	logical or	\n\
8745	HTML-40	cap	ISOtech	intersection	\n\
8746	HTML-40	cup	ISOtech	union or logical sum	\n\
8747	-	Integral	:alias:	ISOtech int	\n\
8747	HTML-40	int	ISOtech	integral operator	\n\
8748	-	Int	ISOtech	double integral operator	\n\
8749	-	iiint	:alias:	ISOtech tint	\n\
8749	-	tint	ISOtech	integral operator	\n\
8750	-	ContourIntegral	:alias:	ISOtech conint	\n\
8750	-	oint	:alias:	ISOtech conint	\n\
8750	-	conint	ISOtech	contour integral operator	\n\
8751	-	DoubleContourIntegral	:alias:	ISOtech Conint	\n\
8751	-	Conint	ISOtech	double contour integral operator	\n\
8752	-	Cconint	ISOtech	triple contour integral operator	\n\
8753	-	cwint	ISOtech	clockwise integral	\n\
8754	-	ClockwiseContourIntegral	:alias:	ISOtech cwconint	\n\
8754	-	cwconint	ISOtech	contour integral, clockwise	\n\
8755	-	CounterClockwiseContourIntegral	:alias:	ISOtech awconint	\n\
8755	-	awconint	ISOtech	contour integral, anti-clockwise	\n\
8756	-	Therefore	:alias:	ISOtech there4	\n\
8756	-	therefore	:alias:	ISOtech there4	\n\
8756	HTML-40	there4	ISOtech	therefore	\n\
8757	-	Because	:alias:	ISOtech becaus	\n\
8757	-	because	:alias:	ISOtech becaus	\n\
8757	-	becaus	ISOtech	because	\n\
8758	-	ratio	ISOamsr	ratio	\n\
8759	-	Proportion	:alias:	ISOamsr Colon	\n\
8759	-	Colon	ISOamsr	two colons	\n\
8760	-	dotminus	:alias:	ISOamsb minusd	\n\
8760	-	minusd	ISOamsb	dotminus; minus sign, dot above	\n\
8762	-	mDDot	ISOamsr	minus with four dots, geometric properties	\n\
8763	-	homtht	ISOamsr	homothetic	\n\
8764	-	Tilde	:alias:	ISOtech sim	\n\
8764	HTML-40	sim	ISOtech	similar	\n\
8765	-	backsim	:alias:	ISOamsr bsim	\n\
8765	-	bsim	ISOamsr	reverse similar	\n\
8766	-	mstpos	ISOamsr	most positive	\n\
8768	-	VerticalTilde	:alias:	ISOamsb wreath	\n\
8768	-	wr	:alias:	ISOamsb wreath	\n\
8768	-	wreath	ISOamsb	wreath product	\n\
8769	-	NotTilde	:alias:	ISOamsn nsim	\n\
8769	-	nsim	ISOamsn	not similar	\n\
8770	-	EqualTilde	:alias:	ISOamsr esim	\n\
8770	-	eqsim	:alias:	ISOamsr esim	\n\
8770	-	esim	ISOamsr	equals, similar	\n\
8771	-	TildeEqual	:alias:	ISOtech sime	\n\
8771	-	simeq	:alias:	ISOtech sime	\n\
8771	-	sime	ISOtech	similar, equals	\n\
8772	-	NotTildeEqual	:alias:	ISOamsn nsime	\n\
8772	-	nsimeq	:alias:	ISOamsn nsime	\n\
8772	-	nsime	ISOamsn	not similar, equals	\n\
8773	-	TildeFullEqual	:alias:	ISOtech cong	\n\
8773	HTML-40	cong	ISOtech	congruent with	\n\
8774	-	simne	ISOamsn	similar, not equals	\n\
8775	-	NotTildeFullEqual	:alias:	ISOamsn ncong	\n\
8775	-	ncong	ISOamsn	not congruent with	\n\
8776	-	TildeTilde	:alias:	ISOtech ap	\n\
8776	-	approx	:alias:	ISOtech ap	\n\
8776	HTML-40	asymp	:alias:	almost equal to; asymptotic to	\n\
8776	-	ap	ISOtech	approximate	\n\
8777	-	NotTildeTilde	:alias:	ISOamsn nap	\n\
8777	-	napprox	:alias:	ISOamsn nap	\n\
8777	-	nap	ISOamsn	not approximate	\n\
8778	-	apE	ISOamsr,:alias:	approximately equal or equal to	\n\
8778	-	approxeq	:alias:	ISOamsr ape	\n\
8778	-	ape	ISOamsr	approximate, equals	\n\
8779	-	apid	ISOamsr	approximately identical to	\n\
8780	-	backcong	:alias:	ISOamsr bcong	\n\
8780	-	bcong	ISOamsr	reverse congruent	\n\
8781	-	CupCap	:alias:	ISOamsr asymp	\n\
8781	-	asymp	ISOamsr	asymptotically equal to	\n\
8782	-	Bumpeq	:alias:	ISOamsr bump	\n\
8782	-	HumpDownHump	:alias:	ISOamsr bump	\n\
8782	-	bump	ISOamsr	bumpy equals	\n\
8783	-	HumpEqual	:alias:	ISOamsr bumpe	\n\
8783	-	bumpeq	:alias:	ISOamsr bumpe	\n\
8783	-	bumpe	ISOamsr	bumpy equals, equals	\n\
8784	-	DotEqual	:alias:	ISOamsr esdot	\n\
8784	-	doteq	:alias:	ISOamsr esdot	\n\
8784	-	esdot	ISOamsr	equals, single dot above	\n\
8785	-	doteqdot	:alias:	ISOamsr eDot	\n\
8785	-	eDot	ISOamsr	eq, even dots	\n\
8786	-	fallingdotseq	:alias:	ISOamsr efDot	\n\
8786	-	efDot	ISOamsr	eq, falling dots	\n\
8787	-	risingdotseq	:alias:	ISOamsr erDot	\n\
8787	-	erDot	ISOamsr	eq, rising dots	\n\
8788	-	coloneq	:alias:	ISOamsr colone	\n\
8788	-	colone	ISOamsr	colon, equals	\n\
8789	-	eqcolon	:alias:	ISOamsr ecolon	\n\
8789	-	ecolon	ISOamsr	equals, colon	\n\
8790	-	eqcirc	:alias:	ISOamsr ecir	\n\
8790	-	ecir	ISOamsr	circle on equals sign	\n\
8791	-	circeq	:alias:	ISOamsr cire	\n\
8791	-	cire	ISOamsr	circle, equals	\n\
8793	-	wedgeq	ISOtech	corresponds to (wedge, equals)	\n\
8794	-	veeeq	ISOtech	logical or, equals	\n\
8795	-	easter	ISOamsr	equal, asterisk above	\n\
8796	-	triangleq	:alias:	ISOamsr trie	\n\
8796	-	trie	ISOamsr	triangle, equals	\n\
8799	-	questeq	:alias:	ISOamsr equest	\n\
8799	-	equest	ISOamsr	equal with questionmark	\n\
8800	-	NotEqual	:alias:	ISOtech ne	\n\
8800	HTML-40	ne	ISOtech	neq; not equal	\n\
8801	-	Congruent	:alias:	ISOtech equiv	\n\
8801	HTML-40	equiv	ISOtech	identical with	\n\
8802	-	NotCongruent	:alias:	ISOamsn nequiv	\n\
8802	-	nequiv	ISOamsn	not identical with	\n\
8804	-	leq	:alias:	ISOtech le	\n\
8804	HTML-40	le	ISOtech	less-than-or-equal	\n\
8805	-	GreaterEqual	:alias:	ISOtech ge	\n\
8805	-	geq	:alias:	ISOtech ge	\n\
8805	HTML-40	ge	ISOtech	greater-than-or-equal	\n\
8806	-	LessFullEqual	:alias:	ISOamsr lE	\n\
8806	-	leqq	:alias:	ISOamsr lE	\n\
8806	-	lE	ISOamsr	less, double equals	\n\
8807	-	GreaterFullEqual	:alias:	ISOamsr gE	\n\
8807	-	geqq	:alias:	ISOamsr gE	\n\
8807	-	gE	ISOamsr	greater, double equals	\n\
8808	-	lnE	ISOamsn,:alias:	less, not double equals	\n\
8808	-	lneq	:alias:	ISOamsn lne	\n\
8808	-	lneqq	:alias:	ISOamsn lnE	\n\
8808	-	lne	ISOamsn	less, not equals	\n\
8809	-	gnE	ISOamsn,:alias:	greater, not double equals	\n\
8809	-	gneq	:alias:	ISOamsn gne	\n\
8809	-	gneqq	:alias:	ISOamsn gnE	\n\
8809	-	gne	ISOamsn	greater, not equals	\n\
8810	-	NestedLessLess	:alias:	ISOamsr Lt	\n\
8810	-	ll	:alias:	ISOamsr Lt	\n\
8810	-	Lt	ISOamsr	double less-than sign	\n\
8811	-	NestedGreaterGreater	:alias:	ISOamsr Gt	\n\
8811	-	gg	:alias:	ISOamsr Gt	\n\
8811	-	Gt	ISOamsr	double greater-than sign	\n\
8812	-	between	:alias:	ISOamsr twixt	\n\
8812	-	twixt	ISOamsr	between	\n\
8813	-	NotCupCap	Lextra,:alias:	mmlextra nasymp	\n\
8814	-	NotLess	:alias:	ISOamsn nlt	\n\
8814	-	nless	:alias:	ISOamsn nlt	\n\
8814	-	nlt	ISOamsn,:alias:	not less-than	\n\
8814	-	nvlt	ISOamsn	not, vert, less-than	\n\
8815	-	NotGreater	:alias:	ISOamsn ngt	\n\
8815	-	ngtr	:alias:	ISOamsn ngt	\n\
8815	-	ngt	ISOamsn,:alias:	not greater-than	\n\
8815	-	nvgt	ISOamsn	not, vert, greater-than	\n\
8816	-	NotGreaterFullEqual	:alias:	ISOamsn nlE	\n\
8816	-	NotLessSlantEqual	:alias:	ISOamsn nles	\n\
8816	-	nleqq	:alias:	ISOamsn nlE	\n\
8816	-	nleqslant	:alias:	ISOamsn nles	\n\
8816	-	nlE	ISOamsn,:alias:	not less, double equals	\n\
8816	-	nles	ISOamsn,:alias:	not less-or-eq, slant	\n\
8816	-	nvle	ISOamsn	not, vert, less-than-or-equal	\n\
8817	-	NotGreaterSlantEqual	:alias:	ISOamsn nges	\n\
8817	-	ngE	ISOamsn,:alias:	not greater, double equals	\n\
8817	-	ngeqq	:alias:	ISOamsn ngE	\n\
8817	-	ngeqslant	:alias:	ISOamsn nges	\n\
8817	-	nges	ISOamsn,:alias:	not gt-or-eq, slanted	\n\
8817	-	nvge	ISOamsn	not, vert, greater-than-or-equal	\n\
8818	-	LessTilde	:alias:	ISOamsr lsim	\n\
8818	-	lap	ISOamsr,:alias:	less, approximate	\n\
8818	-	lessapprox	:alias:	ISOamsr lap	\n\
8818	-	lesssim	:alias:	ISOamsr lsim	\n\
8818	-	lsim	ISOamsr	less, similar	\n\
8819	-	GreaterTilde	:alias:	ISOamsr gsim	\n\
8819	-	gap	ISOamsr,:alias:	greater, approximate	\n\
8819	-	gtrapprox	:alias:	ISOamsr gap	\n\
8819	-	gtrsim	:alias:	ISOamsr gsim	\n\
8819	-	gsim	ISOamsr	greater, similar	\n\
8820	-	NotLessTilde	:alias:	ISOamsn nlsim	\n\
8820	-	nlsim	ISOamsn	not less, similar	\n\
8821	-	NotGreaterTilde	:alias:	ISOamsn ngsim	\n\
8821	-	ngsim	ISOamsn	not greater, similar	\n\
8822	-	LessGreater	:alias:	ISOamsr lg	\n\
8822	-	lessgtr	:alias:	ISOamsr lg	\n\
8822	-	lg	ISOamsr	less, greater	\n\
8823	-	GreaterLess	:alias:	ISOamsr gl	\n\
8823	-	gtrless	:alias:	ISOamsr gl	\n\
8823	-	gl	ISOamsr	greater, less	\n\
8824	-	NotLessGreater	:alias:	ISOamsn ntvlg	\n\
8824	-	ntlg	ISOamsn	not less, greater	\n\
8825	-	NotGreaterLess	:alias:	ISOamsn ntvgl	\n\
8825	-	ntgl	ISOamsn	not greater, less	\n\
8826	-	Precedes	:alias:	ISOamsr pr	\n\
8826	-	prec	:alias:	ISOamsr pr	\n\
8826	-	pr	ISOamsr	precedes	\n\
8827	-	Succeeds	:alias:	ISOamsr sc	\n\
8827	-	succ	:alias:	ISOamsr sc	\n\
8827	-	sc	ISOamsr	succeeds	\n\
8828	-	PrecedesSlantEqual	:alias:	ISOamsr prcue	\n\
8828	-	preccurlyeq	:alias:	ISOamsr prcue	\n\
8828	-	prcue	ISOamsr	precedes, curly eq	\n\
8829	-	SucceedsEqual	:alias:	ISOamsr sce	\n\
8829	-	SucceedsSlantEqual	:alias:	ISOamsr sccue	\n\
8829	-	sccue	ISOamsr,:alias:	succeeds, curly eq	\n\
8829	-	succcurlyeq	:alias:	ISOamsr sccue	\n\
8829	-	succeq	:alias:	ISOamsr sce	\n\
8829	-	sce	ISOamsr	succeeds, equals	\n\
8830	-	PrecedesTilde	:alias:	ISOamsr prsim	\n\
8830	-	precapprox	:alias:	ISOamsr prap	\n\
8830	-	precsim	:alias:	ISOamsr prsim	\n\
8830	-	scE	ISOamsr,:alias:	succeeds, double equals	\n\
8830	-	prap	ISOamsr,:alias:	precedes, approximate	\n\
8830	-	prsim	ISOamsr	precedes, similar	\n\
8831	-	SucceedsTilde	:alias:	ISOamsr scsim	\n\
8831	-	scap	ISOamsr,:alias:	succeeds, approximate	\n\
8831	-	succapprox	:alias:	ISOamsr scap	\n\
8831	-	succsim	:alias:	ISOamsr scsim	\n\
8831	-	scsim	ISOamsr	succeeds, similar	\n\
8832	-	NotPrecedes	:alias:	ISOamsn npr	\n\
8832	-	nprec	:alias:	ISOamsn npr	\n\
8832	-	npr	ISOamsn	not precedes	\n\
8833	-	NotSucceeds	:alias:	ISOamsn nsc	\n\
8833	-	nsucc	:alias:	ISOamsn nsc	\n\
8833	-	nsc	ISOamsn	not succeeds	\n\
8834	-	subset	:alias:	ISOtech sub	\n\
8834	HTML-40	sub	ISOtech	subset or is implied by	\n\
8835	-	Superset	:alias:	ISOtech sup	\n\
8835	-	supset	:alias:	ISOtech sup	\n\
8835	HTML-40	sup	ISOtech	superset or implies	\n\
8836	-	NotSubset	:alias:	ISOamsn vnsub	\n\
8836	-	nsubset	:alias:	ISOamsn vnsub	\n\
8836	-	vnsub	ISOamsn,:alias:	not subset, var	\n\
8836	HTML-40	nsub	ISOamsn	not subset	\n\
8837	-	NotSuperset	:alias:	ISOamsn vnsup	\n\
8837	-	nsupset	:alias:	ISOamsn vnsup	\n\
8837	-	vnsup	ISOamsn,:alias:	not superset, var	\n\
8837	-	nsup	ISOamsn	not superset	\n\
8838	-	SubsetEqual	:alias:	ISOtech sube	\n\
8838	-	subE	ISOamsr,:alias:	subset, double equals	\n\
8838	-	subseteq	:alias:	ISOtech sube	\n\
8838	-	subseteqq	:alias:	ISOamsr subE	\n\
8838	HTML-40	sube	ISOtech	subset, equals	\n\
8839	-	SupersetEqual	:alias:	ISOtech supe	\n\
8839	-	supseteq	:alias:	ISOtech supe	\n\
8839	-	supseteqq	:alias:	ISOamsr supE	\n\
8839	-	supE	ISOamsr,:alias:	superset, double equals	\n\
8839	HTML-40	supe	ISOtech	superset, equals	\n\
8840	-	NotSubsetEqual	:alias:	ISOamsn nsube	\n\
8840	-	nsubseteq	:alias:	ISOamsn nsube	\n\
8840	-	nsubseteqq	:alias:	ISOamsn nsubE	\n\
8840	-	nsubE	ISOamsn,:alias:	not subset, double eq	\n\
8840	-	nsube	ISOamsn	not subset, equals	\n\
8841	-	NotSupersetEqual	:alias:	ISOamsn nsupe	\n\
8841	-	nsupseteq	:alias:	ISOamsn nsupe	\n\
8841	-	nsupseteqq	:alias:	ISOamsn nsupE	\n\
8841	-	nsupE	ISOamsn,:alias:	not superset, double eq	\n\
8841	-	nsupe	ISOamsn	not superset, equals	\n\
8842	-	subsetneq	:alias:	ISOamsn subne	\n\
8842	-	subsetneqq	:alias:	ISOamsn subnE	\n\
8842	-	subnE	ISOamsn,:alias:	subset, not double eq	\n\
8842	-	subne	ISOamsn	subset, not equals	\n\
8843	-	supsetneq	:alias:	ISOamsn supne	\n\
8843	-	supsetneqq	:alias:	ISOamsn supnE	\n\
8843	-	supnE	ISOamsn,:alias:	superset, not double eq	\n\
8843	-	supne	ISOamsn	superset, not equals	\n\
8845	-	cupdot	ISOamsb	union, with dot	\n\
8846	-	UnionPlus	:alias:	ISOamsb uplus	\n\
8846	-	biguplus	:alias:	ISOamsb xuplus	\n\
8846	-	xuplus	ISOamsb,:alias:	big plus sign in union	\n\
8846	-	uplus	ISOamsb	plus sign in union	\n\
8847	-	SquareSubset	:alias:	ISOamsr sqsub	\n\
8847	-	sqsubset	:alias:	ISOamsr sqsub	\n\
8847	-	sqsub	ISOamsr	square subset	\n\
8848	-	SquareSuperset	:alias:	ISOamsr sqsup	\n\
8848	-	sqsupset	:alias:	ISOamsr sqsup	\n\
8848	-	sqsup	ISOamsr	square superset	\n\
8849	-	SquareSubsetEqual	:alias:	ISOamsr sqsube	\n\
8849	-	sqsubseteq	:alias:	ISOamsr sqsube	\n\
8849	-	sqsube	ISOamsr	square subset, equals	\n\
8850	-	SquareSupersetEqual	:alias:	ISOamsr sqsupe	\n\
8850	-	sqsupseteq	:alias:	ISOamsr sqsupe	\n\
8850	-	sqsupe	ISOamsr	square superset, eq	\n\
8851	-	SquareIntersection	:alias:	ISOamsb sqcap	\n\
8851	-	sqcap	ISOamsb	square intersection	\n\
8852	-	SquareUnion	:alias:	ISOamsb sqcup	\n\
8852	-	bigsqcup	:alias:	ISOamsb xsqcup	\n\
8852	-	xsqcup	ISOamsb,:alias:	square union operator	\n\
8852	-	sqcup	ISOamsb	square union	\n\
8853	-	CirclePlus	:alias:	ISOamsb oplus	\n\
8853	-	bigoplus	:alias:	ISOamsb xoplus	\n\
8853	-	xoplus	ISOamsb,:alias:	circle plus operator	\n\
8853	HTML-40	oplus	ISOamsb	plus sign in circle	\n\
8854	-	CircleMinus	:alias:	ISOamsb ominus	\n\
8854	-	ominus	ISOamsb	minus sign in circle	\n\
8855	-	CircleTimes	:alias:	ISOamsb otimes	\n\
8855	-	bigotimes	:alias:	ISOamsb xotime	\n\
8855	-	xotime	ISOamsb,:alias:	circle times operator	\n\
8855	HTML-40	otimes	ISOamsb	multiply sign in circle	\n\
8856	-	osol	ISOamsb	solidus in circle	\n\
8857	-	CircleDot	:alias:	ISOamsb odot	\n\
8857	-	bigodot	:alias:	ISOamsb xodot	\n\
8857	-	xodot	ISOamsb,:alias:	circle dot operator	\n\
8857	-	odot	ISOamsb	middle dot in circle	\n\
8858	-	circledcirc	:alias:	ISOamsb ocir	\n\
8858	-	ocir	ISOamsb	small circle in circle	\n\
8859	-	circledast	:alias:	ISOamsb oast	\n\
8859	-	oast	ISOamsb	asterisk in circle	\n\
8861	-	circleddash	:alias:	ISOamsb odash	\n\
8861	-	odash	ISOamsb	hyphen in circle	\n\
8862	-	boxplus	:alias:	ISOamsb plusb	\n\
8862	-	plusb	ISOamsb	plus sign in box	\n\
8863	-	boxminus	:alias:	ISOamsb minusb	\n\
8863	-	minusb	ISOamsb	minus sign in box	\n\
8864	-	boxtimes	:alias:	ISOamsb timesb	\n\
8864	-	timesb	ISOamsb	multiply sign in box	\n\
8865	-	dotsquare	:alias:	ISOamsb sdotb	\n\
8865	-	sdotb	ISOamsb	boxdot; small dot in box	\n\
8866	-	RightTee	:alias:	ISOamsr vdash	\n\
8866	-	vdash	ISOamsr	vertical, dash	\n\
8867	-	LeftTee	:alias:	ISOamsr dashv	\n\
8867	-	dashv	ISOamsr	dash, vertical	\n\
8868	-	DownTee	:alias:	ISOtech top	\n\
8868	-	top	ISOtech	top	\n\
8869	-	UpTee	:alias:	ISOtech perp	\n\
8869	-	bot	:alias:	ISOtech bottom	\n\
8869	-	bottom	ISOtech,:alias:	bottom	\n\
8869	HTML-40	perp	ISOtech	perpendicular	\n\
8871	-	models	ISOamsr	models	\n\
8872	-	DoubleRightTee	:alias:	ISOamsr vDash	\n\
8872	-	vDash	ISOamsr	vertical, double dash	\n\
8873	-	Vdash	ISOamsr	double vertical, dash	\n\
8874	-	Vvdash	ISOamsr	triple vertical, dash	\n\
8875	-	VDash	ISOamsr	double vert, double dash	\n\
8876	-	nvdash	ISOamsn	not vertical, dash	\n\
8877	-	nvDash	ISOamsn	not vertical, double dash	\n\
8878	-	nVdash	ISOamsn	not double vertical, dash	\n\
8879	-	nVDash	ISOamsn	not double vert, double dash	\n\
8880	-	prurel	ISOamsr	element precedes under relation	\n\
8882	-	LeftTriangle	:alias:	ISOamsr vltri	\n\
8882	-	vartriangleleft	:alias:	ISOamsr vltri	\n\
8882	-	vltri	ISOamsr	left triangle, open, var	\n\
8883	-	RightTriangle	:alias:	ISOamsr vrtri	\n\
8883	-	vartriangleright	:alias:	ISOamsr vrtri	\n\
8883	-	vrtri	ISOamsr	right triangle, open, var	\n\
8884	-	LeftTriangleEqual	:alias:	ISOamsr ltrie	\n\
8884	-	trianglelefteq	:alias:	ISOamsr ltrie	\n\
8884	-	ltrie	ISOamsr	left triangle, eq	\n\
8885	-	RightTriangleEqual	:alias:	ISOamsr rtrie	\n\
8885	-	trianglerighteq	:alias:	ISOamsr rtrie	\n\
8885	-	rtrie	ISOamsr	right triangle, eq	\n\
8886	-	origof	ISOamsa	original of	\n\
8887	-	imof	ISOamsa	image of	\n\
8888	-	multimap	:alias:	ISOamsa mumap	\n\
8888	-	mumap	ISOamsa	multimap	\n\
8889	-	hercon	ISOamsb	hermitian conjugate matrix	\n\
8890	-	intercal	:alias:	ISOamsb intcal	\n\
8890	-	intcal	ISOamsb	intercal	\n\
8891	-	veebar	ISOamsb	logical or, bar below	\n\
8892	-	barwedge	:alias:	ISOamsb barwed	\n\
8892	-	barwed	ISOamsb	logical and, bar above	\n\
8893	-	barvee	ISOamsb	bar, vee	\n\
8894	-	vangrt	ISOtech	right angle, variant	\n\
8896	-	Wedge	:alias:	ISOamsb xwedge	\n\
8896	-	bigwedge	:alias:	ISOamsb xwedge	\n\
8896	-	xwedge	ISOamsb	logical or operator	\n\
8897	-	Vee	:alias:	ISOamsb xvee	\n\
8897	-	bigvee	:alias:	ISOamsb xvee	\n\
8897	-	xvee	ISOamsb	logical and operator	\n\
8898	-	Intersection	:alias:	ISOamsb xcap	\n\
8898	-	bigcap	:alias:	ISOamsb xcap	\n\
8898	-	xcap	ISOamsb	intersection operator	\n\
8899	-	Union	:alias:	ISOamsb xcup	\n\
8899	-	bigcup	:alias:	ISOamsb xcup	\n\
8899	-	xcup	ISOamsb	union operator	\n\
8900	-	Diamond	:alias:	ISOamsb diam	\n\
8900	-	diamond	:alias:	ISOamsb diam	\n\
8900	-	diam	ISOamsb	open diamond	\n\
8901	HTML-40	sdot	ISOamsb	small middle dot	\n\
8902	-	Star	:alias:	ISOamsb sstarf	\n\
8902	-	sstarf	ISOamsb,:alias:	small star, filled	\n\
8902	-	star	ISOpub	star, open	\n\
8903	-	divideontimes	:alias:	ISOamsb divonx	\n\
8903	-	divonx	ISOamsb	division on times	\n\
8904	-	bowtie	ISOamsr	bowtie	\n\
8905	-	ltimes	ISOamsb	times sign, left closed	\n\
8906	-	rtimes	ISOamsb	times sign, right closed	\n\
8907	-	leftthreetimes	:alias:	ISOamsb lthree	\n\
8907	-	lthree	ISOamsb	left three times	\n\
8908	-	rightthreetimes	:alias:	ISOamsb rthree	\n\
8908	-	rthree	ISOamsb	right three times	\n\
8909	-	backsimeq	:alias:	ISOamsr bsime	\n\
8909	-	bsime	ISOamsr	reverse similar, eq	\n\
8910	-	curlyvee	:alias:	ISOamsb cuvee	\n\
8910	-	cuvee	ISOamsb	curly logical or	\n\
8911	-	curlywedge	:alias:	ISOamsb cuwed	\n\
8911	-	cuwed	ISOamsb	curly logical and	\n\
8912	-	Subset	:alias:	ISOamsr Sub	\n\
8912	-	Sub	ISOamsr	double subset	\n\
8913	-	Supset	:alias:	ISOamsr Sup	\n\
8913	-	Sup	ISOamsr	double superset	\n\
8914	-	Cap	ISOamsb	doublecap; double intersection	\n\
8915	-	Cup	ISOamsb	doublecup; double union	\n\
8916	-	pitchfork	:alias:	ISOamsr fork	\n\
8916	-	fork	ISOamsr	pitchfork	\n\
8917	-	epar	ISOtech	parallel, equal; equal or parallel	\n\
8918	-	lessdot	:alias:	ISOamsr ltdot	\n\
8918	-	ltdot	ISOamsr	less than, with dot	\n\
8919	-	gtrdot	:alias:	ISOamsr gtdot	\n\
8919	-	gtdot	ISOamsr	greater than, with dot	\n\
8920	-	Ll	ISOamsr	triple less-than	\n\
8921	-	ggg	:alias:	ISOamsr Gg	\n\
8921	-	Gg	ISOamsr	triple gtr-than	\n\
8922	-	LessEqualGreater	:alias:	ISOamsr leg	\n\
8922	-	lesseqgtr	:alias:	ISOamsr leg	\n\
8922	-	lesseqqgtr	:alias:	ISOamsr lEg	\n\
8922	-	lEg	ISOamsr,:alias:	less, double eq, greater	\n\
8922	-	leg	ISOamsr	less, eq, greater	\n\
8923	-	GreaterEqualLess	:alias:	ISOamsr gel	\n\
8923	-	gtreqless	:alias:	ISOamsr gel	\n\
8923	-	gtreqqless	:alias:	ISOamsr gEl	\n\
8923	-	gEl	ISOamsr,:alias:	gt, double equals, less	\n\
8923	-	gel	ISOamsr	greater, equals, less	\n\
8924	-	eqslantless	:alias:	ISOamsr els	\n\
8924	-	els	ISOamsr	eq-or-less, slanted	\n\
8925	-	eqslantgtr	:alias:	ISOamsr egs	\n\
8925	-	egs	ISOamsr	equal-or-gtr, slanted	\n\
8926	-	curlyeqprec	:alias:	ISOamsr cuepr	\n\
8926	-	cuepr	ISOamsr	curly eq, precedes	\n\
8927	-	curlyeqsucc	:alias:	ISOamsr cuesc	\n\
8927	-	cuesc	ISOamsr	curly eq, succeeds	\n\
8928	-	NotPrecedesSlantEqual	:alias:	ISOamsn nprcue	\n\
8928	-	nprcue	ISOamsn	not curly precedes, eq	\n\
8929	-	NotSucceedsSlantEqual	:alias:	ISOamsn nsccue	\n\
8929	-	nsccue	ISOamsn	not succeeds, curly eq	\n\
8930	-	NotSquareSubsetEqual	:alias:	ISOamsn nsqsube	\n\
8930	-	nsqsube	ISOamsn	not, square subset, equals	\n\
8931	-	NotSquareSupersetEqual	:alias:	ISOamsn nsqsupe	\n\
8931	-	nsqsupe	ISOamsn	not, square superset, equals	\n\
8934	-	lnsim	ISOamsn	less, not similar	\n\
8935	-	gnsim	ISOamsn	greater, not similar	\n\
8936	-	precnapprox	:alias:	ISOamsn prnap	\n\
8936	-	precnsim	:alias:	ISOamsn prnsim	\n\
8936	-	prnap	ISOamsn	precedes, not approx	\n\
8936	-	prnsim	ISOamsn	precedes, not similar	\n\
8937	-	succnapprox	:alias:	ISOamsn scnap	\n\
8937	-	succnsim	:alias:	ISOamsn scnsim	\n\
8937	-	scnap	ISOamsn,:alias:	succeeds, not approx	\n\
8937	-	scnsim	ISOamsn	succeeds, not similar	\n\
8938	-	NotLeftTriangle	:alias:	ISOamsn nltri	\n\
8938	-	ntriangleleft	:alias:	ISOamsn nltri	\n\
8938	-	nltri	ISOamsn	not left triangle	\n\
8939	-	NotRightTriangle	:alias:	ISOamsn nrtri	\n\
8939	-	ntriangleright	:alias:	ISOamsn nrtri	\n\
8939	-	nrtri	ISOamsn	not right triangle	\n\
8940	-	NotLeftTriangleEqual	:alias:	ISOamsn nltrie	\n\
8940	-	ntrianglelefteq	:alias:	ISOamsn nltrie	\n\
8940	-	nltrie	ISOamsn	not left triangle, eq	\n\
8941	-	NotRightTriangleEqual	:alias:	ISOamsn nrtrie	\n\
8941	-	ntrianglerighteq	:alias:	ISOamsn nrtrie	\n\
8941	-	nrtrie	ISOamsn	not right triangle, eq	\n\
8942	-	vellip	ISOpub	vertical ellipsis	\n\
8943	-	ctdot	ISOtech	three dots, centered	\n\
8944	-	utdot	ISOtech	three dots, ascending	\n\
8945	22f1	22F1	-	dtdot	ISOtech	three dots, descending	\n\
	\n\
#	23	\n\
	\n\
8966	-	doublebarwedge	:alias:	ISOamsb Barwed	\n\
8966	-	Barwed	ISOamsb	log and, double bar above	\n\
8968	-	LeftCeiling	:alias:	ISOAMSC lceil	\n\
8968	HTML-40	lceil	ISOamsc	left ceiling	\n\
8969	-	RightCeiling	:alias:	ISOAMSC rceil	\n\
8969	HTML-40	rceil	ISOamsc	right ceiling	\n\
8970	-	LeftFloor	:alias:	ISOAMSC lfloor	\n\
8970	HTML-40	lfloor	ISOamsc	left floor	\n\
8971	-	RightFloor	:alias:	ISOAMSC rfloor	\n\
8971	HTML-40	rfloor	ISOamsc	right floor	\n\
8972	-	drcrop	ISOpub	downward right crop mark 	\n\
8973	-	dlcrop	ISOpub	downward left crop mark 	\n\
8974	-	urcrop	ISOpub	upward right crop mark 	\n\
8975	-	ulcrop	ISOpub	upward left crop mark 	\n\
8976	-	bnot	ISOtech	reverse not	\n\
8978	-	profline	ISOtech	profile of a line	\n\
8979	-	profsurf	ISOtech	profile of a surface	\n\
8981	-	telrec	ISOpub	telephone recorder symbol	\n\
8982	-	target	ISOpub	register mark or target	\n\
8988	-	ulcorner	:alias:	ISOAMSC ulcorn	\n\
8988	-	ulcorn	ISOamsc	upper left corner	\n\
8989	-	urcorner	:alias:	ISOAMSC urcorn	\n\
8989	-	urcorn	ISOamsc	upper right corner	\n\
8990	-	llcorner	:alias:	ISOAMSC dlcorn	\n\
8990	-	dlcorn	ISOamsc	lower left corner	\n\
8991	-	lrcorner	:alias:	ISOAMSC drcorn	\n\
8991	-	drcorn	ISOamsc	lower right corner	\n\
8994	-	frown	ISOamsr	frown; down curve	\n\
8995	-	smile	ISOamsr	smile; up curve	\n\
9001	HTML-40	lang	ISOtech	left-pointing angle bracket; bra	\n\
9002	HTML-40	rang	ISOtech	right-pointing angle bracket; ket	\n\
9005	-	cylcty	ISOtech	cylindricity	\n\
9006	-	profalar	ISOtech	all-around profile	\n\
9014	-	topbot	ISOtech	top and bottom	\n\
	\n\
#	24	\n\
	\n\
9251	-	blank	ISOpub	significant blank symbol	\n\
9416	-	circledS	:alias:	ISOamso oS	\n\
9416	-	oS	ISOamso	capital S in circle	\n\
	\n\
#	25	\n\
	\n\
9472	-	boxh	isobox	horizontal line 	\n\
9474	-	boxv	isobox	vertical line	\n\
9484	-	boxdr	isobox	lower right quadrant	\n\
9488	-	boxdl	isobox	lower left quadrant	\n\
9492	-	boxur	isobox	upper right quadrant	\n\
9496	-	boxul	isobox	upper left quadrant	\n\
9500	-	boxvr	isobox	upper and lower right quadrants	\n\
9508	-	boxvl	isobox	upper and lower left quadrants	\n\
9516	-	boxhd	isobox	lower left and right quadrants	\n\
9524	-	boxhu	isobox	upper left and right quadrants	\n\
9532	-	boxvh	isobox	all four quadrants	\n\
9552	-	boxH	isobox	horizontal line	\n\
9553	-	boxV	isobox	vertical line	\n\
9554	-	boxdR	isobox	lower right quadrant	\n\
9555	-	boxDr	isobox	lower right quadrant	\n\
9556	-	boxDR	isobox	lower right quadrant	\n\
9557	-	boxdL	isobox	lower left quadrant	\n\
9558	-	boxDl	isobox	lower left quadrant	\n\
9559	-	boxDL	isobox	lower left quadrant	\n\
9560	-	boxuR	isobox	upper right quadrant	\n\
9561	-	boxUr	isobox	upper right quadrant	\n\
9562	-	boxUR	isobox	upper right quadrant	\n\
9563	-	boxuL	isobox	upper left quadrant	\n\
9564	-	boxUl	isobox	upper left quadrant	\n\
9565	-	boxUL	isobox	upper left quadrant	\n\
9566	-	boxvR	isobox	upper and lower right quadrants	\n\
9567	-	boxVr	isobox	upper and lower right quadrants	\n\
9568	-	boxVR	isobox	upper and lower right quadrants	\n\
9569	-	boxvL	isobox	upper and lower left quadrants	\n\
9570	-	boxVl	isobox	upper and lower left quadrants	\n\
9571	-	boxVL	isobox	upper and lower left quadrants	\n\
9572	-	boxHd	isobox	lower left and right quadrants	\n\
9573	-	boxhD	isobox	lower left and right quadrants	\n\
9574	-	boxHD	isobox	lower left and right quadrants	\n\
9575	-	boxHu	isobox	upper left and right quadrants	\n\
9576	-	boxhU	isobox	upper left and right quadrants	\n\
9577	-	boxHU	isobox	upper left and right quadrants	\n\
9578	-	boxvH	isobox	all four quadrants	\n\
9579	-	boxVh	isobox	all four quadrants	\n\
9580	-	boxVH	isobox	all four quadrants	\n\
9600	-	uhblk	ISOpub	upper half block	\n\
9604	-	lhblk	ISOpub	lower half block	\n\
9608	-	block	ISOpub	full block	\n\
9617	-	blk14	ISOpub	25% shaded block	\n\
9618	-	blk12	ISOpub	50% shaded block	\n\
9619	-	blk34	ISOpub	75% shaded block	\n\
9633	-	Square	:alias:	for square	\n\
9633	-	squ	ISOpub,:alias:	square, open	\n\
9633	-	square	ISOtech	square	\n\
9642	-	blacksquare	:alias:	ISOtech squarf 	\n\
9642	-	squf	ISOpub,:alias:	square bullet, filled	\n\
9642	-	squarf	ISOtech	blacksquare, square, filled 	\n\
9645	-	rect	ISOpub	rectangle, open	\n\
9646	-	marker	ISOpub	histogram marker	\n\
9651	-	bigtriangleup	:alias:	ISOamsb xutri	\n\
9651	-	xutri	ISOamsb	big up triangle, open	\n\
9652	-	blacktriangle	:alias:	ISOpub utrif	\n\
9652	-	utrif	ISOpub	up triangle, filled	\n\
9653	-	triangle	:alias:	ISOpub utri	\n\
9653	-	utri	ISOpub	up triangle, open	\n\
9654	-	-	-	-	\n\
9655	-	-	-	-	\n\
9656	-	blacktriangleright	:alias:	ISOpub rtrif	\n\
9656	-	rtrif	ISOpub	right triangle, filled	\n\
9657	-	triangleright	:alias:	ISOpub rtri	\n\
9657	-	rtri	ISOpub	right triangle, open	\n\
9658	-	-	-	-	\n\
9659	-	-	-	-	\n\
9660	-	-	-	-	\n\
9661	-	bigtriangledown	:alias:	ISOamsb xdtri	\n\
9661	-	xdtri	ISOamsb	big down triangle, open	\n\
9662	-	blacktriangledown	:alias:	ISOpub dtrif	\n\
9662	-	dtrif	ISOpub	down triangle, filled	\n\
9663	-	triangledown	:alias:	ISOpub dtri	\n\
9663	-	dtri	ISOpub	down triangle, open	\n\
9664	-	-	-	-	\n\
9665	-	-	-	-	\n\
9666	-	blacktriangleleft	:alias:	ISOpub ltrif	\n\
9666	-	ltrif	ISOpub	left triangle, filled	\n\
9667	-	triangleleft	:alias:	ISOpub ltri	\n\
9667	-	ltri	ISOpub	left triangle, open	\n\
9674	-	lozenge	:alias:	ISOpub loz	\n\
9674	HTML-40	loz	ISOpub	lozenge or total mark	\n\
9675	-	cir	ISOpub	circle, open	\n\
9708	-	tridot	ISOamsb	dot in triangle	\n\
9711	-	bigcirc	:alias:	ISOamsb xcirc	\n\
9711	-	xcirc	ISOamsb	large circle	\n\
	\n\
#	26	\n\
	\n\
9733	-	bigstar	:alias:	ISOpub starf 	\n\
9733	-	starf	ISOpub	star, filled 	\n\
9742	-	phone	ISOpub	telephone symbol 	\n\
9792	-	female	ISOpub	female symbol	\n\
9794	-	male	ISOpub	male symbol	\n\
9824	-	spadesuit	:alias:	ISOpub spades 	\n\
9824	HTML-40	spades	ISOpub	spades suit symbol 	\n\
9825	-	heartsuit	:alias:	ISOpub hearts 	\n\
9825	-	hearts	ISOpub	heart suit symbol 	\n\
9826	-	-	-	-	\n\
9827	-	clubsuit	:alias:	ISOpub clubs 	\n\
9827	HTML-40	clubs	ISOpub	club suit symbol 	\n\
9828	-	-	-	-	\n\
9829	HTML-40	hearts	ISO8879	black heart suit; valentine	\n\
9830	HTML-40	diams	ISOpub	diamond suit symbol 	\n\
9830	-	diamondsuit	:alias:	ISOpub diams 	\n\
9834	-	sung	ISOnum	music note (sung text sign)	\n\
9837	-	flat	ISOpub	musical flat	\n\
9838	-	natural	:alias:	ISOpub natur	\n\
9838	-	natur	ISOpub	music natural	\n\
9839	-	sharp	ISOpub	musical sharp	\n\
	\n\
#	27	\n\
	\n\
10003	-	checkmark	:alias:	ISOpub check	\n\
10003	-	check	ISOpub	tick, check mark	\n\
10007	-	cross	ISOpub	ballot cross	\n\
10016	-	maltese	:alias:	ISOpub malt	\n\
10016	-	malt	ISOpub	maltese cross	\n\
10038	-	sext	ISOpub	sextile (6-pointed star)	\n\
	\n\
#	30	\n\
	\n\
12288	-	-	-	ideographic space (for CJK character)	\n\
12296	-	LeftAngleBracket	:alias:	ISOtech lang	\n\
12296	-	langle	:alias:	ISOtech lang	\n\
12296	-	lang	ISOtech	left angle bracket	\n\
12297	-	RightAngleBracket	:alias:	ISOtech rang	\n\
12297	-	rangle	:alias:	ISOtech rang	\n\
12297	-	rang	ISOtech	right angle bracket	\n\
12298	-	Lang	ISOtech	left angle bracket, double	\n\
12299	-	Rang	ISOtech	right angle bracket, double	\n\
12308	-	lbbrk	ISOtech	left broken bracket	\n\
12309	-	rbbrk	ISOtech	right broken bracket	\n\
12310	-	-	-	-	\n\
12311	-	-	-	-	\n\
12312	-	loang	ISOtech	left open angular bracket	\n\
12313	-	roang	ISOtech	right open angular bracket	\n\
12314	-	LeftDoubleBracket	:alias:	left double bracket delimiter	\n\
12314	-	lobrk	ISOtech	left open bracket	\n\
12315	-	RightDoubleBracket	:alias:	right double bracket delimiter	\n\
12315	-	robrk	ISOtech	right open bracket	\n\
	\n\
#	E2	\n\
	\n\
57856	-	DoubleLongLeftArrow	:alias:	ISOamsa xlArr	\n\
57856	-	Longleftarrow	:alias:	ISOamsa xlArr	\n\
57856	-	xlArr	ISOamsa	long left double arrow	\n\
57857	-	LongLeftArrow	:alias:	ISOamsa xlarr	\n\
57857	-	longleftarrow	:alias:	ISOamsa xlarr	\n\
57857	-	xlarr	ISOamsa	long left arrow	\n\
57858	-	DoubleLongLeftRightArrow	:alias:	ISOamsa xhArr	\n\
57858	-	Longleftrightarrow	:alias:	ISOamsa xhArr	\n\
57858	-	xhArr	ISOamsa	long left and right double arrow	\n\
57859	-	LongLeftRightArrow	:alias:	ISOamsa xharr	\n\
57859	-	longleftrightarrow	:alias:	ISOamsa xharr	\n\
57859	-	xharr	ISOamsa	long left and right arrow	\n\
57860	-	DoubleLongRightArrow	:alias:	ISOamsa xrArr	\n\
57860	-	Longrightarrow	:alias:	ISOamsa xrArr	\n\
57860	-	xrArr	ISOamsa	long right double arrow	\n\
57861	-	LongRightArrow	:alias:	ISOamsa xrarr	\n\
57861	-	longrightarrow	:alias:	ISOamsa xrarr	\n\
57861	-	xrarr	ISOamsa	long right arrow	\n\
57862	-	lBarr	ISOamsa	left doubly broken arrow	\n\
57863	-	ac	ISOamsb,:alias:	most positive	\n\
57863	-	dbkarow	:alias:	ISOamsa rBarr	\n\
57863	-	rBarr	ISOamsa	right doubly broken arrow	\n\
57864	-	longmapsto	:alias:	ISOamsa xmap	\n\
57864	-	xmap	ISOamsa	long maps to	\n\
57865	-	drbkarow	:alias:	ISOamsa RBarr	\n\
57865	-	RBarr	ISOamsa	two-headed right broken arrow	\n\
57866	-	hkswarow	:alias:	ISOamsa swarhk	\n\
57866	-	swarhk	ISOamsa	SW arrow-hooked	\n\
57867	-	hksearow	:alias:	ISOamsa searhk	\n\
57867	-	searhk	ISOamsa	SE arrow-hooken	\n\
57868	-	nwarhk	ISOamsa	NW arrow-hooked	\n\
57869	-	nearhk	ISOamsa	NE arrow-hooked	\n\
57870	-	toea	:alias:	ISOamsa nesear	\n\
57870	-	nesear	ISOamsa	NE & SE arrows	\n\
57871	-	tosa	:alias:	ISOamsa seswar	\n\
57871	-	seswar	ISOamsa	SE & SW arrows	\n\
57872	-	swnwar	ISOamsa	SW & NW arrows	\n\
57873	-	nwnear	ISOamsa	NW & NE arrows	\n\
57874	-	Map	ISOamsa	two-headed maps to	\n\
57876	-	lfisht	ISOamsa	left fish tail	\n\
57877	-	rfisht	ISOamsa	right fish tail	\n\
57878	-	DownArrowUpArrow	:alias:	ISOamsa duarr	\n\
57878	-	duarr	ISOamsa	down arrow, up arrow	\n\
57879	-	ReverseUpEquilibrium	:alias:	ISOamsa duhar	\n\
57879	-	duhar	ISOamsa	down harp, up harp	\n\
57880	-	UpEquilibrium	:alias:	ISOamsa udhar	\n\
57880	-	udhar	ISOamsa	up harp, down harp	\n\
57881	-	rdca	ISOamsa	right down curved arrow	\n\
57882	-	ldca	ISOamsa	left down curved arrow	\n\
57883	-	nrarrw	ISOamsa	not right arrow-wavy	\n\
57884	-	rarrc	ISOamsa	right arrow-curved	\n\
57885	-	nrarrc	ISOamsa	not right arrow-curved	\n\
57886	-	rarrpl	ISOamsa	right arrow, plus	\n\
57888	-	larrbfs	ISOamsa	left arrow-bar, filled square	\n\
57889	-	rarrbfs	ISOamsa	right arrow-bar, filled square	\n\
57890	-	larrfs	ISOamsa	left arrow, filled square	\n\
57891	-	rarrfs	ISOamsa	right arrow, filled square	\n\
57892	-	rHar	ISOamsa	right harpoon-up over right harpoon-down	\n\
57893	-	lHar	ISOamsa	left harpoon-up over left harpoon-down	\n\
57894	-	uHar	ISOamsa	up harpoon-left, up harpoon-right	\n\
57895	-	dHar	ISOamsa	down harpoon-left, down harpoon-right	\n\
57896	-	ldrushar	ISOamsa	left-down-right-up harpoon	\n\
57897	-	lurdshar	ISOamsa	left-up-right-down harpoon	\n\
57898	-	ruluhar	ISOamsa	right harpoon-up over left harpoon-up	\n\
57899	-	luruhar	ISOamsa	left harpoon-up over right harpoon-up	\n\
57900	-	ldrdhar	ISOamsa	left harpoon-down over right harpoon-down	\n\
57901	-	rdldhar	ISOamsa	right harpoon-down over left harpoon-down	\n\
57902	-	lharul	ISOamsa	left harpoon-up over long dash	\n\
57903	-	lrhard	ISOamsa	right harpoon-down below long dash	\n\
57904	-	rharul	ISOamsa	right harpoon-up over long dash	\n\
57905	-	llhard	ISOamsa	left harpoon-down below long dash	\n\
57906	-	ShortRightArrow	:alias:	ISOamsa srarr	\n\
57906	-	srarr	ISOamsa	short right arrow	\n\
57907	-	ShortLeftArrow	:alias:	ISOamsa slarr	\n\
57907	-	slarr	ISOamsa	short left arrow	\n\
57908	-	simrarr	ISOamsa	similar, right arrow below	\n\
57909	-	rarrap	ISOamsa	approximate, right arrow above	\n\
57910	-	erarr	ISOamsa	equal, right arrow below	\n\
57911	-	Uarrocir	ISOamsa	up two-headed arrow above circle	\n\
57912	-	DDotrahd	ISOamsa	right arrow with dotted stem	\n\
57913	-	Rarrtl	ISOamsa	right two-headed arrow with tail	\n\
57915	-	rAtail	ISOamsa	right double arrow-tail	\n\
57916	-	latail	ISOamsa	left arrow-tail	\n\
57917	-	lAtail	ISOamsa	left double arrow-tail	\n\
57918	-	cudarrl	ISOamsa	left, curved, down arrow	\n\
57919	-	larrpl	ISOamsa	left arrow, plus	\n\
57920	-	harrcir	ISOamsa	left and right arrow with a circle	\n\
57921	-	roarr	ISOamsa	right open arrow	\n\
57922	-	loarr	ISOamsa	left open arrow	\n\
57923	-	hoarr	ISOamsa	horizontal open arrow	\n\
57924	-	zigrarr	ISOamsa	right zig-zag arrow	\n\
57928	-	angzarr	ISOamsa	angle with down zig-zag arrow	\n\
57929	-	curarrm	ISOamsa	curved right arrow with minus	\n\
57930	-	cularrp	ISOamsa	curved left arrow with plus	\n\
57931	-	ufisht	ISOamsa	up fish tail	\n\
57932	-	dfisht	ISOamsa	down fish tail	\n\
57933	-	rarrsim	ISOamsa	right arrow, similar	\n\
57934	-	larrsim	ISOamsa	left arrow, similar	\n\
57935	-	midcir	ISOamsa	mid, circle below 	\n\
57936	-	cirmid	ISOamsa	circle, mid below	\n\
57937	-	amalg	ISOamsb	amalgamation or coproduct	\n\
57945	-	intprod	:alias:	ISOamsb iprod	\n\
57945	-	iprod	ISOamsb	intprod	\n\
57946	-	plusdu	ISOamsb	plus sign, dot below	\n\
57947	-	minusdu	ISOamsb	minus sign, dot below	\n\
57948	-	loplus	ISOamsb	plus sign in left half circle	\n\
57949	-	roplus	ISOamsb	plus sign in right half circle	\n\
57950	-	lotimes	ISOamsb	multiply sign in left half circle 	\n\
57951	-	rotimes	ISOamsb	multiply sign in right half circle	\n\
57952	-	ohbar	ISOamsb	circle with horizontal bar	\n\
57953	-	capdot	ISOamsb	intersection, with dot	\n\
57954	-	subdot	ISOamsb	subset, with dot	\n\
57955	-	supdot	ISOamsb	superset, with dot	\n\
57956	-	smashp	ISOamsb	smash product	\n\
57957	-	wedbar	ISOamsb	wedge, bar below	\n\
57958	-	pluscir	ISOamsb	plus, small circle above	\n\
57959	-	pluse	ISOamsb	plus, equals	\n\
57960	-	eplus	ISOamsb	equal, plus	\n\
57961	-	plustwo	ISOamsb	plus, two; Nim-addition	\n\
57962	-	plusacir	ISOamsb	plus, circumflex accent above	\n\
57963	-	simplus	ISOamsb	plus, similar above	\n\
57964	-	plussim	ISOamsb	plus, similar below	\n\
57965	-	timesd	ISOamsb	times, dot	\n\
57966	-	cupcap	ISOamsb	union above intersection	\n\
57967	-	capcup	ISOamsb	intersection above union	\n\
57968	-	cupbrcap	ISOamsb	union, bar, intersection	\n\
57969	-	capbrcup	ISOamsb	intersection, bar, union	\n\
57970	-	cupcup	ISOamsb	union, union, joined	\n\
57971	-	capcap	ISOamsb	intersection, intersection, joined	\n\
57972	-	cups	ISOamsb	union, serifs	\n\
57973	-	caps	ISOamsb	intersection, serifs	\n\
57974	-	sqcups	ISOamsb	square union, serifs	\n\
57975	-	sqcaps	ISOamsb	square intersection, serifs	\n\
57976	-	ccups	ISOamsb	closed union, serifs	\n\
57977	-	ccaps	ISOamsb	closed intersection, serifs	\n\
57978	-	ccupssm	ISOamsb	closed union, serifs, smash product	\n\
57979	-	triplus	ISOamsb	plus in triangle	\n\
57980	-	triminus	ISOamsb	minus in triangle	\n\
57981	-	tritime	ISOamsb	multiply in triangle	\n\
57982	-	trisb	ISOamsb	triangle, serifs at bottom	\n\
57983	-	solb	ISOamsb	solidus in square	\n\
57984	-	bsolb	ISOamsb	reverse solidus in square	\n\
57985	-	capand	ISOamsb	intersection, and	\n\
57986	-	cupor	ISOamsb	union, or	\n\
57987	-	ncup	ISOamsb	bar, union	\n\
57988	-	ncap	ISOamsb	bar, intersection	\n\
57989	-	odiv	ISOamsb	divide in circle	\n\
57990	-	odsold	ISOamsb	dot, solidus, dot in circle	\n\
57991	-	ofcir	ISOamsb	filled circle in circle	\n\
57992	-	olt	ISOamsb	less-than in circle	\n\
57993	-	ogt	ISOamsb	greater-than in circle	\n\
57994	-	opar	ISOamsb	parallel in circle	\n\
57995	-	operp	ISOamsb	perpendicular in circle	\n\
57996	-	Otimes	ISOamsb	multiply sign in double circle	\n\
57997	-	otimesas	ISOamsb	multiply sign in circle, circumflex accent	\n\
57998	-	timesbar	ISOamsb	multiply sign, bar below	\n\
58000	-	acE	ISOamsb	most positive, two lines below	\n\
58001	-	rpargt	ISOamsc	right parenthesis, gt	\n\
58002	-	lparlt	ISOamsc	left parenthesis, lt	\n\
58003	-	rmoustache	:alias:	ISOAMSC rmoust	\n\
58003	-	rmoust	ISOamsc	right moustache	\n\
58004	-	lmoustache	:alias:	ISOAMSC lmoust	\n\
58004	-	lmoust	ISOamsc	left moustache	\n\
58005	-	ltrPar	ISOamsc	double right parenthesis, less	\n\
58006	-	gtlPar	ISOamsc	double left parenthesis, greater	\n\
58007	-	langd	ISOamsc	left angle, dot	\n\
58008	-	rangd	ISOamsc	right angle, dot	\n\
58009	-	lbrke	ISOamsc	left bracket, equal	\n\
58010	-	rbrke	ISOamsc	right bracket, equal	\n\
58011	-	lbrkslu	ISOamsc	left bracket, solidus top corner	\n\
58012	-	rbrksld	ISOamsc	right bracket, solidus bottom corner	\n\
58013	-	lbrksld	ISOamsc	left bracket, solidus bottom corner	\n\
58014	-	rbrkslu	ISOamsc	right bracket, solidus top corner	\n\
58015	-	gnap	ISOamsn	greater, not approximate	\n\
58015	-	gnapprox	:alias:	ISOamsn gnap	\n\
58016	-	-	-	-	\n\
58017	-	gvertneqq	:alias:	ISOamsn gvnE	\n\
58017	-	gvnE	ISOamsn	gt, vert, not double eq	\n\
58018	-	lnapprox	:alias:	ISOamsn lnap	\n\
58018	-	lnap	ISOamsn	less, not approximate	\n\
58019	-	-	-	-	\n\
58020	-	lvertneqq	:alias:	ISOamsn lvnE	\n\
58020	-	lvnE	ISOamsn	less, vert, not double eq	\n\
58022	-	NotGreaterEqual	:alias:	ISOamsn nge	\n\
58022	-	ngeq	:alias:	ISOamsn nge	\n\
58022	-	nge	ISOamsn	not greater-than-or-equal	\n\
58023	-	NotLessEqual	:alias:	ISOamsn nle	\n\
58023	-	nleq	:alias:	ISOamsn nle	\n\
58023	-	nle	ISOamsn	not less-than-or-equal	\n\
58026	-	nshortmid	:alias:	ISOamsn nsmid	\n\
58026	-	nsmid	ISOamsn	not short mid	\n\
58027	-	nshortparallel	:alias:	ISOamsn nspar	\n\
58027	-	nspar	ISOamsn	not short par	\n\
58035	-	precneqq	:alias:	ISOamsn prnE	\n\
58035	-	prnE	ISOamsn	precedes, not double eq	\n\
58037	-	succneqq	:alias:	ISOamsn scnE	\n\
58037	-	scnE	ISOamsn	succeeds, not double eq	\n\
58040	-	varsubsetneqq	:alias:	ISOamsn vsubnE	\n\
58040	-	vsubnE	ISOamsn	subset not double eq, var	\n\
58041	-	varsubsetneq	:alias:	ISOamsn vsubne	\n\
58041	-	vsubne	ISOamsn	subset, not eq, var	\n\
58042	-	varsupsetneq	:alias:	ISOamsn vsupne	\n\
58042	-	vsupne	ISOamsn	superset, not eq, var	\n\
58043	-	varsupsetneqq	:alias:	ISOamsn vsupnE	\n\
58043	-	vsupnE	ISOamsn	super not double eq, var	\n\
58044	-	napid	ISOamsn	not approximately identical to	\n\
58053	-	ncongdot	ISOamsn	not congruent, dot	\n\
58054	-	nvap	ISOamsn	not, vert, approximate	\n\
58055	-	napE	ISOamsn	not approximately equal or equal to	\n\
58056	-	parsim	ISOamsn	parallel, similar	\n\
58057	-	nLt	ISOamsn	not, vert, much less than	\n\
58058	-	nGt	ISOamsn	not, vert, much greater than	\n\
58059	-	NotLessLess	:alias:	ISOamsn nLtv	\n\
58059	-	nLtv	ISOamsn	not much less than, variant	\n\
58060	-	NotGreaterGreater	:alias:	ISOamsn nGtv	\n\
58060	-	nGtv	ISOamsn	not much greater than, variant	\n\
58061	-	nLl	ISOamsn	not triple less than	\n\
58062	-	nGg	ISOamsn	not triple greater than	\n\
58063	-	nvrtrie	ISOamsn	not, vert, right triangle, equals	\n\
58064	-	nvltrie	ISOamsn	not, vert, left triangle, equals	\n\
58065	-	rnmid	ISOamsn	reverse nmid	\n\
58066	-	-	-	-	\n\
58067	-	emptyset	:alias:	ISOamso empty	\n\
58067	-	empty	ISOamso	zero, slash	\n\
58068	-	jmath	ISOamso	small j, no dot	\n\
58069	-	hbar	:alias:	ISOamso plank	\n\
58069	-	plank	ISOamso	Planck's over 2pi	\n\
58070	-	ange	ISOamso	angle, equal	\n\
58071	-	range	ISOamso	reverse angle, equal	\n\
58072	-	nang	ISOamso	not, vert, angle	\n\
58073	-	angmsdaa	ISOamso	angle-measured, arrow, up, right	\n\
58074	-	angmsdab	ISOamso	angle-measured, arrow, up, left	\n\
58075	-	angmsdac	ISOamso	angle-measured, arrow, down, right	\n\
58076	-	angmsdad	ISOamso	angle-measured, arrow, down, left	\n\
58077	-	angmsdae	ISOamso	angle-measured, arrow, right, up	\n\
58078	-	angmsdaf	ISOamso	angle-measured, arrow, left, up	\n\
58079	-	angmsdag	ISOamso	angle-measured, arrow, right, down	\n\
58080	-	angmsdah	ISOamso	angle-measured, arrow, left, down	\n\
58081	-	angrtvbd	ISOamso	right angle-measured, dot	\n\
58082	-	urtri	ISOamso	upper right triangle	\n\
58083	-	lrtri	ISOamso	lower right triangle	\n\
58084	-	ultri	ISOamso	upper left triangle	\n\
58085	-	lltri	ISOamso	lower left triangle	\n\
58086	-	boxbox	ISOamso	two joined squares	\n\
58087	-	demptyv	ISOamso	circle, slash, bar above	\n\
58088	-	cemptyv	ISOamso	circle, slash, small circle above	\n\
58089	-	raemptyv	ISOamso	circle, slash, right arrow above	\n\
58090	-	laemptyv	ISOamso	circle, slash, left arrow above	\n\
58091	-	vzigzag	ISOamso	vertical zig-zag line	\n\
58092	-	trpezium	ISOamso	trapezium	\n\
58093	-	bsemi	ISOamso	reverse semi-colon	\n\
58094	-	bbrk	ISOamso	bottom square bracket	\n\
58095	-	tbrk	ISOamso	top square bracket	\n\
58102	-	GreaterSlantEqual	:alias:	ISOamsr ges	\n\
58102	-	geqslant	:alias:	ISOamsr ges	\n\
58102	-	ges	ISOamsr	gt-or-equal, slanted	\n\
58103	-	GreaterGreater	Lextra	alias for GT	\n\
58106	-	LessSlantEqual	:alias:	ISOamsr les	\n\
58106	-	leqslant	:alias:	ISOamsr les	\n\
58106	-	les	ISOamsr	less-than-or-eq, slant	\n\
58107	-	LessLess	Lextra	alias for Lt	\n\
58110	-	PrecedesEqual	:alias:	ISOamsr pre	\n\
58110	-	preceq	:alias:	ISOamsr pre	\n\
58110	-	prE	ISOamsr,:alias:	precedes, double equals	\n\
58110	-	pre	ISOamsr	precedes, equals	\n\
	\n\
#	E3	\n\
	\n\
58113	-	shortmid	:alias:	ISOamsr smid	\n\
58113	-	smid	ISOamsr	short mid	\n\
58114	-	shortparallel	:alias:	ISOamsr spar	\n\
58114	-	spar	ISOamsr	short parallel	\n\
58118	-	thickapprox	:alias:	ISOamsr thkap 	\n\
58118	-	thkap	ISOamsr	thick approximate	\n\
58121	-	ddotseq	:alias:	ISOamsr eDDot	\n\
58121	-	eDDot	ISOamsr	equal with four dots	\n\
58122	-	mlcp	ISOamsr	mlcp	\n\
58123	-	siml	ISOamsr	similar, less	\n\
58124	-	simg	ISOamsr	similar, greater	\n\
58125	-	Vbar	ISOamsr	double vert, bar (under)	\n\
58126	-	Colone	ISOamsr	double colon, equals	\n\
58127	-	DoubleLeftTee	:alias:	for Dashv 	\n\
58127	-	Dashv	ISOamsr	double dash, vertical	\n\
58128	-	vBar	ISOamsr	vert, double bar (under)	\n\
58129	-	Barv	ISOamsr	vert, double bar (over)	\n\
58130	-	vBarv	ISOamsr	double bar, vert over and under	\n\
58131	-	Vdashl	ISOamsr	vertical, dash (long)	\n\
58132	-	congdot	ISOamsr	congruent, dot	\n\
58133	-	-	-	-	\n\
58134	-	bumpE	ISOamsr	bump, equals	\n\
58135	-	Esim	ISOamsr	equal, similar	\n\
58136	-	equivDD	ISOamsr	equivalent, four dots above	\n\
58137	-	sdote	ISOamsr	equal, dot below	\n\
58138	-	mcomma	ISOamsr	minus, comma above	\n\
58139	-	forkv	ISOamsr	fork, variant	\n\
58140	-	topfork	ISOamsr	fork with top	\n\
58141	-	lesdot	ISOamsr	less-than-or-equal, slanted, dot inside	\n\
58142	-	gesdot	ISOamsr	greater-than-or-equal, slanted, dot inside	\n\
58143	-	lesdoto	ISOamsr	less-than-or-equal, slanted, dot above	\n\
58144	-	gesdoto	ISOamsr	greater-than-or-equal, slanted, dot above	\n\
58145	-	lesdotor	ISOamsr	less-than-or-equal, slanted, dot above right	\n\
58146	-	gesdotol	ISOamsr	greater-than-or-equal, slanted, dot above left	\n\
58147	-	elsdot	ISOamsr	equal-or-less, slanted, dot inside	\n\
58148	-	egsdot	ISOamsr	equal-or-greater, slanted, dot inside	\n\
58149	-	ltcir	ISOamsr	less than, circle inside	\n\
58150	-	gtcir	ISOamsr	greater than, circle inside	\n\
58151	-	el	ISOamsr	equal-or-less	\n\
58152	-	eg	ISOamsr	equal-or-greater	\n\
58153	-	ltquest	ISOamsr	less than, questionmark above	\n\
58154	-	gtquest	ISOamsr	greater than, questionmark above	\n\
58155	-	lesg	ISOamsr	less, equal, slanted, greater	\n\
58156	-	gesl	ISOamsr	greater, equal, slanted, less	\n\
58157	-	lgE	ISOamsr	less, greater, equal	\n\
58158	-	glE	ISOamsr	greater, less, equal	\n\
58159	-	glj	ISOamsr	greater, less, overlapping	\n\
58160	-	gla	ISOamsr	greater, less, apart	\n\
58161	-	lesges	ISOamsr	less, equal, slanted, greater, equal, slanted	\n\
58162	-	gesles	ISOamsr	greater, equal, slanted, less, equal, slanted	\n\
58163	-	lsime	ISOamsr	less, similar, equal	\n\
58164	-	gsime	ISOamsr	greater, similar, equal	\n\
58165	-	lsimg	ISOamsr	less, similar, greater	\n\
58166	-	gsiml	ISOamsr	greater, similar, less	\n\
58167	-	simlE	ISOamsr	similar, less, equal	\n\
58168	-	simgE	ISOamsr	similar, greater, equal	\n\
58169	-	smt	ISOamsr	smaller than	\n\
58170	-	lat	ISOamsr	larger than	\n\
58171	-	smte	ISOamsr	smaller than or equal	\n\
58172	-	late	ISOamsr	larger than or equal	\n\
58173	-	smtes	ISOamsr	smaller than or equal, slanted	\n\
58174	-	lates	ISOamsr	larger than or equal, slanted	\n\
58175	-	subrarr	ISOamsr	subset, right arrow	\n\
58176	-	suplarr	ISOamsr	superset, left arrow	\n\
58177	-	subplus	ISOamsr	subset, plus	\n\
58178	-	supplus	ISOamsr	superset, plus	\n\
58179	-	submult	ISOamsr	subset, multiply	\n\
58180	-	supmult	ISOamsr	superset, multiply	\n\
58181	-	subsim	ISOamsr	subset, similar	\n\
58182	-	supsim	ISOamsr	superset, similar	\n\
58183	-	subsup	ISOamsr	subset above superset	\n\
58184	-	supsub	ISOamsr	superset above subset	\n\
58185	-	subsub	ISOamsr	subset above subset	\n\
58186	-	supsup	ISOamsr	superset above superset	\n\
58187	-	suphsub	ISOamsr	superset, subset	\n\
58188	-	supdsub	ISOamsr	superset, subset, dash joining them	\n\
58189	-	bsolhsub	ISOamsr	reverse solidus, subset	\n\
58190	-	suphsol	ISOamsr	superset, solidus	\n\
58191	-	subedot	ISOamsr	subset, equals, dot	\n\
58192	-	supedot	ISOamsr	superset, equals, dot	\n\
58193	-	csub	ISOamsr	subset, closed	\n\
58194	-	csup	ISOamsr	superset, closed	\n\
58195	-	csube	ISOamsr	subset, closed, equals	\n\
58196	-	csupe	ISOamsr	superset, closed, equals	\n\
58197	-	ltcc	ISOamsr	less than, closed by curve	\n\
58198	-	gtcc	ISOamsr	greater than, closed by curve	\n\
58199	-	lescc	ISOamsr	less than, closed by curve, equal, slanted	\n\
58200	-	gescc	ISOamsr	greater than, closed by curve, equal, slanted	\n\
58201	-	rtriltri	ISOamsr	right triangle above left triangle	\n\
58202	-	-	-	-	\n\
58203	-	-	-	-	\n\
58204	-	Pr	ISOamsr	double precedes	\n\
58205	-	Sc	ISOamsr	double succeeds	\n\
58206	-	ltlarr	ISOamsr	less than, left arrow	\n\
58207	-	gtrarr	ISOamsr	greater than, right arrow	\n\
58213	-	iff	ISOtech	if and only if 	\n\
58222	-	andand	ISOtech	two logical and	\n\
58223	-	oror	ISOtech	two logical or	\n\
58224	-	notinva	ISOtech	negated set membership, variant	\n\
58225	-	qprime	ISOtech	quadruple prime	\n\
58226	-	iinfin	ISOtech	infinity sign, incomplete	\n\
58228	-	And	ISOtech	double logical and	\n\
58229	-	Or	ISOtech	double logical or	\n\
58230	-	pointint	ISOtech	integral around a point operator	\n\
58231	-	quatint	ISOtech	quaternion integral operator	\n\
58232	-	iiiint	:alias:	ISOtech qint	\n\
58232	-	qint	ISOtech	quadruple integral operator	\n\
58233	-	lopar	ISOtech	left open parenthesis	\n\
58234	-	ropar	ISOtech	right open parenthesis	\n\
58235	-	notinvb	ISOtech	negated set membership, variant	\n\
58236	-	notinvc	ISOtech	negated set membership, variant	\n\
58237	-	notnivb	ISOtech	contains, variant	\n\
58238	-	notnivc	ISOtech	contains, variant	\n\
58239	-	-	-	-	\n\
58240	-	strns	ISOtech	straightness	\n\
58241	-	fltns	ISOtech	flatness	\n\
58242	-	parsl	ISOtech	parallel, slanted	\n\
58243	-	topcir	ISOtech	top, circle below	\n\
58244	-	eparsl	ISOtech	parallel, slanted, equal; homothetically congruent to	\n\
58245	-	smeparsl	ISOtech	similar, parallel, slanted, equal	\n\
58246	-	eqvparsl	ISOtech	equivalent, equal; congruent and parallel	\n\
58247	-	bnequiv	ISOtech	reverse not equivalent	\n\
58248	-	bne	ISOtech	reverse not equal	\n\
58249	-	nparsl	ISOtech	not parallel, slanted	\n\
58250	-	nedot	ISOtech	not equal, dot	\n\
58251	-	simdot	ISOtech	similar, dot	\n\
58252	-	apacir	ISOtech	approximate, circumflex accent	\n\
58253	-	nhpar	ISOtech	not, horizontal, parallel	\n\
58254	-	nvinfin	ISOtech	not, vert, infinity	\n\
58255	-	-	-	-	\n\
58256	-	npart	ISOtech	not partial differential	\n\
58257	-	andv	ISOtech	and with middle stem	\n\
58258	-	orv	ISOtech	or with middle stem	\n\
58259	-	ord	ISOtech	or, horizontal dash	\n\
58260	-	andd	ISOtech	and, horizontal dash	\n\
58261	-	cirfnint	ISOtech	circulation function	\n\
58262	-	fpartint	ISOtech	finite part integral	\n\
58263	-	rppolint	ISOtech	line integration, rectangular path around pole	\n\
58264	-	scpolint	ISOtech	line integration, semi-circular path around pole	\n\
58265	-	npolint	ISOtech	line integration, not including the pole	\n\
58266	-	intlarhk	ISOtech	integral, left arrow with hook	\n\
58267	-	awint	ISOtech	anti clock-wise integration	\n\
58268	-	isindot	ISOtech	set membership, dot above	\n\
58269	-	notindot	ISOtech	negated set membership, dot above	\n\
58270	-	isinE	ISOtech	set membership, two horizontal strokes	\n\
58271	-	-	-	-	\n\
58272	-	disin	ISOtech	set membership, long horizontal stroke	\n\
58273	-	nisd	ISOtech	contains, long horizontal stroke	\n\
58274	-	isinsv	ISOtech	large set membership, vertical bar on horizontal stroke	\n\
58275	-	xnis	ISOtech	large contains, vertical bar on horizontal stroke	\n\
58276	-	isins	ISOtech	set membership, vertical bar on horizontal stroke	\n\
58277	-	nis	ISOtech	contains, vertical bar on horizontal stroke	\n\
58278	-	acd	ISOtech	ac current	\n\
58279	-	elinters	ISOtech	electrical intersection	\n\
58280	-	olcross	ISOtech	circle, cross	\n\
58281	-	dsol	ISOtech	solidus, bar above	\n\
58282	-	dwangle	ISOtech	large downward pointing angle	\n\
58283	-	uwangle	ISOtech	large upward pointing angle	\n\
58284	-	Not	ISOtech	not with two horizontal strokes	\n\
58285	-	bNot	ISOtech	reverse not with two horizontal strokes	\n\
58286	-	orslope	ISOtech	sloping large or	\n\
58290	-	fjlig	ISOpub	small fj ligature	\n\
	\n\
#	E4	\n\
	\n\
58368	-	cudarrr	ISOamsa	right, curved, down arrow	\n\
58370	-	lbarr	ISOamsa	left broken arrow	\n\
58373	-	bkarow	:alias:	ISOamsa rbarr	\n\
58373	-	rbarr	ISOamsa	right broken arrow	\n\
58377	-	olcir	ISOamsb	large circle in circle	\n\
58378	-	omid	ISOamsb	vertical bar in circle	\n\
58380	-	race	ISOamsb	reverse most positive, line below	\n\
58389	-	nvsim	ISOamsn	not, vert, similar	\n\
58390	-	solbar	ISOamsn	solidus, bar through	\n\
58392	-	angrtvb	ISOamso	right angle-measured	\n\
58393	-	bbrktbrk	ISOamso	bottom above top square bracket	\n\
58394	-	bemptyv	ISOamso	reversed circle, slash	\n\
58395	-	cirE	ISOamso	circle, two horizontal stroked to the right	\n\
58396	-	cirscir	ISOamso	circle, small circle to the right	\n\
58409	-	thicksim	:alias:	ISOamsr thksim	\n\
58409	-	thksim	ISOamsr	thick similar	\n\
	\n\
#	E5	\n\
	\n\
58624	-	Aopf	ISOmopf	Bbb A, open face A	\n\
58625	-	Bopf	ISOmopf	Bbb B, open face B	\n\
58627	-	Dopf	ISOmopf	Bbb D, open face D	\n\
58628	-	Eopf	ISOmopf	Bbb E, open face E	\n\
58629	-	Fopf	ISOmopf	Bbb F, open face F	\n\
58630	-	Gopf	ISOmopf	Bbb G, open face G	\n\
58631	-	Hopf	ISOmopf	Bbb H, open face H	\n\
58632	-	Iopf	ISOmopf	Bbb I, open face I	\n\
58633	-	Jopf	ISOmopf	Bbb J, open face J	\n\
58634	-	Kopf	ISOmopf,:alias:	Bbb K, open face K 	\n\
58634	-	andslope	ISOtech	sloping large and	\n\
58635	-	Lopf	ISOmopf,:alias:	Bbb L, open face L 	\n\
58635	-	imped	ISOtech	impedance	\n\
58636	-	Mopf	ISOmopf,:alias:	Bbb M, open face M 	\n\
58636	-	infintie	ISOtech	tie, infinity	\n\
58637	-	notinE	ISOtech	negated set membership, two horizontal strokes	\n\
58638	-	Oopf	ISOmopf	Bbb O, open face O	\n\
58642	-	Sopf	ISOmopf	Bbb S, open face S	\n\
58643	-	Topf	ISOmopf	Bbb T, open face T	\n\
58644	-	Uopf	ISOmopf	Bbb U, open face U	\n\
58645	-	Vopf	ISOmopf	Bbb V, open face V	\n\
58646	-	Wopf	ISOmopf	Bbb W, open face W	\n\
58647	-	Xopf	ISOmopf	Bbb X, open face X	\n\
58648	-	Yopf	ISOmopf	Bbb Y, open face Y	\n\
58656	-	Ascr	ISOmscr	script letter A	\n\
58657	-	-	-	-	\n\
58658	-	Cscr	ISOmscr	script letter C	\n\
58659	-	Dscr	ISOmscr	script letter D	\n\
58662	-	Gscr	ISOmscr	script letter G	\n\
58665	-	Jscr	ISOmscr	script letter J	\n\
58666	-	Kscr	ISOmscr	script letter K	\n\
58669	-	Nscr	ISOmscr	script letter N	\n\
58670	-	Oscr	ISOmscr	script letter O	\n\
58671	-	Pscr	ISOmscr	script letter P	\n\
58672	-	Qscr	ISOmscr	script letter Q	\n\
58673	-	Rscr	ISOmscr	script letter R	\n\
58674	-	Sscr	ISOmscr	script letter S	\n\
58675	-	Tscr	ISOmscr	script letter T	\n\
58676	-	Uscr	ISOmscr	script letter U	\n\
58677	-	Vscr	ISOmscr	script letter V	\n\
58678	-	Wscr	ISOmscr	script letter W	\n\
58679	-	Xscr	ISOmscr	script letter X	\n\
58680	-	Yscr	ISOmscr	script letter Y	\n\
58681	-	Zscr	ISOmscr	script letter Z	\n\
58688	-	ascr	ISOmscr	script letter a	\n\
58689	-	bscr	ISOmscr	script letter b	\n\
58690	-	cscr	ISOmscr	script letter c	\n\
58691	-	dscr	ISOmscr	script letter d	\n\
58693	-	fscr	ISOmscr	script letter f	\n\
58694	-	gscr	ISOmscr	script letter g	\n\
58695	-	hscr	ISOmscr	script letter h	\n\
58696	-	iscr	ISOmscr	script letter i	\n\
58697	-	jscr	ISOmscr	script letter j	\n\
58698	-	kscr	ISOmscr	script letter k	\n\
58699	-	lscr	ISOmscr	script letter l	\n\
58700	-	mscr	ISOmscr	script letter m	\n\
58701	-	nscr	ISOmscr	script letter n	\n\
58703	-	pscr	ISOmscr	script letter p	\n\
58704	-	qscr	ISOmscr	script letter q	\n\
58705	-	rscr	ISOmscr	script letter r	\n\
58706	-	sscr	ISOmscr	script letter s	\n\
58707	-	tscr	ISOmscr	script letter t	\n\
58708	-	uscr	ISOmscr	script letter u	\n\
58709	-	vscr	ISOmscr	script letter v	\n\
58710	-	wscr	ISOmscr	script letter w	\n\
58711	-	xscr	ISOmscr	script letter x	\n\
58712	-	yscr	ISOmscr	script letter y	\n\
58713	-	zscr	ISOmscr	script letter z	\n\
58720	-	Afr	ISOmfrk	/frak A, upper case a	\n\
58721	-	Bfr	ISOmfrk	/frak B, upper case b	\n\
58722	-	Cfr	ISOmfrk	/frak C, upper case c	\n\
58723	-	Dfr	ISOmfrk	/frak D, upper case d	\n\
58724	-	Efr	ISOmfrk	/frak E, upper case e	\n\
58725	-	Ffr	ISOmfrk	/frak F, upper case f	\n\
58726	-	Gfr	ISOmfrk	/frak G, upper case g	\n\
58729	-	Jfr	ISOmfrk	/frak J, upper case j	\n\
58730	-	Kfr	ISOmfrk	/frak K, upper case k	\n\
58731	-	Lfr	ISOmfrk	/frak L, upper case l	\n\
58732	-	Mfr	ISOmfrk	/frak M, upper case m	\n\
58733	-	Nfr	ISOmfrk	/frak N, upper case n	\n\
58734	-	Ofr	ISOmfrk	/frak O, upper case o	\n\
58735	-	Pfr	ISOmfrk	/frak P, upper case p	\n\
58736	-	Qfr	ISOmfrk	/frak Q, upper case q	\n\
58738	-	Sfr	ISOmfrk	/frak S, upper case s	\n\
58739	-	Tfr	ISOmfrk	/frak T, upper case t	\n\
58740	-	Ufr	ISOmfrk	/frak U, upper case u	\n\
58741	-	Vfr	ISOmfrk	/frak V, upper case v	\n\
58742	-	Wfr	ISOmfrk	/frak W, upper case w	\n\
58743	-	Xfr	ISOmfrk	/frak X, upper case x	\n\
58744	-	Yfr	ISOmfrk	/frak Y, upper case y	\n\
58745	-	Zfr	ISOmfrk	/frak Z, upper case z 	\n\
58752	-	afr	ISOmfrk	/frak a, lower case a	\n\
58753	-	bfr	ISOmfrk	/frak b, lower case b	\n\
58754	-	cfr	ISOmfrk	/frak c, lower case c	\n\
58755	-	dfr	ISOmfrk	/frak d, lower case d	\n\
58756	-	efr	ISOmfrk	/frak e, lower case e	\n\
58757	-	ffr	ISOmfrk	/frak f, lower case f	\n\
58758	-	gfr	ISOmfrk	/frak g, lower case g	\n\
58759	-	hfr	ISOmfrk	/frak h, lower case h	\n\
58760	-	ifr	ISOmfrk	/frak i, lower case i	\n\
58761	-	jfr	ISOmfrk	/frak j, lower case j	\n\
58762	-	kfr	ISOmfrk	/frak k, lower case k	\n\
58763	-	lfr	ISOmfrk	/frak l, lower case l	\n\
58764	-	mfr	ISOmfrk	/frak m, lower case m	\n\
58765	-	nfr	ISOmfrk	/frak n, lower case n	\n\
58766	-	ofr	ISOmfrk	/frak o, lower case o	\n\
58767	-	pfr	ISOmfrk	/frak p, lower case p	\n\
58768	-	qfr	ISOmfrk	/frak q, lower case q	\n\
58769	-	rfr	ISOmfrk	/frak r, lower case r	\n\
58770	-	sfr	ISOmfrk	/frak s, lower case s	\n\
58771	-	tfr	ISOmfrk	/frak t, lower case t	\n\
58772	-	ufr	ISOmfrk	/frak u, lower case u	\n\
58773	-	vfr	ISOmfrk	/frak v, lower case v	\n\
58774	-	wfr	ISOmfrk	/frak w, lower case w	\n\
58775	-	xfr	ISOmfrk	/frak x, lower case x	\n\
58776	-	yfr	ISOmfrk	/frak y, lower case y	\n\
58777	-	zfr	ISOmfrk	/frak z, lower case z	\n\
58844	-	NotPrecedesEqual	:alias:	ISOamsn npre	\n\
58844	-	npreceq	:alias:	ISOamsn npre	\n\
58844	-	npre	ISOamsn	not precedes, equals	\n\
58865	-	NotSucceedsEqual	:alias:	ISOamsn nsce	\n\
58865	-	nsucceq	:alias:	ISOamsn nsce	\n\
58865	-	nsce	ISOamsn	not succeeds, equals	\n\
	\n\
#	E6	\n\
	\n\
58884	-	NotSquareSubset	Lextra	square not subset	\n\
58901	-	NotSquareSuperset	Lextra	negated set-like partial order operator	\n\
58902	-	NotHumpDownHump	Lextra,:alias:	mmlextra nbump	\n\
58902	-	nbump	Lextra	not bumpy equals	\n\
58905	-	Cross	Lextra	cross or vector product	\n\
	\n\
#	E7	\n\
	\n\
59136	-	b.alpha	ISOgrk4	small alpha, Greek	\n\
59137	-	b.beta	ISOgrk4	small beta, Greek	\n\
59138	-	b.gamma	ISOgrk4	small gamma, Greek	\n\
59139	-	b.Gamma	ISOgrk4	capital gamma, Greek	\n\
59140	-	b.gammad	ISOgrk4	digamma, Greek	\n\
59141	-	b.Gammad	ISOgrk4	capital digamma, Greek	\n\
59142	-	b.delta	ISOgrk4	small delta, Greek	\n\
59143	-	b.Delta	ISOgrk4	capital delta, Greek	\n\
59144	-	b.epsi	ISOgrk4	small epsilon, Greek	\n\
59146	-	b.epsiv	ISOgrk4	varepsilion	\n\
59147	-	b.zeta	ISOgrk4	small zeta, Greek	\n\
59148	-	b.eta	ISOgrk4	small eta, Greek	\n\
59149	-	b.thetas	ISOgrk4	straight theta, Greek	\n\
59150	-	b.Theta	ISOgrk4	capital theta, Greek	\n\
59151	-	b.thetav	ISOgrk4	var theta, Greek	\n\
59152	-	b.iota	ISOgrk4	small iota, Greek	\n\
59153	-	b.kappa	ISOgrk4	small kappa, Greek	\n\
59154	-	b.kappav	ISOgrk4	var kappa, Greek	\n\
59155	-	b.lambda	ISOgrk4	small lambda, Greek	\n\
59156	-	b.Lambda	ISOgrk4	capital lambda, Greek	\n\
59157	-	b.mu	ISOgrk4	small mu, Greek	\n\
59158	-	b.nu	ISOgrk4	small nu, Greek	\n\
59159	-	b.xi	ISOgrk4	small xi, Greek	\n\
59160	-	b.Xi	ISOgrk4	capital xi, Greek	\n\
59161	-	b.pi	ISOgrk4	small pi, Greek	\n\
59162	-	b.Pi	ISOgrk4	capital pi, Greek	\n\
59163	-	b.piv	ISOgrk4	varpi	\n\
59164	-	b.rho	ISOgrk4	small rho, Greek	\n\
59165	-	b.rhov	ISOgrk4	varrho	\n\
59166	-	b.sigma	ISOgrk4	small sigma, Greek	\n\
59167	-	b.Sigma	ISOgrk4	capital sigma, Greek	\n\
59168	-	b.sigmav	ISOgrk4	varsigma	\n\
59169	-	b.tau	ISOgrk4	small tau, Greek	\n\
59170	-	b.upsi	ISOgrk4	small upsilon, Greek	\n\
59171	-	b.Upsilon	ISOgrk4	capital upsilon, Greek	\n\
59172	-	b.phis	ISOgrk4	straight phi, Greek	\n\
59173	-	b.Phi	ISOgrk4	capital Phi, Greek	\n\
59174	-	b.phiv	ISOgrk4	varphi	\n\
59175	-	b.chi	ISOgrk4	small chi, Greek	\n\
59176	-	b.psi	ISOgrk4	small psi, Greek	\n\
59177	-	b.Psi	ISOgrk4	capital Psi, Greek	\n\
59178	-	b.omega	ISOgrk4	small omega, Greek	\n\
59179	-	b.Omega	ISOgrk4	capital Omega, Greek	\n\
	\n\
#	E8	\n\
	\n\
59392	-	backepsilon	:alias:	ISOamsr bepsi	\n\
59392	-	bepsi	ISOamsr	backepsilon; such that	\n\
59403	-	blacklozenge	:alias:	ISOpub lozf	\n\
59403	-	lozf	ISOpub	black lozenge; lozenge, filled	\n\
59447	-	NotSucceedsTilde	Lextra	not succeeds or similar	\n\
59448	-	ovbar	ISOamsb	circle with vertical bar	\n\
59460	-	smallsetminus	:alias:	ISOamsb ssetmn	\n\
59460	-	ssetmn	ISOamsb	small set-minus; set-minus reverse solidus	\n\
59469	-	NotHumpEqual	Lextra,:alias:	mmlextra nbumpe	\n\
59469	-	nbumpe	Lextra	not bumpy single equals	\n\
59470	-	NotEqualTilde	Lextra,:alias:	mmlextra nesim	\n\
59470	-	nesim	Lextra	not equal or similar	\n\
59472	-	LeftSkeleton	Lextra	start of missing information	\n\
59473	-	RightSkeleton	Lextra	end of missing information	\n\
59481	-	HorizontalLine	Lextra	short horizontal line 	\n\
59482	-	VerticalLine	Lextra	short vertical line 	\n\
59483	-	Assign	Lextra	assignment operator	\n\
59484	-	VerticalSeparator	Lextra	vertical separating operator	\n\
59519	-	ShortDownArrow	Lextra	short down arrow	\n\
59520	-	ShortUpArrow	Lextra	sort up arrow 	\n\
59537	-	IndentingNewLine	Lextra	force a line break and indent appropriately on next line	\n\
59538	-	NoBreak	Lextra	never break line here	\n\
59539	-	GoodBreak	Lextra	if a linebreak is needed, here is a good spot	\n\
59540	-	BadBreak	Lextra	if a linebreak is needed, try to avoid breaking here	\n\
59543	-	ThickSpace	Lextra	space of width 5/18 em	\n\
59544	-	NegativeVeryThinSpace	Lextra	space of width -1/18 em	\n\
59545	-	NegativeThinSpace	Lextra	space of width -3/18 em	\n\
59546	-	NegativeMediumSpace	Lextra	space of width -4/18 em	\n\
59547	-	NegativeThickSpace	Lextra	space of width -5/18 em	\n\
59548	-	InvisibleComma	Lextra	used as a separator, e.g., in indices	\n\
59548	-	ic	Lextra,:alias:	short form of InvisibleComma	\n\
59550	-	it	Lextra,:alias:	short form of InvisibleTimes	\n\
59550	-	InvisibleTimes	Lextra	marks multiplication when it is understood without a mark	\n\
59552	-	af	Lextra,:alias:	short form of ApplyFunction	\n\
59552	-	ApplyFunction	Lextra	character showing function application in presentation tagging	\n\
59559	-	false	Lextra	logical constant false	\n\
59562	-	NotANumber	Lextra	used in MML 4.3.2.9	\n\
59563	-	true	Lextra	logical constant true	\n\
	\n\
#	F4	\n\
	\n\
62480	-	LeftTriangleBar	Lextra	not left triangle, vertical bar	\n\
62481	-	RightTriangleBar	Lextra	vertical bar, right triangle	\n\
62482	-	NotLeftTriangleBar	Lextra	not left triangle, vertical bar	\n\
62483	-	NotRightTriangleBar	Lextra	not vertical bar, right triangle	\n\
62499	-	NotNestedLessLess	Lextra	not double less-than sign	\n\
62504	-	NotNestedGreaterGreater	Lextra	not double greater-than sign	\n\
62513	-	Equal	Lextra	two consecutive equal signs	\n\
	\n\
#	F5	\n\
	\n\
62723	-	UpArrowBar	Lextra	up arrow to bar	\n\
62724	-	DownArrowBar	Lextra	down arrow to bar	\n\
62725	-	LeftRightVector	Lextra	left-up-right-up harpoon	\n\
62727	-	LeftVectorBar	Lextra	left-up harpoon to bar	\n\
62728	-	RightVectorBar	Lextra	up-right harpoon to bar	\n\
62729	-	LeftTeeVector	Lextra	left-up harpoon from bar	\n\
62730	-	RightTeeVector	Lextra	right-up harpoon from bar	\n\
62731	-	DownLeftRightVector	Lextra	left-down-right-down harpoon	\n\
62732	-	DownLeftVectorBar	Lextra	left-down harpoon to bar	\n\
62733	-	DownRightVectorBar	Lextra	right-down harpoon to bar	\n\
62734	-	DownLeftTeeVector	Lextra	left-down harpoon from bar	\n\
62735	-	DownRightTeeVector	Lextra	right-down harpoon from bar	\n\
62736	-	RightUpDownVector	Lextra	up-right-down-right harpoon	\n\
62737	-	RightUpVectorBar	Lextra	up-right harpoon to bar	\n\
62738	-	RightDownVectorBar	Lextra	down-right harpoon to bar	\n\
62739	-	RightUpTeeVector	Lextra	up-right harpoon from bar	\n\
62740	-	RightDownTeeVector	Lextra	down-right harpoon from bar	\n\
62741	-	LeftUpDownVector	Lextra	up-left-down-left harpoon	\n\
62742	-	LeftUpVectorBar	Lextra	up-left harpoon to bar	\n\
62743	-	LeftDownVectorBar	Lextra	down-left harpoon to bar	\n\
62744	-	LeftUpTeeVector	Lextra	up-left harpoon from bar	\n\
62745	-	LeftDownTeeVector	Lextra	down-left harpoon from bar	\n\
62751	-	RuleDelayed	Lextra	rule-delayed (colon right arrow)	\n\
62756	-	RoundImplies	Lextra	round implies	\n\
62759	-	EmptySmallSquare	Lextra	empty small square	\n\
62760	-	FilledSmallSquare	Lextra	filled small square	\n\
62761	-	FilledVerySmallSquare	Lextra	filled very small square	\n\
62768	-	EmptyVerySmallSquare	Lextra	empty small square	\n\
	\n\
#	F6	\n\
	\n\
62979	-	LeftBracketingBar	Lextra	left vertical delimiter	\n\
62980	-	RightBracketingBar	Lextra	right vertical delimiter	\n\
62981	-	LeftDoubleBracketingBar	Lextra	left double vertical delimiter	\n\
62982	-	RightDoubleBracketingBar	Lextra	right double vertical delimiter	\n\
62992	-	OverParenthesis	Lextra	over parenthesis	\n\
62993	-	UnderParenthesis	Lextra	under parenthesis	\n\
62994	-	OverBrace	Lextra	over brace 	\n\
62995	-	UnderBrace	Lextra	under brace 	\n\
62996	-	OverBracket	Lextra	over bracket	\n\
62997	-	UnderBracket	Lextra	under bracket	\n\
	\n\
#	F7	\n\
	\n\
63307	-	DD	Lextra,:alias:	short form of CapitalDifferentialD	\n\
63307	-	CapitalDifferentialD	Lextra	D for use in differentials, e.g., within integrals	\n\
63308	-	dd	Lextra,:alias:	short form of DifferentialD	\n\
63308	-	DifferentialD	Lextra	d for use in differentials, e.g., within integrals	\n\
63309	-	ee	Lextra,:alias:	short form of ExponentialE	\n\
63309	-	ExponentialE	Lextra	e use for the exponential base of the natural logarithms	\n\
63310	-	ii	Lextra,:alias:	short form of ImaginaryI	\n\
63310	-	ImaginaryI	Lextra	i for use as a square root of -1	\n\
	\n\
#	FB	\n\
	\n\
64256	-	fflig	ISOpub	small ff ligature	\n\
64257	-	filig	ISOpub	small fi ligature	\n\
64258	-	fllig	ISOpub	small fl ligature	\n\
64259	-	ffilig	ISOpub	small ffi ligature	\n\
64260	-	ffllig	ISOpub	small ffl ligature	\n\
	\n\
#	BOM (private definitions:)	\n\
	\n\
65279	-	bom	Lextra	byte order mark	\n\
65279	-	bomBE	Lextra,:alias:	byte order mark, big endian	\n\
65534	-	zwnbsp	Lextra	zero width non-breaking space	\n\
65534	-	bomLE	Lextra,:alias:	byte order mark, little endian	\n\
#------+-------+---------------+---------------+----------      \n\
# Int           standard        Entity          Group           Description	\n\
#------+-------+---------------+---------------+----------      \n\
	\n\
"; // _charsMapAll}

// {_charsMapSpace
/* all Unicode space characters
 * Note: this table must be in sync with _charsMapAll above
 * aliases should be used first, then the one to be displayed
 * (see EnDeGUI.js)
 */
var _charsMapSpace = "	\n\
	\n\
#------+----------------------------------	\n\
# Int           Description	\n\
#------+----------------------------------	\n\
160	no break (required) space	\n\
5760	ogham space mark	\n\
6158	mongolian vowel separator	\n\
8192	en space (1/2-em)	\n\
8193	em space	\n\
8194	en space (1/2-em)	\n\
8195	em space	\n\
8196	1/3-em space	\n\
8197	1/4-em space	\n\
8198	1/6-em space	\n\
8199	figure space	\n\
8199	tabular width space	\n\
8199	digit space (width of a number)	\n\
8200	punctuation space (width of comma))	\n\
8201	thin space (1/6-em)	\n\
8201	space of width 3/18 em (ISOpub thinsp	\n\
8202	hair space	\n\
8202	space of width 1/18 em (ISOpub hairsp)	\n\
8203	zero width space	\n\
8204	zero width non-joiner	\n\
8205	zero width joiner	\n\
8234	left to right embedding	\n\
8235	--unknown--	\n\
8236	--unknown--	\n\
8237	--unknown--	\n\
8238	right to left override	\n\
8239	narrow no break space	\n\
8287	4/18-em space (medium mathematical space)	\n\
8298	--unknown--	\n\
8299	--unknown--	\n\
12288	ideographic space (for CJK character)	\n\
59543	space of width 5/18 em	\n\
59544	space of width -1/18 em	\n\
59545	space of width -3/18 em	\n\
59546	space of width -4/18 em	\n\
59547	space of width -5/18 em	\n\
65279	byte order mark, big endian	\n\
65534	zero width non-breaking space	\n\
#------+----------------------------------	\n\
"; // _charsMapSpace

// {_uhwMapAll
/*             -- not used, just for documentation
var _uhwMapAll = "	\n\
	\n\
#------+-------+------------------------------------------	\n\
# Int	ASCII	\n\
#------+-------+------------------------------------------	\n\
# 65280	-undef-	\n\
65281	!	\n\
65282	\"	\n\
65283	#	\n\
65284	$	\n\
65285	%	\n\
65286	&	\n\
65287	'	\n\
65288	(	\n\
65289	)	\n\
65290	*	\n\
65291	+	\n\
65292	,	\n\
65293	-	\n\
65294	.	\n\
65295	/	\n\
65296	0	\n\
65297	1	\n\
65298	2	\n\
65299	3	\n\
65300	4	\n\
65301	5	\n\
65302	6	\n\
65303	7	\n\
65304	8	\n\
65305	9	\n\
65306	:	\n\
65307	;	\n\
65308	<	\n\
65309	=	\n\
65310	>	\n\
65311	?	\n\
65312	@	\n\
65313	A	\n\
65314	B	\n\
65315	C	\n\
65316	D	\n\
65317	E	\n\
65318	F	\n\
65319	G	\n\
65320	H	\n\
65321	I	\n\
65322	J	\n\
65323	K	\n\
65324	L	\n\
65325	M	\n\
65326	N	\n\
65327	O	\n\
65328	P	\n\
65329	Q	\n\
65330	R	\n\
65331	S	\n\
65332	T	\n\
65333	U	\n\
65334	V	\n\
65335	W	\n\
65336	X	\n\
65337	Y	\n\
65338	Z	\n\
65339	[	\n\
65340	\	\n\
65341	]	\n\
65342	^	\n\
65343	_	\n\
65344	`	\n\
65345	a	\n\
65346	b	\n\
65347	c	\n\
65348	d	\n\
65349	e	\n\
65350	f	\n\
65351	g	\n\
65352	h	\n\
65353	i	\n\
65354	j	\n\
65355	k	\n\
65356	l	\n\
65357	m	\n\
65358	n	\n\
65359	o	\n\
65360	p	\n\
65361	q	\n\
65362	r	\n\
65363	s	\n\
65364	t	\n\
65365	u	\n\
65366	v	\n\
65367	w	\n\
65368	x	\n\
65369	y	\n\
65370	z	\n\
65371	{	\n\
65372	|	\n\
65373	}	\n\
65374	~	\n\
# 65375	// ToDo: .. \n\
# 65534	// ToDo: .. \n\
# 65471	-undef-	\n\
# 65472	-undef-	\n\
# 65473	-undef-	\n\
# 65480	-undef-	\n\
# 65481	-undef-	\n\
# 65488	-undef-	\n\
# 65489	-undef-	\n\
# 65496	-undef-	\n\
# 65497	-undef-	\n\
# 65501	-undef-	\n\
# 65502	-undef-	\n\
# 65503	-undef-	\n\
# 65511	-undef-	\n\
# 65535	-undef-	\n\
#------+-------+------------------------------------------	\n\
	\n\
";
*/
// _uhwMapAll}

/*
 * about  EnDe.winMap[]  -- see CP-1252  above ...
 * the intMap column is the reference to above table
 * currently (26jun07) from: http://www.i18nguy.com/markup/ncrs.html#t128159
 *      Achtung: 8214 falsch, muss 8218 sein
#  See also: http://unicode.org/Public/MAPPINGS/VENDORS/MICSFT/WindowsBestFit/bestfit1252.txt
 */
// {_winMapAll
var _winMapAll = "	\n\
	\n\
#------+-------+-------+-------+--------------------------	\n\
# Int	intMap	Entity	Group	Description			\n\
#------+-------+-------+-------+--------------------------	\n\
0	0	#0	CP1252	small ae ligature	\n\
13	13	nl	CP1252	carriage return	\n\
32	32	space	CP1252	space	\n\
#32	32	sp	CP1252	space	\n\
126	126	-	CP1252	tilde	\n\
127	127	delete	CP1252	delete	\n\
128	8364	euro	CP1252	euro sign	\n\
129	129	-	CP1252	HOP	\n\
130	8218	sbquo	CP1252	left lower single quotation mark	\n\
131	402	fnof	CP1252	function of, florian	\n\
132	8222	bdquo	CP1252	left lower double quotation mark	\n\
#132	8222	ldquor	CP1252	left lower double quotation mark	\n\
133	8230	hellip	CP1252	ellipsis, three dots	\n\
#133	8230	ldots	CP1252	ellipsis, three dots	\n\
134	8224	dagger	CP1252	dagger	\n\
135	8225	Dagger	CP1252	double dagger	\n\
136	710	circ	CP1252	caret	\n\
137	8240	permil	CP1252	permil	\n\
138	352	Scaron	CP1252	capital S with caron	\n\
139	8249	lsaquo	CP1252	left arrow	\n\
140	338	OElig	CP1252	capital OE ligature	\n\
141	141	-	CP1252	RI	\n\
142	381	#x17d	CP1252	capital Z with caron	\n\
#142	381	Zcaron	CP1252	capital Z with caron	\n\
143	143	-	CP1252	SS3	\n\
144	144	-	CP1252	DCS	\n\
145	8216	rsquo	CP1252	right upper single quotation mark	\n\
#145	8216	rsquor	CP1252	right upper single quotation mark	\n\
146	8217	lsquo	CP1252	left upper single quotation mark	\n\
147	8220	rdquo	CP1252	right upper double quotation mark	\n\
148	8221	ldquo	CP1252	left upper double quotation mark	\n\
149	8226	bull	CP1252	big middle dot	\n\
150	8211	ndash	CP1252	dash	\n\
#150	8211	endash	CP1252	dash	\n\
151	8212	mdash	CP1252	double dash	\n\
#151	8212	emdash	CP1252	double dash	\n\
152	732	tilde	CP1252	tilde	\n\
153	8482	trade	CP1252	TM, trade mark	\n\
154	353	scaron	CP1252	small S with caron	\n\
155	8250	rsaquo	CP1252	right arrow	\n\
156	339	oelig	CP1252	small oe ligature	\n\
157	157	-	CP1252	OSC	\n\
158	382	#x17e	CP1252	small z with caron	\n\
#158	382	zcaron	CP1252	small z with caron	\n\
159	376	Yuml	CP1252	yen?	\n\
	\n\
# 160..255 are the same as in _charsMapAll	\n\
"; // _winMapAll}

// {_asciiMapAll
var _asciiMapAll = "	\n\
	\n\
#------+-------+------------------------------------------	\n\
# Int	ASCII	Description	\n\
#------+-------+------------------------------------------	\n\
0	NUL	Null	\n\
1	SOH	Start of Heading (CC)	\n\
2	STX	Start of Text (CC)	\n\
3	ETX	End of Text (CC)	\n\
4	EOT	End of Transmission (CC)	\n\
5	ENQ	Enquiry (CC)	\n\
6	ACK	Acknowledge (CC)	\n\
7	BEL	Bell	\n\
8	BS	Backspace (FE)	\n\
9	HT	Horizontal Tabulation (FE)	\n\
10	LF	Line Feed (FE)	\n\
11	VT	Vertical Tabulation (FE)	\n\
12	FF	Form Feed (FE)	\n\
13	CR	Carriage Return (FE)	\n\
14	SO	Shift Out	\n\
15	SI	Shift In	\n\
16	DLE	Data Link Escape (CC)	\n\
17	DC1	Device Control 1	\n\
18	DC2	Device Control 2	\n\
19	DC3	Device Control 3	\n\
20	DC4	Device Control 4	\n\
21	NAK	Negative Acknowledge (CC)	\n\
22	SYN	Synchronous Idle (CC)	\n\
23	ETB	End of Transmission Block (CC)	\n\
24	CAN	Cancel	\n\
25	EM	End of Medium	\n\
26	SUB	Substitute	\n\
27	ESC	Escape	\n\
28	FS	File Separator (IS)	\n\
29	GS	Group Separator (IS)	\n\
30	RS	Record Separator (IS)	\n\
31	US	Unit Separator (IS)	\n\
32	SP	Space	\n\
33	!	Exclamation Point	\n\
34	\"	Quotation Mark	\n\
35	#	Number Sign, Octothorp, 'pound'	\n\
36	$	Dollar Sign	\n\
37	%	Percent	\n\
38	&	Ampersand	\n\
39	'	Apostrophe, Prime	\n\
40	(	Left Parenthesis	\n\
41	)	Right Parenthesis	\n\
42	*	Asterisk, 'star'	\n\
43	+	Plus Sign	\n\
44	,	Comma	\n\
45	-	Hyphen, Minus Sign	\n\
46	.	Period, Decimal Point, dot	\n\
47	/	Slash, Virgule	\n\
48	0	0	\n\
49	1	1	\n\
50	2	2	\n\
51	3	3	\n\
52	4	4	\n\
53	5	5	\n\
54	6	6	\n\
55	7	7	\n\
56	8	8	\n\
57	9	9	\n\
58	:	Colon	\n\
59	;	Semicolon	\n\
60	<	Less-than Sign	\n\
61	=	Equal Sign	\n\
62	>	Greater-than Sign	\n\
63	?	Question Mark	\n\
64	@	At Sign	\n\
65	A	A	\n\
66	B	B	\n\
67	C	C	\n\
68	D	D	\n\
69	E	E	\n\
70	F	F	\n\
71	G	G	\n\
72	H	H	\n\
73	I	I	\n\
74	J	J	\n\
75	K	K	\n\
76	L	L	\n\
77	M	M	\n\
78	N	N	\n\
79	O	O	\n\
80	P	P	\n\
81	Q	Q	\n\
82	R	R	\n\
83	S	S	\n\
84	T	T	\n\
85	U	U	\n\
86	V	V	\n\
87	W	W	\n\
88	X	X	\n\
89	Y	Y	\n\
90	Z	Z	\n\
91	[	Opening Bracket	\n\
92	\	Reverse Slant	\n\
93	]	Closing Bracket	\n\
94	^	Circumflex, Caret	\n\
95	_	Underline, Underscore	\n\
96	`	Grave Accent	\n\
97	a	a	\n\
98	b	b	\n\
99	c	c	\n\
100	d	d	\n\
101	e	e	\n\
102	f	f	\n\
103	g	g	\n\
104	h	h	\n\
105	i	i	\n\
106	j	j	\n\
107	k	k	\n\
108	l	l	\n\
109	m	m	\n\
110	n	n	\n\
111	o	o	\n\
112	p	p	\n\
113	q	q	\n\
114	r	r	\n\
115	s	s	\n\
116	t	t	\n\
117	u	u	\n\
118	v	v	\n\
119	w	w	\n\
120	x	x	\n\
121	y	y	\n\
122	z	z	\n\
123	{	Opening Brace	\n\
124	|	Vertical Line	\n\
125	}	Closing Brace	\n\
126	~	Tilde	\n\
127	DEL	Delete	\n\
# 128	 	Reserved	\n\
# 129	 	Reserved	\n\
# 130	 	Reserved	\n\
# 131	 	Reserved	\n\
132	IND	Index (FE)	\n\
133	NEL	Next Line (FE)	\n\
134	SSA	Start of Selected Area	\n\
135	ESA	End of Selected Area	\n\
136	HTS	Horizontal Tabulation Set (FE)	\n\
137	HTJ	Horizontal Tabulation with Justification (FE)	\n\
138	VTS	Vertical Tabulation Set (FE)	\n\
139	PLD	Partial Line Down (FE)	\n\
140	PLU	Partial Line Up (FE)	\n\
141	RI	Reverse Index (FE)	\n\
142	SS2	Single Shift Two (1)	\n\
143	SS3	Single Shift Three (1)	\n\
144	DCS	Device Control String (2)	\n\
145	PU1	Private Use One	\n\
146	PU2	Private Use Two	\n\
147	STS	Set Transmit State	\n\
148	CCH	Cancel Character	\n\
149	MW	Message Waiting	\n\
150	SPA	Start of Protected Area	\n\
151	EPA	End of Protected Area	\n\
# 152	 	Reserved	\n\
# 153	 	Reserved	\n\
# 154	 	Reserved	\n\
155	CSI	Control Sequence Introducer (1)	\n\
156	ST	String Terminator (2)	\n\
157	OSC	Operating System Command (2)	\n\
158	PM	Privacy Message (2)	\n\
159	APC	Application Program Command (2)	\n\
# 160 - 255	 	 	\n\
#------+-------+------------------------------------------	\n\
	\n\
"; // _asciiMapAll}

// {_ebcdicMapAll
var _ebcdicMapAll = "	\n\
	\n\
#------+-------+------------------------------------------	\n\
# Int	EBCDIC	Description	\n\
#------+-------+------------------------------------------	\n\
0	NUL	Null	\n\
1	SOH	Start of Heading	\n\
2	STX	Start of Text	\n\
3	ETX	End of Text	\n\
4	PF	Punch Off	\n\
5	HT	Horizontal Tab	\n\
6	LC	Lower Case	\n\
7	DEL	Delete	\n\
# 8	  	 	\n\
# 9	  	 	\n\
10	SMM	Start of Manual Message	\n\
11	VT	Vertical Tab	\n\
12	FF	Form Feed	\n\
13	CR	Carriage Return	\n\
14	SO	Shift Out	\n\
15	SI	Shift In	\n\
16	DLE	Data Link Escape	\n\
17	DC1	Device Control 1	\n\
18	DC2	Device Control 2	\n\
19	TM	Tape Mark	\n\
20	RES	Restore	\n\
21	NL	New Line	\n\
22	BS	Backspace	\n\
23	IL	Idle	\n\
24	CAN	Cancel	\n\
25	EM	End of Medium	\n\
26	CC	Cursor Control	\n\
27	CU1	Customer Use 1	\n\
28	IFS	Interchange File Separator	\n\
29	IGS	Interchange Group Separator	\n\
30	IRS	Interchange Record Separator	\n\
31	IUS	Interchange Unit Separator	\n\
32	DS	Digit Select	\n\
33	SOS	Start of Significance	\n\
34	FS	Field Separator	\n\
# 35	  	 	\n\
36	BYP	Bypass	\n\
37	LF	Line Feed	\n\
38	ETB	End of Transmission Block	\n\
39	ESC	Escape	\n\
# 40	  	 	\n\
# 41	  	 	\n\
42	SM	Set Mode	\n\
43	CU2	Customer Use 2	\n\
# 44	  	 	\n\
45	ENQ	Enquiry	\n\
46	ACK	Acknowledge	\n\
47	BEL	Bell	\n\
# 48	  	 	\n\
# 49	  	 	\n\
50	SYN	Synchronous Idle	\n\
# 51	 	 	\n\
52	PN	Punch On	\n\
53	RS	Reader Stop	\n\
54	UC	Upper Case	\n\
55	EOT	End of Transmission	\n\
# 56	  	 	\n\
# 57	  	 	\n\
# 58	  	 	\n\
59	CU3	Customer Use 3	\n\
60	DC4	Device Control 4	\n\
61	NAK	Negative Acknowledge	\n\
# 62				\n\
63	SUB	Substitute	\n\
64	 	Space	\n\
# 65 - 73	  	 	\n\
74		Cent Sign	\n\
75	.	Period, Decimal Point	\n\
76	<	Less-than Sign	\n\
77	(	Left Parenthesis	\n\
78	+	Plus Sign	\n\
# 79	  	 	\n\
80	&	Ampersand	\n\
# 81 - 89	  	 	\n\
90	!	Exclamation Point	\n\
91	$	Dollar Sign	\n\
92	*	Asterisk	\n\
93	)	Right Parenthesis	\n\
94	;	Semicolon	\n\
95	^	Logical NOT	\n\
96	-	Hyphen, Minus Sign	\n\
97	/	Slash, Virgule	\n\
# 98 - 105	  	 	\n\
106	|	Logical OR	\n\
107	,	Comma	\n\
108	%	Percent	\n\
109	_	Underline, Underscore	\n\
110	>	Greater-than Sign	\n\
111	?	Question Mark	\n\
# 112 - 121	  	 	\n\
122	:	Colon	\n\
123	#	Number Sign, Octothorp, 'pound'	\n\
124	@	At Sign	\n\
125	'	Apostrophe, Prime	\n\
126	=	Equal Sign	\n\
127	\"	Quotation Mark	\n\
# 128	  	 	\n\
129	a	a	\n\
130	b	b	\n\
131	c	c	\n\
132	d	d	\n\
133	e	e	\n\
134	f	f	\n\
135	g	g	\n\
136	h	h	\n\
137	i	i	\n\
# 138 - 144	  	 	\n\
145	j	j	\n\
146	k	k	\n\
147	l	l	\n\
148	m	m	\n\
149	n	n	\n\
150	o	o	\n\
151	p	p	\n\
152	q	q	\n\
153	r	r	\n\
# 154 - 160	  	 	\n\
161	~ 	Tilde	\n\
162	s	s	\n\
163	t	t	\n\
164	u	u	\n\
165	v	v	\n\
166	w	w	\n\
167	x	x	\n\
168	y	y	\n\
169	z	z	\n\
# 170 - 184	  	 	\n\
185	`	Grave Accent	\n\
# 186 - 191	  	 	\n\
192	{	Opening Brace	\n\
193	A	A	\n\
194	B	B	\n\
195	C	C	\n\
196	D	D	\n\
197	E	E	\n\
198	F	F	\n\
199	G	G	\n\
200	H	H	\n\
201	I	I	\n\
# 202 - 207	   	 	\n\
208	}	Closing Brace	\n\
209	J	J	\n\
210	K	K	\n\
211	L	L	\n\
212	M	M	\n\
213	N	N	\n\
214	O	O	\n\
215	P	P	\n\
216	Q	Q	\n\
217	R	R	\n\
# 218 - 223	  	 	\n\
224	\\	Reverse Slant	\n\
# 225	  	 	\n\
226	S	S	\n\
227	T	T	\n\
228	U	U	\n\
229	V	V	\n\
230	W	W	\n\
231	X	X	\n\
232	Y	Y	\n\
233	Z	Z	\n\
# 234 - 239	  	 	\n\
240	0	0	\n\
241	1	1	\n\
242	2	2	\n\
243	3	3	\n\
244	4	4	\n\
245	5	5	\n\
246	6	6	\n\
247	7	7	\n\
248	8	8	\n\
249	9	9	\n\
# 250 - 254	   	 	\n\
255	APC	APC	\n\
#------+-------+------------------------------------------	\n\
	\n\
"; // _ebcdicMapAll}

// {_ebcdicMapUTF
var _ebcdicMapUTF = "	\n\
	\n\
#------+-------+------------------------------------------	\n\
# Int	EBCDIC	Description	\n\
#------+-------+------------------------------------------	\n\
62	PM	PM	\n\
74			\n\
79	|	Logical OR	\n\
106			\n\
121	`	Grave Accent	\n\
173	[	Opening Bracket	\n\
185			\n\
189	[	Closing Bracket	\n\
"; // _ebcdicMapUTF}

/*
 * XML entity characters
 */
EnDe.xmlMap[34]       = 'quot';
EnDe.xmlMap[38]       = 'amp';
EnDe.xmlMap[39]       = 'apos';
EnDe.xmlMap[60]       = 'lt';
EnDe.xmlMap[62]       = 'gt';

/*
 * Morse characters
 */
EnDe.sosMap['0']='_____';
EnDe.sosMap['1']='.____'; EnDe.sosMap['a']='._';   EnDe.sosMap['j']='.___'; EnDe.sosMap['s']='...';
EnDe.sosMap['2']='..___'; EnDe.sosMap['b']='_...'; EnDe.sosMap['k']='_._';  EnDe.sosMap['t']='_';
EnDe.sosMap['3']='...__'; EnDe.sosMap['c']='_._.'; EnDe.sosMap['l']='._..'; EnDe.sosMap['u']='.._';
EnDe.sosMap['4']='...._'; EnDe.sosMap['d']='_..';  EnDe.sosMap['m']='__';   EnDe.sosMap['v']='..._';
EnDe.sosMap['5']='.....'; EnDe.sosMap['e']='.';    EnDe.sosMap['n']='_.';   EnDe.sosMap['w']='.__';
EnDe.sosMap['6']='_....'; EnDe.sosMap['f']='.._.'; EnDe.sosMap['o']='___';  EnDe.sosMap['x']='_.._';
EnDe.sosMap['7']='__...'; EnDe.sosMap['g']='__.';  EnDe.sosMap['p']='.__.'; EnDe.sosMap['y']='_.__';
EnDe.sosMap['8']='___..'; EnDe.sosMap['h']='....'; EnDe.sosMap['q']='__._'; EnDe.sosMap['z']='__..';
EnDe.sosMap['9']='____.'; EnDe.sosMap['i']='..';   EnDe.sosMap['r']='._.';  // EnDe.sosMap[' ']=' ';
/*
 * extended Morse characters
 */
EnDe.sosMap['.']='._._._'; EnDe.sosMap[':']='___...'; EnDe.sosMap['=']='_..._';
EnDe.sosMap['-']='_...._'; EnDe.sosMap['@']='.__._.'; EnDe.sosMap['/']='_.._.';
EnDe.sosMap[',']='__..__'; EnDe.sosMap['?']='..__..';
EnDe.sosMap['"']='._.._.'; EnDe.sosMap["'"]='.____.';
EnDe.sosMap[')']='_.__._'; EnDe.sosMap['+']='._._.';
/* not reversable, see above */
EnDe.sosMap['(']='_.__._'; // same as ')'
	/* note that '(' is defined, but not ')', we use it the other way arround
	 * so that decoding returns '(' always 
	 * seems to be a hash problem in JavaScript
	 */

/*
 * Baudot characters
 */
EnDe.figsMap['00000'] = "";
EnDe.figsMap['00001'] = "3";
EnDe.figsMap['00010'] = "\n";
EnDe.figsMap['00011'] = "-";
EnDe.figsMap['00100'] = " ";
EnDe.figsMap['00101'] = "'";
EnDe.figsMap['00110'] = "8";
EnDe.figsMap['00111'] = "7";
EnDe.figsMap['01000'] = "\r";
EnDe.figsMap['01001'] = "$";
EnDe.figsMap['01010'] = "4";
EnDe.figsMap['01011'] = " ";
EnDe.figsMap['01100'] = ",";
EnDe.figsMap['01101'] = "!";
EnDe.figsMap['01110'] = ":";
EnDe.figsMap['01111'] = "(";
EnDe.figsMap['10000'] = "5";
EnDe.figsMap['10001'] = '"';
EnDe.figsMap['10010'] = ")";
EnDe.figsMap['10011'] = "2";
EnDe.figsMap['10100'] = "#";
EnDe.figsMap['10101'] = "6";
EnDe.figsMap['10110'] = "0";
EnDe.figsMap['10111'] = "1";
EnDe.figsMap['11000'] = "9";
EnDe.figsMap['11001'] = "?";
EnDe.figsMap['11010'] = "&";
EnDe.figsMap['11011'] = "FIGS";
EnDe.figsMap['11100'] = ".";
EnDe.figsMap['11101'] = "/";
EnDe.figsMap['11110'] = ";";
EnDe.figsMap['11111'] = "LTRS";
EnDe.ltrsMap['00000'] = "";
EnDe.ltrsMap['00001'] = "E";
EnDe.ltrsMap['00010'] = "\n";
EnDe.ltrsMap['00011'] = "A";
EnDe.ltrsMap['00100'] = " ";
EnDe.ltrsMap['00101'] = "S";
EnDe.ltrsMap['00110'] = "I";
EnDe.ltrsMap['00111'] = "U";
EnDe.ltrsMap['01000'] = "\r";
EnDe.ltrsMap['01001'] = "D";
EnDe.ltrsMap['01010'] = "R";
EnDe.ltrsMap['01011'] = "J";
EnDe.ltrsMap['01100'] = "N";
EnDe.ltrsMap['01101'] = "F";
EnDe.ltrsMap['01110'] = "C";
EnDe.ltrsMap['01111'] = "K";
EnDe.ltrsMap['10000'] = "T";
EnDe.ltrsMap['10001'] = "Z";
EnDe.ltrsMap['10010'] = "L";
EnDe.ltrsMap['10011'] = "W";
EnDe.ltrsMap['10100'] = "H";
EnDe.ltrsMap['10101'] = "Y";
EnDe.ltrsMap['10110'] = "P";
EnDe.ltrsMap['10111'] = "Q";
EnDe.ltrsMap['11000'] = "O";
EnDe.ltrsMap['11001'] = "B";
EnDe.ltrsMap['11010'] = "G";
EnDe.ltrsMap['11011'] = "FIGS";
EnDe.ltrsMap['11100'] = "M";
EnDe.ltrsMap['11101'] = "X";
EnDe.ltrsMap['11110'] = "V";
EnDe.ltrsMap['11111'] = "LTRS";

/*
 * Braille characters
 */
/*
	following Braille character: 1-3-4-5

		**
		 *
		*

	will be written as:

		'**\n *\n* '
		 14  25  36
 */
/* References
 * http://www.dotlessbraille.org/FACInfo.htm
 * http://www.brl.org/
 * http://de.wikipedia.org/wiki/Brailleschrift
 * http://en.wikipedia.org/wiki/Braille_ASCII
*/
/*
 * Standard (dotless) Braille characters
 */
EnDe.DbrMap['digit']=' *\n *\n**'; // used as prefix to digits same as '#'
EnDe.DbrMap[' ']='  \n* \n  ';
EnDe.DbrMap['1']='* \n  \n  ';
EnDe.DbrMap['2']='* \n* \n  ';
EnDe.DbrMap['3']='**\n  \n  ';
EnDe.DbrMap['4']='**\n *\n  ';
EnDe.DbrMap['5']='* \n *\n  ';
EnDe.DbrMap['6']='**\n* \n  ';
EnDe.DbrMap['7']='**\n**\n  ';
EnDe.DbrMap['8']='* \n**\n  ';
EnDe.DbrMap['9']=' *\n* \n  ';
EnDe.DbrMap['0']=' *\n**\n  ';
//
EnDe.DbrMap['a']='* \n  \n  ';
EnDe.DbrMap['b']='* \n* \n  ';
EnDe.DbrMap['c']='**\n  \n  ';
EnDe.DbrMap['d']='**\n *\n  ';
EnDe.DbrMap['e']='* \n *\n  ';
EnDe.DbrMap['f']='**\n* \n  ';
EnDe.DbrMap['g']='**\n**\n  ';
EnDe.DbrMap['h']='* \n**\n  ';
EnDe.DbrMap['i']=' *\n* \n  ';
EnDe.DbrMap['j']=' *\n**\n  ';
//
EnDe.DbrMap['k']='* \n  \n* ';
EnDe.DbrMap['l']='* \n* \n* ';
EnDe.DbrMap['m']='**\n  \n* ';
EnDe.DbrMap['n']='**\n *\n* ';
EnDe.DbrMap['o']='* \n *\n* ';
EnDe.DbrMap['p']='**\n* \n* ';
EnDe.DbrMap['q']='**\n**\n* ';
EnDe.DbrMap['r']='* \n**\n* ';
EnDe.DbrMap['s']=' *\n* \n* ';
EnDe.DbrMap['t']=' *\n**\n* ';
//
EnDe.DbrMap['u']='* \n  \n**';
EnDe.DbrMap['v']='* \n* \n**';
EnDe.DbrMap['w']=' *\n**\n *';
EnDe.DbrMap['x']='**\n  \n**';
EnDe.DbrMap['y']='**\n *\n**';
EnDe.DbrMap['z']='* \n *\n**';
EnDe.DbrMap['']=' *\n* \n**';
EnDe.DbrMap['st']=' *\n**\n**';
EnDe.DbrMap['au']='* \n  \n *';
EnDe.DbrMap['eu']='* \n* \n *';
//
EnDe.DbrMap[',']='  \n* \n  ';
EnDe.DbrMap[';']='  \n* \n* ';
EnDe.DbrMap[':']='  \n**\n  ';
EnDe.DbrMap['.']='  \n  \n* ';
EnDe.DbrMap['?']='  \n* \n *';
EnDe.DbrMap['!']='  \n**\n* ';
EnDe.DbrMap['(']='  \n**\n**';
EnDe.DbrMap[')']='  \n**\n**';
EnDe.DbrMap['"']='  \n* \n**';
EnDe.DbrMap['"']='  \n *\n**';
EnDe.DbrMap['-']='  \n  \n**';
/*
 * ASCII Braille characters
 */
EnDe.AbrMap     = EnDe.DbrMap;
EnDe.AbrMap[' ']='  \n  \n *';
//
EnDe.AbrMap[',']='  \n  \n *';
EnDe.AbrMap[';']='  \n *\n *';
EnDe.AbrMap[':']='* \n *\n *';
EnDe.AbrMap['.']=' *\n  \n *';
EnDe.AbrMap['?']='**\n *\n *';
EnDe.AbrMap['!']=' *\n* \n**';
EnDe.AbrMap['(']='* \n**\n**';
EnDe.AbrMap[')']=' *\n**\n**';
EnDe.AbrMap['"']='  \n *\n  ';
EnDe.AbrMap['"']='  \n  \n *';
EnDe.AbrMap['-']='  \n  \n**';
//
EnDe.AbrMap['#']=' *\n *\n**';
EnDe.AbrMap['$']='**\n* \n *';
EnDe.AbrMap['%']='**\n  \n *';
EnDe.AbrMap['&']='**\n* \n**';
EnDe.AbrMap["'"]='  \n  \n* ';
EnDe.AbrMap['*']='* \n  \n *';
EnDe.AbrMap['+']=' *\n  \n**';
EnDe.AbrMap['/']=' *\n  \n* ';
EnDe.AbrMap['<']='* \n* \n *';
EnDe.AbrMap['=']='**\n**\n**';
EnDe.AbrMap['>']=' *\n *\n* ';
EnDe.AbrMap['@']=' *\n  \n  ';
EnDe.AbrMap['[']=' *\n* \n *';
EnDe.AbrMap['\\']='* \n**\n *';
EnDe.AbrMap[']']='**\n**\n *';
EnDe.AbrMap['^']=' *\n *\n  ';
EnDe.AbrMap['_']=' *\n *\n* ';

/*
 * Dada Urka  http://www.leiber.ws/dirk/dadaurka.htm
 */
EnDe.DadMap['a']='|   \n|   \n+---';
EnDe.DadMap['b']='|   \n|*  \n+---';
EnDe.DadMap['c']='|   \n|** \n+---';
EnDe.DadMap['d']='   |\n   |\n---+';
EnDe.DadMap['e']='   |\n  *|\n---+';
EnDe.DadMap['f']='   |\n **|\n---+';
EnDe.DadMap['g']='---+\n   |\n   |';
EnDe.DadMap['h']='---+\n  *|\n   |';
EnDe.DadMap['i']='---+\n **|\n   |';
EnDe.DadMap['j']='+---\n|   \n|   ';
EnDe.DadMap['k']='+---\n|*  \n|   ';
EnDe.DadMap['l']='+---\n|**|\n|   ';
EnDe.DadMap['m']='|  |\n|  |\n+--+';
EnDe.DadMap['n']='|  |\n|* |\n+--+';
EnDe.DadMap['o']='|  |\n|**|\n+--+';
EnDe.DadMap['p']='---+\n   |\n---+';
EnDe.DadMap['q']='---+\n  *|\n---+';
EnDe.DadMap['r']='---+\n **|\n---+';
EnDe.DadMap['s']='+--+\n|  |\n|  |';
EnDe.DadMap['t']='+--+\n|* |\n|  |';
EnDe.DadMap['u']='+--+\n|**|\n|  |';
EnDe.DadMap['v']='+---\n|*  \n+---';
EnDe.DadMap['w']='+--+\n|  |\n+--+';
EnDe.DadMap['x']='+--+\n|* |\n+--+';
EnDe.DadMap['y']='+--+\n|**|\n+--+';
EnDe.DadMap['z']='+---\n|   \n+---';
EnDe.DadMap['A']='|   \n|   \n+---';
EnDe.DadMap['B']='|   \n|*  \n+---';
EnDe.DadMap['C']='|   \n|   \n+---';
EnDe.DadMap['D']='   |\n   |\n---+';
EnDe.DadMap['E']='   |\n  *|\n---+';
EnDe.DadMap['F']='   |\n **|\n---+';
EnDe.DadMap['G']='---+\n   |\n   |';
EnDe.DadMap['H']='---+\n  *|\n   |';
EnDe.DadMap['I']='---+\n **|\n   |';
EnDe.DadMap['J']='+---\n|   \n|   ';
EnDe.DadMap['K']='+---\n|*  \n|   ';
EnDe.DadMap['L']='+---\n|**|\n|   ';
EnDe.DadMap['M']='|  |\n|  |\n+--+';
EnDe.DadMap['N']='|  |\n|* |\n+--+';
EnDe.DadMap['O']='|  |\n|**|\n+--+';
EnDe.DadMap['P']='---+\n   |\n---+';
EnDe.DadMap['Q']='---+\n  *|\n---+';
EnDe.DadMap['R']='---+\n **|\n---+';
EnDe.DadMap['S']='+--+\n|  |\n|  |';
EnDe.DadMap['T']='+--+\n|* |\n|  |';
EnDe.DadMap['U']='+--+\n|**|\n|  |';
EnDe.DadMap['V']='+---\n|*  \n+---';
EnDe.DadMap['W']='+--+\n|  |\n+--+';
EnDe.DadMap['X']='+--+\n|* |\n+--+';
EnDe.DadMap['Y']='+--+\n|**|\n+--+';
EnDe.DadMap['Z']='+---\n|   \n+---';

/*
 * Unicode code point ranges, see http://www.unicode.org/Public/UNIDATA/Blocks.txt
 */
EnDe.rangeMap['0x000000-0x00007F'] = 'Basic Latin';
EnDe.rangeMap['0x000080-0x0000FF'] = 'C1 Controls and Latin-1 Supplement';
EnDe.rangeMap['0x000100-0x00017F'] = 'Latin Extended-A';
EnDe.rangeMap['0x000180-0x00024F'] = 'Latin Extended-B';
EnDe.rangeMap['0x000250-0x0002AF'] = 'IPA Extensions';
EnDe.rangeMap['0x0002B0-0x0002FF'] = 'Spacing Modifier Letters';
EnDe.rangeMap['0x000300-0x00036F'] = 'Combining Diacritical Marks';
EnDe.rangeMap['0x000370-0x0003FF'] = 'Greek/Coptic';
EnDe.rangeMap['0x000400-0x0004FF'] = 'Cyrillic';
EnDe.rangeMap['0x000500-0x00052F'] = 'Cyrillic Supplement';
EnDe.rangeMap['0x000530-0x00058F'] = 'Armenian';
EnDe.rangeMap['0x000590-0x0005FF'] = 'Hebrew';
EnDe.rangeMap['0x000600-0x0006FF'] = 'Arabic';
EnDe.rangeMap['0x000700-0x00074F'] = 'Syriac';
EnDe.rangeMap['0x000750-0x00077F'] = 'Arabic Supplement';
EnDe.rangeMap['0x000780-0x0007BF'] = 'Thaana';
EnDe.rangeMap['0x0007C0-0x0007FF'] = 'Nko';
EnDe.rangeMap['0x0008C0-0x0008FF'] = 'Undefined';
EnDe.rangeMap['0x000900-0x00097F'] = 'Devanagari';
EnDe.rangeMap['0x000980-0x0009FF'] = 'Bengali/Assamese';
EnDe.rangeMap['0x000A00-0x000A7F'] = 'Gurmukhi';
EnDe.rangeMap['0x000A80-0x000AFF'] = 'Gujarati';
EnDe.rangeMap['0x000B00-0x000B7F'] = 'Oriya';
EnDe.rangeMap['0x000B80-0x000BFF'] = 'Tamil';
EnDe.rangeMap['0x000C00-0x000C7F'] = 'Telugu';
EnDe.rangeMap['0x000C80-0x000CFF'] = 'Kannada';
EnDe.rangeMap['0x000D00-0x000D7F'] = 'Malayalam';
EnDe.rangeMap['0x000D80-0x000DFF'] = 'Sinhala';
EnDe.rangeMap['0x000E00-0x000E7F'] = 'Thai';
EnDe.rangeMap['0x000E80-0x000EFF'] = 'Lao';
EnDe.rangeMap['0x000F00-0x000FFF'] = 'Tibetan';
EnDe.rangeMap['0x001000-0x00109F'] = 'Myanmar';
EnDe.rangeMap['0x0010A0-0x0010FF'] = 'Georgian';
EnDe.rangeMap['0x001100-0x0011FF'] = 'Hangul Jamo';
EnDe.rangeMap['0x001200-0x00137F'] = 'Ethiopic';
EnDe.rangeMap['0x001380-0x00139F'] = 'Ethiopic Supplement';
EnDe.rangeMap['0x0013A0-0x0013FF'] = 'Cherokee';
EnDe.rangeMap['0x001400-0x00167F'] = 'Unified Canadian Aboriginal Syllabics';
EnDe.rangeMap['0x001680-0x00169F'] = 'Ogham';
EnDe.rangeMap['0x0016A0-0x0016FF'] = 'Runic';
EnDe.rangeMap['0x001700-0x00171F'] = 'Tagalog';
EnDe.rangeMap['0x001720-0x00173F'] = 'Hanunoo';
EnDe.rangeMap['0x001740-0x00175F'] = 'Buhid';
EnDe.rangeMap['0x001760-0x00177F'] = 'Tagbanwa';
EnDe.rangeMap['0x001780-0x0017FF'] = 'Khmer';
EnDe.rangeMap['0x001800-0x0018AF'] = 'Mongolian';
EnDe.rangeMap['0x0018B0-0x0018FF'] = 'Undefined';
EnDe.rangeMap['0x001900-0x00194F'] = 'Limbu';
EnDe.rangeMap['0x001950-0x00197F'] = 'Tai Le';
EnDe.rangeMap['0x001980-0x0019DF'] = 'New Tai Lue';
EnDe.rangeMap['0x0019E0-0x0019FF'] = 'Khmer Symbols';
EnDe.rangeMap['0x001A00-0x001A1F'] = 'Buginese';
EnDe.rangeMap['0x001A20-0x001AFF'] = '-undefined-';
EnDe.rangeMap['0x001B00-0x001B7F'] = 'Balinese';
EnDe.rangeMap['0x001B80-0x001BBF'] = 'Sundanese';
EnDe.rangeMap['0x001BC0-0x001BFF'] = '-undefined-';
EnDe.rangeMap['0x001C00-0x001C4F'] = 'Lepcha';
EnDe.rangeMap['0x001C50-0x001C7F'] = 'Ol Chiki';
EnDe.rangeMap['0x001C80-0x001CFF'] = '-undefined-';
EnDe.rangeMap['0x001D00-0x001D7F'] = 'Phonetic Extensions';
EnDe.rangeMap['0x001D80-0x001DBF'] = 'Phonetic Extensions Supplement';
EnDe.rangeMap['0x001DC0-0x001DFF'] = 'Combining Diacritical Marks Supplement';
EnDe.rangeMap['0x001E00-0x001EFF'] = 'Latin Extended Additional';
EnDe.rangeMap['0x001F00-0x001FFF'] = 'Greek Extended';
EnDe.rangeMap['0x002000-0x00206F'] = 'General Punctuation';
EnDe.rangeMap['0x002070-0x00209F'] = 'Superscripts and Subscripts';
EnDe.rangeMap['0x0020A0-0x0020CF'] = 'Currency Symbols';
EnDe.rangeMap['0x0020D0-0x0020FF'] = 'Combining Diacritical Marks for Symbols';
EnDe.rangeMap['0x002100-0x00214F'] = 'Letterlike Symbols';
EnDe.rangeMap['0x002150-0x00218F'] = 'Number Forms';
EnDe.rangeMap['0x002190-0x0021FF'] = 'Arrows';
EnDe.rangeMap['0x002200-0x0022FF'] = 'Mathematical Operators';
EnDe.rangeMap['0x002300-0x0023FF'] = 'Miscellaneous Technical';
EnDe.rangeMap['0x002400-0x00243F'] = 'Control Pictures';
EnDe.rangeMap['0x002440-0x00245F'] = 'Optical Character Recognition';
EnDe.rangeMap['0x002460-0x0024FF'] = 'Enclosed Alphanumerics';
EnDe.rangeMap['0x002500-0x00257F'] = 'Box Drawing';
EnDe.rangeMap['0x002580-0x00259F'] = 'Block Elements';
EnDe.rangeMap['0x0025A0-0x0025FF'] = 'Geometric Shapes';
EnDe.rangeMap['0x002600-0x0026FF'] = 'Miscellaneous Symbols';
EnDe.rangeMap['0x002700-0x0027BF'] = 'Dingbats';
EnDe.rangeMap['0x0027C0-0x0027EF'] = 'Miscellaneous Mathematical Symbols-A';
EnDe.rangeMap['0x0027F0-0x0027FF'] = 'Supplemental Arrows-A';
EnDe.rangeMap['0x002800-0x0028FF'] = 'Braille Patterns';
EnDe.rangeMap['0x002900-0x00297F'] = 'Supplemental Arrows-B';
EnDe.rangeMap['0x002980-0x0029FF'] = 'Miscellaneous Mathematical Symbols-B';
EnDe.rangeMap['0x002A00-0x002AFF'] = 'Supplemental Mathematical Operators';
EnDe.rangeMap['0x002B00-0x002BFF'] = 'Miscellaneous Symbols and Arrows';
EnDe.rangeMap['0x002C00-0x002C5F'] = 'Glagolitic';
EnDe.rangeMap['0x002C60-0x002C7F'] = 'Latin Extended-C';
EnDe.rangeMap['0x002C80-0x002CFF'] = 'Coptic';
EnDe.rangeMap['0x002D00-0x002D2F'] = 'Georgian Supplement';
EnDe.rangeMap['0x002D30-0x002D7F'] = 'Tifinagh';
EnDe.rangeMap['0x002D80-0x002DDF'] = 'Ethiopic Extended';
EnDe.rangeMap['0x002DE0-0x002DFF'] = 'Cyrillic Extended-A';
EnDe.rangeMap['0x002E00-0x002E7F'] = 'Supplemental Punctuation';
EnDe.rangeMap['0x002E80-0x002EFF'] = 'CJK Radicals Supplement';
EnDe.rangeMap['0x002F00-0x002FDF'] = 'Kangxi Radicals';
EnDe.rangeMap['0x002FE0-0x002FEF'] = '-undefined-';
EnDe.rangeMap['0x002FF0-0x002FFF'] = 'Ideographic Description Characters';
EnDe.rangeMap['0x003000-0x00303F'] = 'CJK Symbols and Punctuation';
EnDe.rangeMap['0x003040-0x00309F'] = 'Hiragana';
EnDe.rangeMap['0x0030A0-0x0030FF'] = 'Katakana';
EnDe.rangeMap['0x003100-0x00312F'] = 'Bopomofo';
EnDe.rangeMap['0x003130-0x00318F'] = 'Hangul Compatibility Jamo';
EnDe.rangeMap['0x003190-0x00319F'] = 'Kanbun (Kunten)';
EnDe.rangeMap['0x0031A0-0x0031BF'] = 'Bopomofo Extended';
EnDe.rangeMap['0x0031C0-0x0031EF'] = 'CJK Strokes';
EnDe.rangeMap['0x0031F0-0x0031FF'] = 'Katakana Phonetic Extensions';
EnDe.rangeMap['0x003200-0x0032FF'] = 'Enclosed CJK Letters and Months';
EnDe.rangeMap['0x003300-0x0033FF'] = 'CJK Compatibility';
EnDe.rangeMap['0x003400-0x004DBF'] = 'CJK Unified Ideographs Extension A';
EnDe.rangeMap['0x004DC0-0x004DFF'] = 'Yijing Hexagram Symbols';
EnDe.rangeMap['0x004E00-0x009FFF'] = 'CJK Unified Ideographs';
EnDe.rangeMap['0x00A000-0x00A48F'] = 'Yi Syllables';
EnDe.rangeMap['0x00A490-0x00A4CF'] = 'Yi Radicals';
EnDe.rangeMap['0x00A4D0-0x00A4FF'] = '-undefined-';
EnDe.rangeMap['0x00A500-0x00A63F'] = 'Vai';
EnDe.rangeMap['0x00A640-0x00A69F'] = 'Cyrillic Extended-B';
EnDe.rangeMap['0x00A700-0x00A71F'] = 'Modifier Tone Letters';
EnDe.rangeMap['0x00A720-0x00A7FF'] = 'Latin Extended-D';
EnDe.rangeMap['0x00A800-0x00A82F'] = 'Syloti Nagri';
EnDe.rangeMap['0x00A830-0x00A83F'] = '-undefined-';
EnDe.rangeMap['0x00A840-0x00A87F'] = 'Phags-pa';
EnDe.rangeMap['0x00A880-0x00A8DF'] = 'Saurashtra';
EnDe.rangeMap['0x00A900-0x00A92F'] = 'Kayah Li';
EnDe.rangeMap['0x00A930-0x00A95F'] = 'Rejang';
EnDe.rangeMap['0x00A960-0x00A9FF'] = '-undefined-';
EnDe.rangeMap['0x00AA00-0x00AA5F'] = 'Cham';
EnDe.rangeMap['0x00AA60-0x00ABFF'] = '-undefined-';
EnDe.rangeMap['0x00AC00-0x00D7AF'] = 'Hangul Syllables';
EnDe.rangeMap['0x00D7B0-0x00D7FF'] = '-undefined-';
EnDe.rangeMap['0x00D800-0x00DB7F'] = 'High Surrogate Area';
EnDe.rangeMap['0x00DB80-0x00DBFF'] = 'High Private Use Surrogate Area';
EnDe.rangeMap['0x00DC00-0x00DFFF'] = 'Low Surrogate Area';
EnDe.rangeMap['0x00E000-0x00E07F'] = 'Private Use Area - Conlang Tengwar';
EnDe.rangeMap['0x00E080-0x00E0FF'] = 'Private Use Area - Conlang brie.Cirth';
EnDe.rangeMap['0x00E2E0-0x00E2FF'] = 'Private Use Area - Conlang Xaini';
EnDe.rangeMap['0x00E300-0x00E33F'] = 'Private Use Area - Conlang Mizarian';
EnDe.rangeMap['0x00E340-0x00E35F'] = 'Private Use Area - Conlang Ziri:nka';
EnDe.rangeMap['0x00E360-0x00E37F'] = 'Private Use Area - Conlang Sarkai';
EnDe.rangeMap['0x00E380-0x00E3AF'] = 'Private Use Area - Conlang Thelwik';
EnDe.rangeMap['0x00E3B0-0x00E3FF'] = 'Private Use Area - Conlang Olaetyan';
EnDe.rangeMap['0x00E400-0x00E42F'] = 'Private Use Area - Conlang Niskloz';
EnDe.rangeMap['0x00E430-0x00E44F'] = 'Private Use Area - Conlang Kazat ?akkorou';
EnDe.rangeMap['0x00E450-0x00E46F'] = 'Private Use Area - Conlang Kazvarad';
EnDe.rangeMap['0x00E470-0x00E48F'] = 'Private Use Area - Conlang Zarkhand';
EnDe.rangeMap['0x00E490-0x00E4BF'] = 'Private Use Area - Conlang Rozhxh';
EnDe.rangeMap['0x00E4C0-0x00E4EF'] = 'Private Use Area - Conlang Serivelna';
EnDe.rangeMap['0x00E4F0-0x00E4FF'] = 'Private Use Area - Conlang Kelwathi';
EnDe.rangeMap['0x00E500-0x00E51F'] = 'Private Use Area - Conlang Saklor';
EnDe.rangeMap['0x00E520-0x00E54F'] = 'Private Use Area - Conlang Rynnan';
EnDe.rangeMap['0x00E550-0x00E57F'] = 'Private Use Area - Conlang Alzetjan';
EnDe.rangeMap['0x00E580-0x00E59F'] = 'Private Use Area - Conlang Telarasso';
EnDe.rangeMap['0x00E5A0-0x00E5BF'] = 'Private Use Area - Conlang Ssuraki';
EnDe.rangeMap['0x00E5C0-0x00E5DF'] = 'Private Use Area - Conlang Gargoyle';
EnDe.rangeMap['0x00E5E0-0x00E5FF'] = 'Private Use Area - Conlang Ophidian';
EnDe.rangeMap['0x00E600-0x00E62F'] = 'Private Use Area - Conlang Ferengi';
EnDe.rangeMap['0x00E630-0x00E64F'] = 'Private Use Area - Conlang Seussian Latin Extensions';
EnDe.rangeMap['0x00E650-0x00E68F'] = 'Private Use Area - ??';
EnDe.rangeMap['0x00E690-0x00E6CF'] = 'Private Use Area - Conlang Ewellic';
EnDe.rangeMap['0x00E6D0-0x00E6FF'] = 'Private Use Area - Conlang Phaistos Disc';
EnDe.rangeMap['0x00E740-0x00E76F'] = 'Private Use Area - Conlang Unifon';
EnDe.rangeMap['0x00E770-0x00E77F'] = 'Private Use Area - Conlang Solresol';
EnDe.rangeMap['0x00E780-0x00E7FF'] = 'Private Use Area - Conlang Visible Speech';
EnDe.rangeMap['0x00E800-0x00E82F'] = 'Private Use Area - Conlang Monofon';
EnDe.rangeMap['0x00E830-0x00EDFF'] = 'Private Use Area - ??';
EnDe.rangeMap['0x00EE00-0x00EEAF'] = 'Private Use Area - Hungarian Runes';
EnDe.rangeMap['0x00EEB0-0x00F89F'] = 'Private Use Area - ??';
EnDe.rangeMap['0x00F8A0-0x00F8CF'] = 'Private Use Area - Conlang Aiha';
EnDe.rangeMap['0x00F8D0-0x00F8FF'] = 'Private Use Area - Conlang Klingon';
EnDe.rangeMap['0x00E000-0x00F8FF'] = 'Private Use Area - ??';
EnDe.rangeMap['0x00F900-0x00FAFF'] = 'CJK Compatibility Ideographs';
EnDe.rangeMap['0x00FB00-0x00FB4F'] = 'Alphabetic Presentation Forms';
EnDe.rangeMap['0x00FB50-0x00FDFF'] = 'Arabic Presentation Forms-A';
EnDe.rangeMap['0x00FE00-0x00FE0F'] = 'Variation Selectors';
EnDe.rangeMap['0x00FE10-0x00FE1F'] = 'Vertical Forms';
EnDe.rangeMap['0x00FE20-0x00FE2F'] = 'Combining Half Marks';
EnDe.rangeMap['0x00FE30-0x00FE4F'] = 'CJK Compatibility Forms';
EnDe.rangeMap['0x00FE50-0x00FE6F'] = 'Small Form Variants';
EnDe.rangeMap['0x00FE70-0x00FEFF'] = 'Arabic Presentation Forms-B';
EnDe.rangeMap['0x00FF00-0x00FFEF'] = 'Halfwidth and Fullwidth Forms';
EnDe.rangeMap['0x00FFF0-0x00FFFF'] = 'Specials';
EnDe.rangeMap['0x010000-0x01007F'] = 'Linear B Syllabary';
EnDe.rangeMap['0x010080-0x0100FF'] = 'Linear B Ideograms';
EnDe.rangeMap['0x010100-0x01013F'] = 'Aegean Numbers';
EnDe.rangeMap['0x010140-0x01018F'] = 'Ancient Greek Numbers';
EnDe.rangeMap['0x010190-0x0101CF'] = 'Ancient Symbols';
EnDe.rangeMap['0x0101D0-0x0101FF'] = 'Phaistos Disc';
EnDe.rangeMap['0x010200-0x01027F'] = '-undefined-';
EnDe.rangeMap['0x010280-0x01029F'] = 'Lycian';
EnDe.rangeMap['0x0102A0-0x0102DF'] = 'Carian';
EnDe.rangeMap['0x0102E0-0x0102FF'] = '-undefined-';
EnDe.rangeMap['0x010300-0x01032F'] = 'Old Italic';
EnDe.rangeMap['0x010330-0x01034F'] = 'Gothic';
EnDe.rangeMap['0x010380-0x01039F'] = 'Ugaritic';
EnDe.rangeMap['0x0103A0-0x0103DF'] = 'Old Persian';
EnDe.rangeMap['0x010400-0x01044F'] = 'Deseret';
EnDe.rangeMap['0x010450-0x01047F'] = 'Shavian';
EnDe.rangeMap['0x010480-0x0104AF'] = 'Osmanya';
EnDe.rangeMap['0x0104B0-0x0107FF'] = '-undefined-';
EnDe.rangeMap['0x010800-0x01083F'] = 'Cypriot Syllabary';
EnDe.rangeMap['0x010840-0x0108FF'] = '-undefined-';
EnDe.rangeMap['0x010900-0x01091F'] = 'Phoenician';
EnDe.rangeMap['0x010920-0x01093F'] = 'Lydian';
EnDe.rangeMap['0x010940-0x0109FF'] = '-undefined-';
EnDe.rangeMap['0x010A00-0x010A5F'] = 'Kharoshthi';
EnDe.rangeMap['0x010A60-0x011FFF'] = '-undefined-';
EnDe.rangeMap['0x012000-0x0123FF'] = 'Cuneiform';
EnDe.rangeMap['0x012400-0x01247F'] = 'Cuneiform Numbers and Punctuation';
EnDe.rangeMap['0x012480-0x01CFFF'] = '-undefined-';
EnDe.rangeMap['0x01D000-0x01D0FF'] = 'Byzantine Musical Symbols';
EnDe.rangeMap['0x01D100-0x01D1FF'] = 'Musical Symbols';
EnDe.rangeMap['0x01D200-0x01D24F'] = 'Ancient Greek Musical Notation';
EnDe.rangeMap['0x01D250-0x01D2FF'] = '-undefined-';
EnDe.rangeMap['0x01D300-0x01D35F'] = 'Tai Xuan Jing Symbols';
EnDe.rangeMap['0x01D360-0x01D37F'] = 'Counting Rod Numerals';
EnDe.rangeMap['0x01D380-0x01D3FF'] = '-undefined-';
EnDe.rangeMap['0x01D400-0x01D7FF'] = 'Mathematical Alphanumeric Symbols';
EnDe.rangeMap['0x01D800-0x01EFFF'] = '-undefined-';
EnDe.rangeMap['0x01F000-0x01F02F'] = 'Mahjong Tiles';
EnDe.rangeMap['0x01F030-0x01F09F'] = 'Domino Tiles';
EnDe.rangeMap['0x01F0A0-0x01FFFF'] = '-undefined-';
EnDe.rangeMap['0x020000-0x02A6DF'] = 'CJK Unified Ideographs Extension B';
EnDe.rangeMap['0x02A6E0-0x02F7FF'] = '-undefined-';
EnDe.rangeMap['0x02F800-0x02FA1F'] = 'CJK Compatibility Ideographs Supplement';
EnDe.rangeMap['0x02FAB0-0x0DFFFF'] = '-unused-';
EnDe.rangeMap['0x0E0000-0x0E007F'] = 'Tags';
EnDe.rangeMap['0x0E0080-0x0E00FF'] = '-unused-';
EnDe.rangeMap['0x0E0100-0x0E01EF'] = 'Variation Selectors Supplement';
EnDe.rangeMap['0x0E01F0-0x0EFFFF'] = '-unused-';
EnDe.rangeMap['0x0F0000-0x0F0E69'] = 'Supplementary Private Use Area-A Conlang Kinya Syllables';
EnDe.rangeMap['0x0F0E70-0x0F16AF'] = 'Supplementary Private Use Area-A Conlang Pikto';
EnDe.rangeMap['0x0F16B0-0x0FFFFF'] = 'Supplementary Private Use Area-A';
EnDe.rangeMap['0x0FFFFE-0x0FFFFF'] = 'Unused';
EnDe.rangeMap['0x100000-0x10FFFD'] = 'Supplementary Private Use Area-B';

